/*****************************************************************************/
/*
                                  address.c

An address list is a string containing one or more email addresses (separated
by newline characters).  Any newline character should be preceded by a comma
(making it a comma-separated list) and therefore directly usable in an RFC2822
header field.


COPYRIGHT
---------
Copyright (C) 2005,2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
14-MAR-2006  MGD  AddressListMassage() refine address reduction (yep - again!)
                    also allow for appending a default host/domain if the
                    configuration directive [SMTP-defaulthost] is set
15-FEB-2006  MGD  AddressListMassage() refine "<address>, <address>" reduction
01-FEB-2005  MGD  initial
*/
/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* VMS related header files */
#include <ssdef.h>
#include <stsdef.h>

/* application header file */
#include <soymail.h>
#include <address.h>
#include <config.h>

#define FI_LI __FILE__, __LINE__

/* global storage */

/* prototypes */

/* external storage */

extern BOOL  Debug,
             WatchEnabled;
extern CONFIG_DATA SoyMailConfig;

/*****************************************************************************/
/*
This function appends the string supplied as 'AddrPtr' to the list of addresses
pointed to by 'ListPtr'using dynamic memory.  The memory previously allocated
by 'ListPtr' is freed.  Returns a pointer to the new list.  Expcts to be
subsequently massaged by AddressListMassage().
*/

char* AddressListAppend
(
char *ListPtr,
char *AddrPtr
)
{
   int  len;
   char  *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AddressListAppend() |%s|\n", AddrPtr);

   if (!AddrPtr || !AddrPtr[0]) return (ListPtr);

   if (ListPtr)
      len = strlen(ListPtr) + strlen(AddrPtr) + 3;
   else
      len = strlen(AddrPtr) + 1;
   sptr = calloc (1, len);
   if (ListPtr && ListPtr[0])
      sprintf (sptr, "%s,\n%s", ListPtr, AddrPtr);
   else
      strcpy (sptr, AddrPtr);
   free (ListPtr);

   if (Debug) fprintf (stdout, "|%s|\n", sptr);
   return (sptr);
}

/*****************************************************************************/
/*
This function ensures that the string list conforms to the address list
requirements described in the prologue to this module.   Returns a pointer to a
reformatted, compliant string.  Returns NULL if there was a significant problem
(e.g. buffer overflow).  If the parameter string is dynamic it is up to the
calling function to dispose of it.
*/

char* AddressListMassage
(
char* ListPtr,
BOOL CommaSeparate
)
{
   int  ccnt, lcnt,
        NewListSize;
   char  *cptr, *lptr, *nlptr, *sptr, *zptr,
         *NewListPtr;
   char  abuf [256];

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled)
      WatchThis ("MASSAGE from |!AZ", ListPtr ? ListPtr : "(null)");

   if (!(lptr = ListPtr)) return (NULL);

   NewListSize = 0;
   /* allocate an empty list (just in case) */
   NewListPtr = nlptr = calloc (8, 1);
   if (!NewListPtr) ErrorExit (vaxc$errno, FI_LI);

   lptr = ListPtr;
   while (*lptr)
   {
      /* skip leading white-space */
      while (*lptr && (*lptr == ' ' || *lptr == '\t')) lptr++;

      /* copy into the scratch buffer */
      zptr = (sptr = abuf) + sizeof(abuf);
      while (*lptr && *lptr != '\n' && *lptr != '\r' && sptr < zptr)
      {
         if (*lptr != '>')
         {
            *sptr++ = *lptr++;
            continue;
         }
         *sptr++ = *lptr++;
         if (*lptr != ',' && *lptr != ' ' && *lptr != '\t' &&
             *lptr != '\n' && *lptr != '\r') continue;
         while (*lptr == ' ' || *lptr == '\t') lptr++;
         if (*lptr != ',') continue;
         while (*lptr == ',') lptr++;
         /* looks like "<address1>, <address2>" */
         break;
      }
      if (sptr > zptr) return (NULL);
      *sptr = '\0';

      /* skip carriage-control */
      while (*lptr && (*lptr == '\n' || *lptr == '\r')) lptr++;

      /* if an empty line */
      if (!abuf[0]) continue;

      /* if the configuration specifies unqualified username is RFC address */
      if (SoyMailConfig.SmtpDefaultHost)
      {
         /* check for an RFC-style address or VMS Mail 'escape sequence' */
         for (cptr = abuf; *cptr &&
              *cptr != '@' && *(unsigned short*)cptr != '::';
              cptr++);
         if (!*cptr)
         {
            /* no, create a RFC-style address */
            if (sptr < zptr) *sptr++ = '@';
            for (cptr = SoyMailConfig.SmtpDefaultHost;
                 *cptr && sptr < zptr;
                 *sptr++ = *cptr++);
            if (sptr > zptr) return (NULL);
            *sptr = '\0';
         }
      }

      ccnt = sptr - abuf;
      NewListPtr = realloc (NewListPtr, NewListSize+ccnt+8);
      if (!NewListPtr) ErrorExit (vaxc$errno, FI_LI);
      nlptr = NewListPtr + NewListSize;

      if (nlptr > NewListPtr)
      {
         /* append a comma and newline to previous entry */
         if (CommaSeparate) *nlptr++ = ',';
         *nlptr++ = '\n';
      }
      for (cptr = abuf; *cptr; *nlptr++ = *cptr++);
      NewListSize = nlptr - NewListPtr;
   }
   *nlptr = '\0';

   AddressListPurge (NewListPtr);

   if (WatchEnabled)
      WatchThis ("MASSAGE to |!AZ", NewListPtr ? NewListPtr : "(null)");

   return (NewListPtr);
}

/*****************************************************************************/
/*
Remove exact duplicates from an address list.
Assumes it has been massaged into an address list format.
Returns the number of characters in the purged list.
*/

int AddressListPurge (char *ListPtr)

{
   char  ch1, ch2;
   char  *cptr, *lptr, *sptr, *tptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AddressListPurge() |%s|\n", ListPtr);

   if (!ListPtr) return (0);
   if (!(lptr = ListPtr)) return (0);

   while (*lptr)
   {
      /* scan to the end of the next address */
      for (cptr = lptr; *cptr && *cptr != '\n'; cptr++);
      /* if the last address in the list */
      if (!*cptr || !*(cptr+1)) break;
      ch1 = *cptr;
      *cptr++ = '\0';
      /* scan through the rest of the list comparing addresses to that one */
      while (*cptr)
      {
         for (sptr = cptr; *cptr && *cptr != '\n'; cptr++);
         ch2 = *cptr;
         *cptr = '\0';
         if (!strcmp (lptr, sptr))
         {
            /* same-same, eliminate by copying the rest over the top of it */
            *cptr = ch2;
            if (*cptr) cptr++;
            for (tptr = sptr; *cptr; *tptr++ = *cptr++);
            *tptr = '\0';
            cptr = sptr;
         }
         else
         {
            /* not the same, just resume */
            *cptr = ch2;
            if (*cptr) cptr++;
            sptr = cptr;
         }
      }
      while (*lptr) lptr++;
      *lptr = ch1;
      if (*lptr) lptr++;
   }
   if (Debug) fprintf (stdout, "AddressListPurge() |%s|\n", ListPtr);

   return (lptr - ListPtr);
}

/*****************************************************************************/
/*
Does the supplied address look like VMS style. 
*/

BOOL AddressIsVms (char* cptr)

{
   if (!cptr) return (FALSE);
   while (*cptr && isspace (*cptr)) cptr++;
   if (!*cptr) return (FALSE); 

   if (AddressIsRfc (cptr)) return (FALSE); 

   return (TRUE);
}

/*****************************************************************************/
/*
Does the supplied address look like Internet style? 
*/

BOOL AddressIsRfc (char *cptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AddressIsRfc() |%s|\n", cptr);

   if (!cptr) return (FALSE);

   while (*cptr)
   {
      if (*cptr == '\"')
      {
         /* ignore anything inside of quotes */
         cptr++;
         while (*cptr && *cptr != '\"') cptr++;
         if (*cptr) cptr++;
      }
      else
      if (*cptr == '\'')
      {
         /* ditto */
         cptr++;
         while (*cptr && *cptr != '\'') cptr++;
         if (*cptr) cptr++;
      }
      else
      {
         /* looks like a component of an Internet-style address */
         if (*cptr == '<' || *cptr == '@') return (TRUE);
         cptr++;
      }
   }
   return (FALSE);
}

/*****************************************************************************/

