/*****************************************************************************/
/*
                                 CallMail.c

All of the functions interacting with the VMS callable Mail API.


COPYRIGHT
---------
Copyright (C) 2005,2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
25-FEB-2006  MGD  CallMailGoToNewMail() if the new message count differs
                    from the messages selected count then adjust it
                  bugfix; CallMailSetUserProfile() VMS option copy-send
                  bugfix; CallMailMessageDelete() close mail file to
                    empty WASTEBASKET before continuing request
                  bugfix; CallMailMessageCopyMove() adjust new message
                    count regardless of whether it's a copy or a move
01-FEB-2005  MGD  initial
*/

/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <file.h>
#include <stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unixio.h>

/* VMS related header files */
#include <descrip.h>
#include <maildef.h>
#include <mailmsgdef.h>
#include <rms.h>
#include <ssdef.h>
#include <starlet.h>
#include <stsdef.h>

/* application header file */
#include "soymail.h"
#include "callmail.h"
#include "other.h"
#include "request.h"

#define FI_LI __FILE__, __LINE__

/* global storage */

char  CallMailNoSuchUser [] =
"An entry for this user could not be found in VMSMAIL_PROFILE (new account?)";

VMS_ITEM_LIST3  CallMailNullItem = {0,0,0,0};

VMS_ITEM_LIST3  CallMailNoSignalItem [] = {
   { 0, MAIL$_NOSIGNAL, 0, 0 },
   { 0,0,0,0 }
};

/* external storage */

extern BOOL  Debug;
extern int  VmsVersion;

/* required prototypes (some older VAXen have no MAIL$ROUTINES.H) */
int mail$message_begin (__unknown_params);
int mail$message_copy (__unknown_params);
int mail$message_delete (__unknown_params);
int mail$message_end (__unknown_params);
int mail$message_get (__unknown_params);
int mail$message_info (__unknown_params);
int mail$message_select (__unknown_params);
int mail$mailfile_begin (__unknown_params);
int mail$mailfile_close (__unknown_params);
int mail$mailfile_end (__unknown_params);
int mail$mailfile_info_file (__unknown_params);
int mail$mailfile_open (__unknown_params);
int mail$mailfile_purge_waste (__unknown_params);
int mail$user_begin (__unknown_params);
int mail$user_get_info (__unknown_params);
int mail$user_set_info (__unknown_params);
int mail$user_end (__unknown_params);

/* some older VAXen do not have these either */
#ifndef MAIL$_USER_SIGFILE
#  define MAIL$_USER_SIGFILE 3141
#endif
#ifndef MAIL$_USER_SET_SIGFILE
#  define MAIL$_USER_SET_SIGFILE 3110
#endif

/*****************************************************************************/
/*
Mail status returns may be found in 'MailVmsStatus'.

A list of 'muptr->ListMessageIdStart' messages from 'muptr->ListMessageIdStart'
is placed into a linked list of VMS_MAIL_MSG structures pointed to by
'muptr->MessageList' beginning with the most recent (larger ID numbers). 

Automatically closes the message context on an error (including
no-more-messages) so an explicit close call is not required.  It could be
closed peremtorially by setting 'MailVmsStatus' to an error value can calling.
*/

int CallMailMessageList (VMS_MAIL_USER *muptr)

{
   int  hit, status;
   long  MessageIdVector;
   unsigned long  MessageCount,
                  MessageId;
   VMS_MAIL_MSG  *vmptr, *vmlptr,
                 *PrevMessagePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CallMailMessageList() %d %d %d\n",
               muptr->ListMessageIdStart, muptr->ListMessageIdCount,
               muptr->ListMessageIdVector);

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   vmptr = &muptr->VmsMailMsg;

   if (!vmptr->MessageContext)
   {
      status = CallMailMessageContext (muptr, vmptr);
      if (VMSnok (status)) return (status);
   }

   if (muptr->ListMessageIdStart == RANGE_ALL_MSGS)
   {
      /* absolutely all messages from last to first */
      MessageId = MessageCount = vmptr->MessageSelectedCount;
      MessageIdVector = -1;
   }
   else
   if (muptr->ListMessageIdStart == RANGE_MOST_RECENT ||
       muptr->ListMessageIdStart > vmptr->MessageSelectedCount)
   {
      /* from the highest numbered message down for the count */
      MessageId = vmptr->MessageSelectedCount;
      MessageCount = muptr->ListMessageIdCount;
      MessageIdVector = -1;
   }
   else
   if (muptr->ListMessageIdVector < 0)
   {
      /* stepping from higher to lower numbers (more to less recent) */
      MessageId = muptr->ListMessageIdStart;
      MessageCount = muptr->ListMessageIdCount;
      MessageIdVector = -1;
   }
   else
   {
      /* stepping from lower to higher numbers (less to more recent) */
      MessageId = muptr->ListMessageIdStart;
      MessageCount = muptr->ListMessageIdCount;
      MessageIdVector = +1;
   }
   /* when searching this puts a limit on the number of hits before results */
   if (muptr->ListMessageIdMax) MessageCount = muptr->ListMessageIdMax;

   /********************/
   /* process messages */
   /********************/

   vmlptr = NULL;
   muptr->MessageListCount = 0;
   while (VMSok (muptr->MailVmsStatus))
   {
      if (!MessageCount || !MessageId) break;

      if (!vmlptr) vmlptr = calloc (1, sizeof(VMS_MAIL_MSG));
      if (!vmlptr)
      {
         muptr->MailVmsStatus = vaxc$errno;
         break;
      }

      vmlptr->MessageId = MessageId;
      /* either up or down */
      MessageId += MessageIdVector;

      /* borrow the message context from the main structure */
      vmlptr->MessageContext = vmptr->MessageContext;

      CallMailMessageHeader (muptr, vmlptr);
      if (VMSnok (muptr->MailVmsStatus)) break;

      if (muptr->SearchFunction)
      {
         if (vmlptr->ErrorOpenIn)
            hit = 0;
         else
         {
            CallMailMessageGet (muptr, vmlptr);
            if (VMSnok (muptr->MailVmsStatus)) break;
            hit = muptr->SearchFunction (vmlptr, muptr->SearchDataPtr);
            CallMailMessageFree (vmlptr);
            if (Debug) fprintf (stdout, "HIT: %d\n", hit);
         }

         if (!hit)
         {
            /* save potentially freeing and reallocating the memory */
            memset (vmlptr, 0, sizeof(VMS_MAIL_MSG));
            continue;
         }
      }

      /* ensure the most recent is always at the front of the list */
      if (!muptr->MessageList)
      {
         /* first entry in list */
         muptr->MessageList = vmlptr;
      }
      else
      if (MessageIdVector < 0)
      {
         /* message IDs will be decreasing, append this to the list */
         vmlptr->PrevPtr = PrevMessagePtr;
         PrevMessagePtr->NextPtr = vmlptr;
      }
      else
      {
         /* message IDs will be increasing, prepend this to the list */
         PrevMessagePtr->PrevPtr = vmlptr;
         vmlptr->NextPtr = PrevMessagePtr;
         muptr->MessageList = vmlptr;
      }
      PrevMessagePtr = vmlptr;

      /* make sure the context borrow is not propagated in any way */
      vmlptr->MessageContext = 0;

      /* this memory has been used, more will be allocated */
      vmlptr = NULL;

      /* and on to the next */
      muptr->MessageListCount++;
      MessageCount--;
   }

   /* just to be tidy */
   if (vmlptr) free (vmlptr);

   /* if we run out before reaching the end-of-range */
   if (muptr->MailVmsStatus == MAIL$_NOMOREMSG)
      muptr->MailVmsStatus = SS$_NORMAL;

   if (VMSnok (muptr->MailVmsStatus))
   {
      /*************************/
      /* close message context */
      /*************************/

      status = mail$message_end (&vmptr->MessageContext,
                                 &CallMailNullItem, &CallMailNullItem);
      if (Debug) fprintf (stdout, "mail$message_end() %%X%08.08X\n", status);
      vmptr->MessageContext = 0;
   }

   return (muptr->MailVmsStatus);
}

/*****************************************************************************/
/*
Mail status returns may be found in 'MailVmsStatus'.

Read the VMS Mail header fields from the specified message ID.

Generate a 32 bit hash 'MessageHash' from selected fields of the header and if
'ConfirmHash' is non-zero compare the two, returning SS$_BADCHECKSUM if they
are different. 
*/

int CallMailMessageHeader
(
VMS_MAIL_USER *muptr,
VMS_MAIL_MSG *vmptr
)
{
   int  status;
   VMS_ITEM_LIST3  *itptr;
   /* reminder: these require one extra element for the null item */
   VMS_ITEM_LIST3  MessageIdItem [3];
   VMS_ITEM_LIST3  MessageInfoItem [12];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailMessageHeader()\n");

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   /* default to the current message */
   if (!vmptr) vmptr = &muptr->VmsMailMsg;

   itptr = &MessageIdItem[0];
   memset (itptr, 0, sizeof(MessageIdItem));

   itptr->buf_len = sizeof(vmptr->MessageId);
   itptr->item = MAIL$_MESSAGE_ID;
   itptr->buf_addr = &vmptr->MessageId;
   itptr++;

   itptr->item = MAIL$_NOSIGNAL;

   itptr = &MessageInfoItem[0];
   memset (itptr, 0, sizeof(MessageInfoItem));

   itptr->buf_len = sizeof(vmptr->CurrentId);
   itptr->item = MAIL$_MESSAGE_CURRENT_ID;
   itptr->buf_addr = (void*)&vmptr->CurrentId;
   itptr++;

   itptr->buf_len = sizeof(vmptr->ExtId)-1;
   itptr->item = MAIL$_MESSAGE_EXTID;
   itptr->buf_addr = vmptr->ExtId;
   itptr->ret_len = (void*)&vmptr->ExtIdLength;
   itptr++;

   itptr->buf_len = sizeof(vmptr->MessageFlags);
   itptr->item = MAIL$_MESSAGE_RETURN_FLAGS;
   itptr->buf_addr = (void*)&vmptr->MessageFlags;
   itptr++;

   itptr->buf_len = sizeof(vmptr->SizeInRecords);
   itptr->item = MAIL$_MESSAGE_SIZE;
   itptr->buf_addr = (void*)&vmptr->SizeInRecords;
   itptr++;

   itptr->buf_len = sizeof(vmptr->BinDate);
   itptr->item = MAIL$_MESSAGE_BINARY_DATE;
   itptr->buf_addr = &vmptr->BinDate;
   itptr++;

   itptr->buf_len = sizeof(vmptr->Date)-1;
   itptr->item = MAIL$_MESSAGE_DATE;
   itptr->buf_addr = vmptr->Date;
   itptr->ret_len = (void*)&vmptr->DateLength;
   itptr++;

   itptr->buf_len = sizeof(vmptr->From)-1;
   itptr->item = MAIL$_MESSAGE_FROM;
   itptr->buf_addr = vmptr->From;
   itptr->ret_len = (void*)&vmptr->FromLength;
   itptr++;

   itptr->buf_len = sizeof(vmptr->Subject)-1;
   itptr->item = MAIL$_MESSAGE_SUBJECT;
   itptr->buf_addr = vmptr->Subject;
   itptr->ret_len = (void*)&vmptr->SubjectLength;
   itptr++;

   itptr->buf_len = sizeof(vmptr->To)-1;
   itptr->item = MAIL$_MESSAGE_TO;
   itptr->buf_addr = vmptr->To;
   itptr->ret_len = (void*)&vmptr->ToLength;
   itptr++;

   itptr->buf_len = sizeof(vmptr->Cc)-1;
   itptr->item = MAIL$_MESSAGE_CC;
   itptr->buf_addr = vmptr->Cc;
   itptr->ret_len = (void*)&vmptr->CcLength;
   itptr++;

   itptr->item = MAIL$_NOSIGNAL;

   status = mail$message_get (&vmptr->MessageContext,
                              &MessageIdItem,
                              &MessageInfoItem);
   if (Debug) fprintf (stdout, "mail$message_get() %%X%08.08X\n", status);

   /* if there's an external mail file and it's not found keep going */
   if (status == MAIL$_OPENIN)
   {
      vmptr->ErrorOpenIn = 1;
      status = mail$message_info (&vmptr->MessageContext,
                                  &MessageIdItem,
                                  &MessageInfoItem);
      if (Debug) fprintf (stdout, "mail$message_info() %%X%08.08X\n", status);
   }
   else
      vmptr->ErrorOpenIn = 0;

   if (VMSok (status))
   {
      vmptr->Cc[vmptr->CcLength] = '\0';
      vmptr->Date[vmptr->DateLength] = '\0';
      vmptr->From[vmptr->FromLength] = '\0';
      vmptr->Subject[vmptr->SubjectLength] = '\0';
      vmptr->To[vmptr->ToLength] = '\0';
      vmptr->ExtId[vmptr->ExtIdLength] = '\0';

      vmptr->MessageHash = CallMailHeaderHash (vmptr);

      if (Debug)
         fprintf (stdout,
"id:%d size:%d hash:%d/%d flags:0x%04.04X\n\
|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n",
                  vmptr->CurrentId, vmptr->SizeInRecords,
                  vmptr->MessageHash, vmptr->ConfirmHash,
                  vmptr->MessageFlags,
                  vmptr->Date, vmptr->ExtId, vmptr->From, vmptr->Subject,
                  vmptr->To, vmptr->Cc);

      if (vmptr->ConfirmHash &&
          vmptr->ConfirmHash != vmptr->MessageHash)
         status = SS$_BADCHECKSUM;
   }

   return (muptr->MailVmsStatus = status);
}

/*****************************************************************************/
/*
Read the message body into a dynamically allocated buffer pointed at by
'vmptr->BodyPtr' and 'vmptr->BodyLength' in length.

Mail status returns may be found in 'MailVmsStatus'.

Automatically closes the message context on an error (including NOMOREREC) so
an explicit close call is not required.  It could be closed peremtorially by
setting 'MailVmsStatus' to an error value can calling.
*/

int CallMailMessageGet
(
VMS_MAIL_USER *muptr,
VMS_MAIL_MSG *vmptr
)
{
   /* at an unknown VMS version max record size went from 255 to 2048 */
   static unsigned long  MaxRecordSize = 2048;

   static VMS_ITEM_LIST3  MessageContinueItem [] = {
      { 0, MAIL$_MESSAGE_CONTINUE, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  efd, retval, status,
        BytesAllocated;
   unsigned long  rlen;
   char  *cptr, *rptr, *sptr, *zptr;
   char  ExtFileName [256];
   stat_t  StatBuffer;
   VMS_ITEM_LIST3  *itptr;
   /* reminder: these require one extra element for the null item */
   VMS_ITEM_LIST3  MessageIdItem [3];
   VMS_ITEM_LIST3  MessageInfoItem [12];
   VMS_ITEM_LIST3  MessageRecordItem [3];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailMessageGet()\n");

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   /* default to the current message */
   if (!vmptr) vmptr = &muptr->VmsMailMsg;

   if (!vmptr->MessageContext)
   {
      /****************/
      /* open message */
      /****************/

      status = CallMailMessageContext (muptr, vmptr);
      if (VMSnok (status)) return (status);

      CallMailMessageHeader (muptr, vmptr);

      /* if it wasn't the message we were seeking */
      if (muptr->MailVmsStatus == SS$_BADCHECKSUM)
         return (muptr->MailVmsStatus);
   }
   else
      muptr->MailVmsStatus = SS$_NORMAL;

   /*******************/
   /* process message */
   /*******************/

   if (VMSok (muptr->MailVmsStatus))
   {
      /********************/
      /* buffer record(s) */
      /********************/

      if (vmptr->ExtIdLength)
      {
         zptr = (sptr = ExtFileName) + sizeof(ExtFileName)-1;
         for (cptr = muptr->VmsMailFullDirectory;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         for (cptr = vmptr->ExtId; *cptr && sptr < zptr; *sptr++ = *cptr++);
         *sptr = '\0';
         if (Debug) fprintf (stdout, "stat() |%s|\n", ExtFileName);
         if (stat (ExtFileName, &StatBuffer) < 0)
         {
            muptr->MailVmsStatus = status = vaxc$errno;
            if (Debug) fprintf (stdout, "stat() %%X%08.08X\n", status);
            BytesAllocated = 0;
         }
         else
            BytesAllocated = StatBuffer.st_size;
      }
      else
         BytesAllocated = CALLMAIL_MSG_SIZE;

      if (Debug) fprintf (stdout, "%d bytes\n", BytesAllocated);
      vmptr->BodyPtr = rptr = calloc (1, BytesAllocated+32);
      if (!vmptr->BodyPtr) muptr->MailVmsStatus = vaxc$errno;
      vmptr->BodyLength = 0;
   }

   if (VMSok (muptr->MailVmsStatus))
   {
      if (vmptr->ExtIdLength)
      {
         /**********************/
         /* external mail file */
         /**********************/

         /* avoid the overhead of mail$message_get */
         efd = open (ExtFileName, O_RDONLY, 0, "ctx=stm", "shr=get");
         if (!efd)
         {
            muptr->MailVmsStatus = vaxc$errno;
            if (Debug) fprintf (stdout, "open() %%X%08.08X\n", vaxc$errno);
         }
         else
         {
            retval = read (efd, vmptr->BodyPtr, BytesAllocated);
            if (Debug) fprintf (stdout, "retval: %d\n", retval);
            if (retval == -1)
            {
               muptr->MailVmsStatus = vaxc$errno;
               if (Debug) fprintf (stdout, "read() %%X%08.08X\n", vaxc$errno);
               close (efd);
            }
            else
            if (vmptr->MessageFlags & MAIL$M_EXTNSTD)
            {
               /* must be a fixed 512 byte foreign message */
               close (efd);
               vmptr->BodyLength = retval;
            }
            else
            {
               /* just assume it's variable-length record mode */
               close (efd);
               zptr = (sptr = rptr = vmptr->BodyPtr) + BytesAllocated;
               for (;;)
               {
                  rlen = (unsigned int)*(unsigned short*)rptr;
                  if (rlen == 0xffff) break;
                  rptr += sizeof(unsigned short);
                  memcpy (sptr, rptr, rlen);
                  sptr += rlen;
                  *sptr++ = '\n';
                  rptr += rlen;
                  /* records are always on even byte (word) boundaries */
                  if ((int)rptr & 1) rptr++;
                  /* sanity check */
                  if (rptr >= zptr) break;
               }
               *sptr = '\0';
               vmptr->BodyLength = sptr - vmptr->BodyPtr;
            }
         }
      }
      else
      {
         /************************/
         /* use MAIL$MESSAGE_GET */
         /************************/

         itptr = &MessageRecordItem[0];
         memset (itptr, 0, sizeof(MessageRecordItem));
         /* might be a longword specification but only a short is written! */
         rlen = 0;

         while (VMSok (muptr->MailVmsStatus))
         {
            itptr = &MessageRecordItem[0];

            if (BytesAllocated >= MaxRecordSize)
            itptr->buf_len = MaxRecordSize;
            else
               itptr->buf_len = BytesAllocated;
            itptr->item = MAIL$_MESSAGE_RECORD;
            itptr->buf_addr = rptr;
            itptr->ret_len = &rlen;
            itptr++;

            itptr->item = MAIL$_NOSIGNAL;

            status = mail$message_get (&vmptr->MessageContext,
                                       &MessageContinueItem,
                                       &MessageRecordItem);
            if (Debug)
               fprintf (stdout, "mail$message_get() %%X%08.08X %d/%d bytes\n",
                        status, rlen, MaxRecordSize);

            if (VMSok (status))
            {
               /* not interested in the VMS Mail header */
               if (vmptr->RecordCount++ < 5)
               {
                  rptr[0] = '\0';
                  continue;
               }
               rptr[rlen++] = '\n';
               rptr += rlen;
               rptr[0] = '\0';
               vmptr->BodyLength += rlen;
               BytesAllocated -= rlen;
               continue;
            }
            if (status == MAIL$_INVITMLEN && MaxRecordSize > 255)
            {
               /* drop back from 2048 to 255 and try again */
               MaxRecordSize = 255;
               continue;
            }
            muptr->MailVmsStatus = status;
            break;
         }
      }

      if (muptr->MailVmsStatus == MAIL$_NOMOREREC)
         muptr->MailVmsStatus = SS$_NORMAL;
   }

   if (Debug)
      fprintf (stdout, "%d |%s|\n", vmptr->BodyLength, vmptr->BodyPtr);

   if (vmptr->RecordCount >= 5)
      vmptr->RecordCount -= 5;
   else
      vmptr->RecordCount = 0;

   if (VMSnok (muptr->MailVmsStatus))
   {
      /*****************/
      /* close message */
      /*****************/

      status = mail$message_end (&vmptr->MessageContext,
                                 &CallMailNullItem, &CallMailNullItem);
      if (Debug) fprintf (stdout, "mail$message_end() %%X%08.08X\n", status);
      vmptr->MessageContext = 0;
   }

   return (muptr->MailVmsStatus);
}

/*****************************************************************************/
/*
Release memory allocated by CallMailMessageGet().
*/

void CallMailMessageFree (VMS_MAIL_MSG *vmptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailMessageFree()\n");

   if (!vmptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   free (vmptr->BodyPtr);
   vmptr->BodyPtr = NULL;
   vmptr->BodyLength = 0;
}

/*****************************************************************************/
/*
Establish a message context.
Then select message(s) based on item number, subject contents, etc.
This context can then be used to list messages, read a message, etc.
*/

int CallMailMessageContext
(
VMS_MAIL_USER *muptr,
VMS_MAIL_MSG *vmptr
)
{
   int  status;
   VMS_ITEM_LIST3  *itptr;
   /* reminder: these require one extra element for the null item */
   VMS_ITEM_LIST3  MessageBeginItem [3];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailMessageContext()\n");

   if (!muptr->VmsMailFileContext)
   {
      status = CallMailFileOpen (muptr);
      if (VMSnok (status)) return (status);
   }

   itptr = &MessageBeginItem[0];
   memset (itptr, 0, sizeof(MessageBeginItem));

   itptr->buf_len = sizeof(muptr->VmsMailFileContext);
   itptr->item = MAIL$_MESSAGE_FILE_CTX;
   itptr->buf_addr = (void*)&muptr->VmsMailFileContext;
   itptr++;

   itptr->item = MAIL$_NOSIGNAL;

   status = mail$message_begin (&vmptr->MessageContext,
                                &MessageBeginItem,
                                &CallMailNullItem);
   if (Debug) fprintf (stdout, "mail$message_begin() %%X%08.08X\n", status);

   if (VMSok (status)) status = CallMailMessageSelect (muptr);

   muptr->MailVmsStatus = status;
   return (status);
}

/*****************************************************************************/
/*
Close a message context.
*/

int CallMailMessageContextEnd
(
VMS_MAIL_USER *muptr,
VMS_MAIL_MSG *vmptr
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailMessageContextEnd()\n");

   if (!vmptr->MessageContext) return (SS$_NORMAL);;

   status = mail$message_end (&vmptr->MessageContext,
                              &CallMailNullItem, &CallMailNullItem);
   if (Debug) fprintf (stdout, "mail$message_end() %%X%08.08X\n", status);
   vmptr->MessageContext = 0;
   return (muptr->MailVmsStatus = status);
}

/*****************************************************************************/
/*
Reset all of the input items that control message selection.
*/

void CallMailMessageSelectReset (VMS_MAIL_USER *muptr)

{
   VMS_MAIL_MSG  *vmptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailMessageSelectReset()\n");

   vmptr = &muptr->VmsMailMsg;

   vmptr->CcLength = vmptr->FromLength = vmptr->SubjectLength =
      vmptr->ToLength = vmptr->TimeBeforeLength = vmptr->TimeSinceLength = 0;

   vmptr->Cc[0] = vmptr->From[0] = vmptr->Subject[0] = 
      vmptr->To[0] = vmptr->TimeBefore[0] = vmptr->TimeSince[0] = '\0';
}

/*****************************************************************************/
/*
Select message(s) based on item number, subject contents, etc.
*/

int CallMailMessageSelect (VMS_MAIL_USER *muptr)

{
   int  status;
   VMS_MAIL_MSG  *vmptr;
   VMS_ITEM_LIST3  *itptr;
   /* reminder: these require one extra element for the null item */
   VMS_ITEM_LIST3  MessageSelectItem [7];
   VMS_ITEM_LIST3  MessageSelectedItem [7];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CallMailMessageSelect() %d |%s|\n",
               muptr->VmsMailFolderNameLength, muptr->VmsMailFolderName);

   /* the current msg structure is being kept in the user data area */
   vmptr = &muptr->VmsMailMsg;

   if (!vmptr->MessageContext)
   {
      status = CallMailMessageContext (muptr, vmptr);
      if (VMSnok (status)) return (status);
   }

   /**********/
   /* select */
   /**********/

   if (!vmptr->MessageContext) ErrorExit (SS$_BUGCHECK, FI_LI);

   itptr = &MessageSelectItem[0];
   memset (itptr, 0, sizeof(MessageSelectItem));

   itptr->buf_len = muptr->VmsMailFolderNameLength;
   itptr->item = MAIL$_MESSAGE_FOLDER;
   itptr->buf_addr = muptr->VmsMailFolderName;
   itptr++;

   if (muptr->SelectNotSearch)
   {
      if (vmptr->CcLength)
      {
         itptr->buf_len = vmptr->CcLength;
         itptr->item = MAIL$_MESSAGE_CC_SUBSTRING;
         itptr->buf_addr = vmptr->Cc;
         itptr++;
      }

      if (vmptr->FromLength)
      {
         itptr->buf_len = vmptr->FromLength;
         itptr->item = MAIL$_MESSAGE_FROM_SUBSTRING;
         itptr->buf_addr = vmptr->From;
         itptr++;
      }

      if (vmptr->SubjectLength)
      {
         itptr->buf_len = vmptr->SubjectLength;
         itptr->item = MAIL$_MESSAGE_SUBJ_SUBSTRING;
         itptr->buf_addr = vmptr->Subject;
         itptr++;
      }

      if (vmptr->ToLength)
      {
         itptr->buf_len = vmptr->ToLength;
         itptr->item = MAIL$_MESSAGE_TO_SUBSTRING;
         itptr->buf_addr = vmptr->To;
         itptr++;
      }

      if (vmptr->TimeBeforeLength)
      {
         itptr->buf_len = vmptr->TimeBeforeLength;
         itptr->item = MAIL$_MESSAGE_BEFORE;
         itptr->buf_addr = vmptr->TimeBefore;
         itptr++;
      }

      if (vmptr->TimeSinceLength)
      {
         itptr->buf_len = vmptr->TimeSinceLength;
         itptr->item = MAIL$_MESSAGE_SINCE;
         itptr->buf_addr = vmptr->TimeSince;
         itptr++;
      }
   }

   itptr->item = MAIL$_NOSIGNAL;

   itptr = &MessageSelectedItem[0];
   memset (itptr, 0, sizeof(MessageSelectedItem));

   itptr->buf_len = sizeof(vmptr->MessageSelectedCount);
   itptr->item = MAIL$_MESSAGE_SELECTED;
   itptr->buf_addr = &vmptr->MessageSelectedCount;
   itptr++;

   itptr->item = MAIL$_NOSIGNAL;

   status = mail$message_select (&vmptr->MessageContext,
                                 &MessageSelectItem,
                                 &MessageSelectedItem);
   if (Debug)
      fprintf (stdout, "mail$message_select() %%X%08.08X selected: %d\n",
               status, vmptr->MessageSelectedCount);

   muptr->MailVmsStatus = status;
   return (status);
}

/*****************************************************************************/
/*
Open the Mail file in 'VmsMailFileName'.  If this contains a ':' then consider it
a full VMS file specification (e.g. USER_DISK:[ME.MAIL]TEST.MAI), otherwise a
mail file based in the current Mail directory (e.g. MAIL).  The ".MAI"
extension defaults and so is not needed.

This function tracks which mail file is currently open.  If a different file is
specified in a subsequent call the current one is closed and the new one open. 
If the same mail file then just return because it's already open.
*/

int CallMailFileOpen (VMS_MAIL_USER *muptr)

{
   int  status;
   unsigned long  MailFileNameLength;
   char  *cptr, *sptr, *zptr;
   char  Scratch [255+1];
   VMS_ITEM_LIST3  *itptr;
   /* reminder: these require one extra element for the null item */
   VMS_ITEM_LIST3  MailFileItem [3];
   VMS_ITEM_LIST3  WasteBasketItem [3];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailFileOpen()\n");

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   if (!strchr (muptr->VmsMailFileName, ':'))
   {
      /* build a file specification from a name-only */
      zptr = (sptr = Scratch) + sizeof(Scratch)-1;
      for (cptr = muptr->VmsMailFullDirectory;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      for (cptr = muptr->VmsMailFileName;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      *sptr = '\0';
      strcpy (muptr->VmsMailFileName, Scratch);
      muptr->VmsMailFileNameLength = sptr - Scratch;
   }

   if (Debug)
      fprintf (stdout, "%%X%08.08X |%s|\n",
               muptr->VmsMailFileContext, muptr->VmsMailFileName);

   if (muptr->VmsMailFileContext)
   {
      /* already open */
      muptr->MailVmsStatus = SS$_NORMAL;
      return (SS$_NORMAL);
   }

   muptr->MailDotMai = 0;

   status = mail$mailfile_begin (&muptr->VmsMailFileContext,
                                 &CallMailNullItem, &CallMailNullItem);
   if (Debug) fprintf (stdout, "mail$mailfile_begin() %%X%08.08X\n", status);
   if (VMSnok (status))
   {
      muptr->MailVmsStatus = status;
      return (status);
   }

   itptr = &MailFileItem[0];
   memset (itptr, 0, sizeof(MailFileItem));

   itptr->buf_len = muptr->VmsMailFileNameLength;
   itptr->item = MAIL$_MAILFILE_NAME;
   itptr->buf_addr = muptr->VmsMailFileName;
   itptr->ret_len = &MailFileNameLength;
   itptr++;

   itptr->item = MAIL$_NOSIGNAL;

   itptr = &WasteBasketItem[0];
   memset (itptr, 0, sizeof(WasteBasketItem));

   itptr->buf_len = sizeof(muptr->VmsMailWasteBasketName)-1;
   itptr->item = MAIL$_MAILFILE_WASTEBASKET;
   itptr->buf_addr = muptr->VmsMailWasteBasketName;
   itptr->ret_len = &muptr->VmsMailWasteBasketNameLength;

   itptr->item = MAIL$_NOSIGNAL;

   status = mail$mailfile_open (&muptr->VmsMailFileContext,
                                &MailFileItem,
                                &WasteBasketItem);
   if (Debug) fprintf (stdout, "mail$mailfile_open() %%X%08.08X\n", status);
   if (VMSok (status))
   {
      /* set a boolean if it's the primary "MAIL.MAI" */
      for (cptr = muptr->VmsMailFileName + muptr->VmsMailFileNameLength;
           cptr > muptr->VmsMailFileName && *cptr != ']';
           cptr--);
      if (*cptr == ']') cptr++;
      if (!memcmp (cptr, "MAIL\0", 5) ||
          !memcmp (cptr, "MAIL.", 5))
         muptr->MailDotMai = 1;

      muptr->VmsMailWasteBasketName[muptr->VmsMailWasteBasketNameLength] = '\0';
      if (Debug)
         fprintf (stdout, "muptr->VmsMailWasteBasketName |%s|\n",
                  muptr->VmsMailWasteBasketName);
   }
   else
   {
      /* couldn't open the mail file, end the context */
      mail$mailfile_end (&muptr->VmsMailFileContext,
                         &CallMailNullItem, &CallMailNullItem);
      if (Debug) fprintf (stdout, "mail$mailfile_end() %%X%08.08X\n", status);
      muptr->VmsMailFileContext = 0;
   }

   muptr->MailVmsStatus = status;
   return (status);
}

/*****************************************************************************/
/*
Close the currently open mail file.
If the request involved a message delete the also purge the WASTEBASKET.
*/ 

CallMailFileClose (VMS_MAIL_USER *muptr)

{
   static unsigned long  TotalReclaim;

   static VMS_ITEM_LIST3  FullCloseItem [] = {
      { 0, MAIL$_MAILFILE_FULL_CLOSE, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static VMS_ITEM_LIST3  TotalReclaimItem [] = {
      { sizeof(TotalReclaim), MAIL$_MAILFILE_TOTAL_RECLAIM, &TotalReclaim, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailFileClose()\n");

   if (!muptr->VmsMailFileContext) ErrorExit (SS$_BUGCHECK, FI_LI);

   muptr->TotalReclaimBytes = 0;

   if (muptr->VmsMailAutoPurge)
      status = mail$mailfile_close (&muptr->VmsMailFileContext,
                                    &FullCloseItem,
                                    &TotalReclaimItem);
   else
      status = mail$mailfile_close (&muptr->VmsMailFileContext,
                                    &CallMailNoSignalItem, &CallMailNullItem);
   if (Debug) fprintf (stdout, "mail$mailfile_close() %%X%08.08X\n", status);
   if (VMSnok (status))
   {
      muptr->MailVmsStatus = status;
      return (status);
   }

   muptr->TotalReclaimBytes = TotalReclaim;

   status = mail$mailfile_end (&muptr->VmsMailFileContext,
                               &CallMailNoSignalItem, &CallMailNullItem);
   if (Debug) fprintf (stdout, "mail$mailfile_end() %%X%08.08X\n", status);

   muptr->VmsMailFileContext = 0;

   muptr->MailVmsStatus = status;
   return (status);
}

/*****************************************************************************/
/*
This function retrieves the folder names in the current mail file context.
It places into the data area each null-terminated folder name, with the list
itself terminated by an empty name (i.e. "FOLDER1\0FOLDER2\0\0").
*/

int CallMailFolderList (VMS_MAIL_USER *muptr)

{
   int  status;
   VMS_ITEM_LIST3  *itptr;
   /* reminder: these require one extra element for the null item */
   VMS_ITEM_LIST3  CallbackItem [4];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailFolderList()\n");

   if (!muptr->VmsMailFileContext)
   {
      status = CallMailFileOpen (muptr);
      if (VMSnok (status)) return (status);
   }

   if (muptr->FolderListPtr)
   {
      free (muptr->FolderListPtr);
      muptr->FolderListPtr = NULL;
      muptr->FolderListCount = muptr->FolderListLength =
         muptr->FolderListSize = 0;
   }

   muptr->DataBufferLength = 0;
   muptr->DataBufferPtr = muptr->DataBuffer;

   itptr = &CallbackItem[0];
   memset (itptr, 0, sizeof(CallbackItem));

   itptr->buf_len = sizeof(void*);
   itptr->item = MAIL$_MAILFILE_FOLDER_ROUTINE;
   itptr->buf_addr = (void*)CallMailFolderListCallback;
   itptr++;

   itptr->buf_len = sizeof(VMS_MAIL_USER*);
   itptr->item = MAIL$_MAILFILE_USER_DATA;
   itptr->buf_addr = muptr;
   itptr++;

   itptr->item = MAIL$_NOSIGNAL;

   status = mail$mailfile_info_file (&muptr->VmsMailFileContext,
                                     &CallbackItem,
                                     &CallMailNullItem);
   if (Debug)
      fprintf (stdout, "mail$mailfile_info_file() %%X%08.08X\n", status);

   if (VMSnok (status))
   {
      /* ensure the list is empty in the case of an error */
      if (muptr->FolderListPtr) muptr->FolderListPtr[0] = '\0';
      muptr->FolderListLength = muptr->FolderListCount = 0;
   }

   muptr->MailVmsStatus = status;
   return (status);
}

/*****************************************************************************/
/*
Called once from CallMailFolderList() for each folder name in the current mail
file.
*/

int CallMailFolderListCallback
(
VMS_MAIL_USER *muptr,
struct dsc$descriptor *FolderDscPtr
)
{
   int  status,
        NewListSize;
   char  *cptr, *dzptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CallMailFolderListCallback() %d %*.*s\n",
               FolderDscPtr->dsc$w_length, FolderDscPtr->dsc$w_length,
               FolderDscPtr->dsc$w_length, FolderDscPtr->dsc$a_pointer);

   if (!FolderDscPtr->dsc$w_length) return (SS$_NORMAL);

   NewListSize = muptr->FolderListSize + FolderDscPtr->dsc$w_length + 1;
   if (NewListSize > muptr->FolderListSize)
   {
      muptr->FolderListSize += 1024;
      muptr->FolderListPtr = realloc (muptr->FolderListPtr,
                                      muptr->FolderListSize);
      if (!muptr->FolderListPtr) exit (vaxc$errno);
   }

   sptr = muptr->FolderListPtr + muptr->FolderListLength;
   dzptr = FolderDscPtr->dsc$a_pointer + FolderDscPtr->dsc$w_length;
   for (cptr = FolderDscPtr->dsc$a_pointer; cptr < dzptr; *sptr++ = *cptr++);
   *sptr = '\0';
   muptr->FolderListLength += FolderDscPtr->dsc$w_length + 1;
   muptr->FolderListCount++;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Copy or copy-then-delete (move) messages between folders.
*/

int CallMailMessageCopy (VMS_MAIL_USER *muptr) {
   return (CallMailMessageCopyMove (muptr, 0));
}

int CallMailMessageMove (VMS_MAIL_USER *muptr) {
   return (CallMailMessageCopyMove (muptr, 1));
}

int CallMailMessageCopyMove
(
VMS_MAIL_USER *muptr,
BOOL DeleteAfterCopy
)
{
   int  status;
   unsigned long  MsgCnt;
   char  *cptr;
   VMS_ITEM_LIST3  *itptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_MSG  VmsMailMsg;
   /* reminder: these require one extra element for the null item */
   VMS_ITEM_LIST3  MessageIdItem [6];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CallMailMessageCopyMove() |%s|\n",
               muptr->MsgIdListPtr);

   muptr->MessageCopiedCount = 0;

   vmptr = &muptr->VmsMailMsg;

   if (!vmptr->MessageContext)
   {
      /************************/
      /* open message context */
      /************************/

      status = CallMailMessageContext (muptr, vmptr);
      if (VMSnok (status)) return (status);
   }

   /**********************/
   /* process message(s) */
   /**********************/

   if (!(cptr = muptr->MsgIdListPtr)) ErrorExit (SS$_BUGCHECK, FI_LI);

   memset (&VmsMailMsg, 0, sizeof(VmsMailMsg));
   VmsMailMsg.MessageContext = vmptr->MessageContext;

   while (*cptr)
   {
      if (*cptr == '*')
         VmsMailMsg.MessageId++;
      else
      {
         VmsMailMsg.MessageId = atol(cptr);
         while (*cptr && *cptr != '.') cptr++;
         if (*cptr != '.') ErrorExit (SS$_BUGCHECK, FI_LI);
         cptr++;
         VmsMailMsg.ConfirmHash = atol(cptr);
         while (*cptr && *cptr != ',') cptr++;
         if (*cptr) cptr++;

         CallMailMessageHeader (muptr, &VmsMailMsg);

         /* if it wasn't the message we were seeking */
         if (muptr->MailVmsStatus == SS$_BADCHECKSUM) break;
      }

      itptr = &MessageIdItem[0];
      memset (itptr, 0, sizeof(MessageIdItem));

      itptr->buf_len = sizeof(VmsMailMsg.MessageId);
      itptr->item = MAIL$_MESSAGE_ID;
      itptr->buf_addr = &VmsMailMsg.MessageId;
      itptr++;

      if (muptr->VmsMailDestFileNameLength &&
          strcmp (muptr->VmsMailFileName, muptr->VmsMailDestFileName))
      {
         /* if source and destination file names differ */
         itptr->buf_len = muptr->VmsMailDestFileNameLength;
         itptr->item = MAIL$_MESSAGE_FILENAME;
         itptr->buf_addr = &muptr->VmsMailDestFileName;
         itptr++;
      }

      itptr->buf_len = muptr->VmsMailDestFolderNameLength;
      itptr->item = MAIL$_MESSAGE_FOLDER;
      itptr->buf_addr = &muptr->VmsMailDestFolderName;
      itptr++;

      if (DeleteAfterCopy)
      {
         itptr->item = MAIL$_MESSAGE_DELETE;
         itptr++;
      }

      itptr->item = MAIL$_NOSIGNAL;

      status = mail$message_copy (&VmsMailMsg.MessageContext,
                                  &MessageIdItem,
                                  &CallMailNullItem);
      if (Debug) fprintf (stdout, "mail$message_copy() %%X%08.08X\n", status);

      muptr->MailVmsStatus = status;
      if (VMSnok (status)) break;

      muptr->MessageCopiedCount++;
   }

   if (*cptr == '*' && (status == MAIL$_NOMSGS ||
                        status == MAIL$_NOMOREMSG))
      status = muptr->MailVmsStatus = SS$_NORMAL;

   /***************************************/
   /* update NEWMAIL count (if necessary) */
   /***************************************/

   MsgCnt = 0;

   if (muptr->MailDotMai && !strcmp (muptr->VmsMailFolderName, "NEWMAIL"))
      MsgCnt = 0 - muptr->MessageCopiedCount;

   if (!MsgCnt)
   {
      for (cptr = muptr->VmsMailDestFileName +
                  muptr->VmsMailDestFileNameLength;
           cptr > muptr->VmsMailDestFileName && *cptr != ']';
           cptr--);
      if (*cptr == ']') cptr++;
      if (!memcmp (cptr, "MAIL\0", 5) || !memcmp (cptr, "MAIL.", 5))
          if (!strcmp (muptr->VmsMailDestFolderName, "NEWMAIL"))
             MsgCnt = muptr->MessageCopiedCount;
   }

   if (MsgCnt)
   {
      status = CallMailSetUserNewMessages (muptr, MsgCnt);
      if (VMSok (muptr->MailVmsStatus))
         if (VMSnok (status))
            muptr->MailVmsStatus = status;
   }

   /*************************/
   /* close message context */
   /*************************/

   status = mail$message_end (&vmptr->MessageContext,
                              &CallMailNoSignalItem,
                              &CallMailNullItem);
   if (Debug) fprintf (stdout, "mail$message_end() %%X%08.08X\n", status);
   vmptr->MessageContext = 0;

   return (muptr->MailVmsStatus);
}

/*****************************************************************************/
/*
Delete the specified message(s).
*/

int CallMailMessageDelete (VMS_MAIL_USER *muptr)

{
   int  status;
   char  *cptr;
   VMS_ITEM_LIST3  *itptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_MSG  VmsMailMsg;
   /* reminder: these require one extra element for the null item */
   VMS_ITEM_LIST3  MessageIdItem [3];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "CallMailMessageDelete() |%s|\n",
               muptr->MsgIdListPtr);

   muptr->MessageDeletedCount = 0;

   vmptr = &muptr->VmsMailMsg;

   if (!vmptr->MessageContext)
   {
      /************************/
      /* open message context */
      /************************/

      status = CallMailMessageContext (muptr, vmptr);
      if (VMSnok (status)) return (status);
   }

   if (!(cptr = muptr->MsgIdListPtr)) ErrorExit (SS$_BUGCHECK, FI_LI);

   if (*cptr == '*' &&
       muptr->MailDotMai &&
       !strcmp (muptr->VmsMailFolderName, "WASTEBASKET"))
   {
      /*********************/
      /* purge wastebasket */
      /*********************/

      itptr = &MessageIdItem[0];
      memset (itptr, 0, sizeof(MessageIdItem));

      itptr->item = MAIL$_NOSIGNAL;

      status = mail$mailfile_purge_waste (&muptr->VmsMailFileContext,
                                          &MessageIdItem,
                                          &CallMailNullItem);
      if (Debug)
         fprintf (stdout, "mail$mailfile_purge_waste() %%X%08.08X\n", status);

      if (VMSok (status)) muptr->MessageDeletedCount = 999999;
      muptr->MailVmsStatus = status;
      return (muptr->MailVmsStatus);
   }

   /**********************/
   /* process message(s) */
   /**********************/

   memset (&VmsMailMsg, 0, sizeof(VmsMailMsg));
   VmsMailMsg.MessageContext = vmptr->MessageContext;

   while (*cptr)
   {
      if (*cptr == '*')
         VmsMailMsg.MessageId++;
      else
      {
         VmsMailMsg.MessageId = atol(cptr);
         while (*cptr && *cptr != '.') cptr++;
         if (*cptr != '.') ErrorExit (SS$_BUGCHECK, FI_LI);
         cptr++;
         VmsMailMsg.ConfirmHash = atol(cptr);
         while (*cptr && *cptr != ',') cptr++;
         if (*cptr) cptr++;

         CallMailMessageHeader (muptr, &VmsMailMsg);

         /* if it wasn't the message we were seeking */
         if (muptr->MailVmsStatus == SS$_BADCHECKSUM) break;
      }

      itptr = &MessageIdItem[0];
      memset (itptr, 0, sizeof(MessageIdItem));

      itptr->buf_len = sizeof(VmsMailMsg.MessageId);
      itptr->item = MAIL$_MESSAGE_ID;
      itptr->buf_addr = &VmsMailMsg.MessageId;
      itptr++;

      itptr->item = MAIL$_NOSIGNAL;

      status = mail$message_delete (&VmsMailMsg.MessageContext,
                                    &MessageIdItem,
                                    &CallMailNullItem);
      if (Debug) fprintf (stdout, "mail$message_delete() %%X%08.08X\n", status);

      muptr->MailVmsStatus = status;
      if (VMSnok (status)) break;

      muptr->MessageDeletedCount++;
   }

   if (*cptr == '*' && (status == MAIL$_NOMSGS ||
                        status == MAIL$_NOMOREMSG))
      muptr->MailVmsStatus = status = SS$_NORMAL;

   /***************************************/
   /* update NEWMAIL count (if necessary) */
   /***************************************/

   if (muptr->MailDotMai && !strcmp (muptr->VmsMailFolderName, "NEWMAIL"))
   {
      /* NEWMAIL folder in primary MAIL.MAI file */
      status = CallMailSetUserNewMessages (muptr, -muptr->MessageDeletedCount);
      if (Debug)
         fprintf (stdout, "CallMailSetUserNewMessages() %%X%08.08X\n", status);
      if (VMSok (muptr->MailVmsStatus))
         if (VMSnok (status))
            muptr->MailVmsStatus = status;
   }

   /*************************/
   /* close message context */
   /*************************/

   status = mail$message_end (&vmptr->MessageContext,
                              &CallMailNoSignalItem,
                              &CallMailNullItem);
   if (Debug) fprintf (stdout, "mail$message_end() %%X%08.08X\n", status);
   vmptr->MessageContext = 0;

   /* prevent the message complaining you can't delete from the WASTEBASKET */
   if (status == MAIL$_DELWASTE) muptr->MailVmsStatus = SS$_NORMAL;
   /* 'cause you can! */

   CallMailFileClose (muptr);

   return (muptr->MailVmsStatus);
}

/*****************************************************************************/
/*
Move the message context to standard VMS Mail "NEWMAIL" folder.
*/

int CallMailGoToNewMail (VMS_MAIL_USER *muptr)

{
   int  cnt, status;
   VMS_ITEM_LIST3  *itptr;
   VMS_MAIL_MSG  *vmptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailGoToNewMail()\n");

   vmptr = &muptr->VmsMailMsg;

   /* close any existing message context */
   if (vmptr->MessageContext)
   {
      status = mail$message_end (&vmptr->MessageContext,
                                 &CallMailNoSignalItem,
                                 &CallMailNullItem);
      if (Debug) fprintf (stdout, "mail$message_end() %%X%08.08X\n", status);
      vmptr->MessageContext = 0;
   }

   /* open message context */
   strcpy (muptr->VmsMailFileName, "MAIL");
   muptr->VmsMailFileNameLength = strlen(muptr->VmsMailFileName);

   strcpy (muptr->VmsMailFolderName, "NEWMAIL");
   muptr->VmsMailFolderNameLength = strlen(muptr->VmsMailFolderName);

   status = CallMailMessageContext (muptr, vmptr);
   if (VMSnok (status))
   {
      if (status == MAIL$_NOTEXIST)
      {
         /* NEWMAIL folder does not exist */
         muptr->MailVmsStatus = SS$_NORMAL;
         if (muptr->VmsMailNewMessages)
         {
            /* why then does it have a new message count? reset to zero! */
            CallMailSetUserNewMessages (muptr, -999999);
         }
      }
      return (muptr->MailVmsStatus);
   }
   else
   {
      if (muptr->VmsMailNewMessages != vmptr->MessageSelectedCount)
      {
         /* if the counts don't match then adjust by difference */
         cnt = -(muptr->VmsMailNewMessages - vmptr->MessageSelectedCount);
         CallMailSetUserNewMessages (muptr, cnt);
      }
   }

   return (muptr->MailVmsStatus);
}

/*****************************************************************************/
/*
This function retrieves user profile information.
*/

int CallMailGetUserProfile (VMS_MAIL_USER *muptr)

{
   int  status;
   unsigned long  UserNameLength;
   VMS_ITEM_LIST3  *itptr;
   /* reminder: these require one extra element for the null item */
   VMS_ITEM_LIST3  ProfileItem [14];
   VMS_ITEM_LIST3  SigFileItem [3];
   VMS_ITEM_LIST3  UserNameItem [3];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "VmsMailGetUserProfile() |%s|\n", muptr->UserName);

   /* if the user context has already been established */
   if (muptr->VmsMailUserContext) return (SS$_NORMAL);

   itptr = &UserNameItem[0];
   memset (itptr, 0, sizeof(UserNameItem));

   itptr->buf_len = muptr->UserNameLength;
   itptr->item = MAIL$_USER_USERNAME;
   itptr->buf_addr = muptr->UserName;
   itptr->ret_len = &UserNameLength;
   itptr++;

   itptr->item = MAIL$_NOSIGNAL;

   if (!muptr->VmsMailUserContext)
   {
      status = mail$user_begin (&muptr->VmsMailUserContext,
                                &CallMailNullItem, &CallMailNullItem);
      if (Debug) fprintf (stdout, "mail$user_begin() %%X%08.08X\n", status);
      if (VMSnok (status))
      {
         muptr->MailVmsStatus = status;
         return (status);
      }
   }

   itptr = &ProfileItem[0];
   memset (itptr, 0, sizeof(ProfileItem));

   itptr->buf_len = sizeof(muptr->VmsMailNewMessages);
   itptr->item = MAIL$_USER_NEW_MESSAGES;
   itptr->buf_addr = &muptr->VmsMailNewMessages;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailAutoPurge);
   itptr->item = MAIL$_USER_AUTO_PURGE;
   itptr->buf_addr = &muptr->VmsMailAutoPurge;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailCcPrompt);
   itptr->item = MAIL$_USER_CC_PROMPT;
   itptr->buf_addr = &muptr->VmsMailCcPrompt;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailCopyForward);
   itptr->item = MAIL$_USER_COPY_FORWARD ;
   itptr->buf_addr = &muptr->VmsMailCopyForward;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailCopyReply);
   itptr->item = MAIL$_USER_COPY_REPLY;
   itptr->buf_addr = &muptr->VmsMailCopyReply;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailCopySend);
   itptr->item = MAIL$_USER_COPY_SEND;
   itptr->buf_addr = &muptr->VmsMailCopySend;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailForm)-1;
   itptr->item = MAIL$_USER_FORM;
   itptr->buf_addr = muptr->VmsMailForm;
   itptr->ret_len = &muptr->VmsMailFormLength;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailEditor)-1;
   itptr->item = MAIL$_USER_EDITOR;
   itptr->buf_addr = muptr->VmsMailEditor;
   itptr->ret_len = &muptr->VmsMailEditorLength;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailForwarding)-1;
   itptr->item = MAIL$_USER_FORWARDING;
   itptr->buf_addr = muptr->VmsMailForwarding;
   itptr->ret_len = &muptr->VmsMailForwardingLength;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailFullDirectory)-1;
   itptr->item = MAIL$_USER_FULL_DIRECTORY;
   itptr->buf_addr = muptr->VmsMailFullDirectory;
   itptr->ret_len = &muptr->VmsMailFullDirectoryLength;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailPersonalName)-1;
   itptr->item = MAIL$_USER_PERSONAL_NAME;
   itptr->buf_addr = muptr->VmsMailPersonalName;
   itptr->ret_len = &muptr->VmsMailPersonalNameLength;
   itptr++;

   itptr->buf_len = sizeof(muptr->VmsMailQueue)-1;
   itptr->item = MAIL$_USER_QUEUE;
   itptr->buf_addr = muptr->VmsMailQueue;
   itptr->ret_len = &muptr->VmsMailQueueLength;
   itptr++;

   itptr->item = MAIL$_NOSIGNAL;

   status = mail$user_get_info (&muptr->VmsMailUserContext,
                                &UserNameItem,
                                &ProfileItem);
   if (Debug) fprintf (stdout, "mail$user_get_info() %%X%08.08X\n", status);

   if (VMSok (status))
   {
      muptr->VmsMailEditor[muptr->VmsMailEditorLength] = '\0';
      muptr->VmsMailForwarding[muptr->VmsMailForwardingLength] = '\0';
      muptr->VmsMailForm[muptr->VmsMailFormLength] = '\0';
      muptr->VmsMailFullDirectory[muptr->VmsMailFullDirectoryLength] = '\0';
      muptr->VmsMailPersonalName[muptr->VmsMailPersonalNameLength] = '\0';
      muptr->VmsMailQueue[muptr->VmsMailQueueLength] = '\0';
      muptr->VmsMailAutoPurge = muptr->VmsMailAutoPurge & 1;
      muptr->VmsMailCcPrompt = muptr->VmsMailCcPrompt & 1;
      muptr->VmsMailCopyForward = muptr->VmsMailCopyForward & 1;
      muptr->VmsMailCopyReply = muptr->VmsMailCopyReply & 1;
      muptr->VmsMailCopySend = muptr->VmsMailCopySend & 1;
      if (Debug)
         fprintf (stdout,
"%d %d %d %d %d %d\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n|%s|\n",
                 muptr->VmsMailNewMessages,
                 muptr->VmsMailAutoPurge,
                 muptr->VmsMailCcPrompt,
                 muptr->VmsMailCopyForward,
                 muptr->VmsMailCopyReply, 
                 muptr->VmsMailCopySend, 
                 muptr->VmsMailEditor,
                 muptr->VmsMailForm,
                 muptr->VmsMailForwarding,
                 muptr->VmsMailFullDirectory,
                 muptr->VmsMailPersonalName,
                 muptr->VmsMailQueue);
   }

   muptr->VmsMailSigFile[0] = '\0';
   muptr->VmsMailSigFileLength = 0;

   if (VmsVersion >= 700 && VMSok (status))
   {
      itptr = &SigFileItem[0];
      memset (itptr, 0, sizeof(SigFileItem));

      itptr->buf_len = sizeof(muptr->VmsMailSigFile)-1;
      itptr->item = MAIL$_USER_SIGFILE;
      itptr->buf_addr = muptr->VmsMailSigFile;
      itptr->ret_len = &muptr->VmsMailSigFileLength;
      itptr++;

      itptr->item = MAIL$_NOSIGNAL;

      status = mail$user_get_info (&muptr->VmsMailUserContext,
                                   &UserNameItem,
                                   &SigFileItem);
      if (Debug) fprintf (stdout, "mail$user_get_info() %%X%08.08X\n", status);
      if (VMSok (status))
      {
         muptr->VmsMailSigFile[muptr->VmsMailSigFileLength] = '\0';
         if (Debug) fprintf (stdout, "|%s|\n", muptr->VmsMailSigFile);
      }
      status = SS$_NORMAL;
   }
   else
   {
      /* just substitute a fallback signature file name */
      strcpy (muptr->VmsMailSigFile, SOY_SIGFILE);
      muptr->VmsMailSigFileLength = strlen(muptr->VmsMailSigFile);
   }

   /* leave the user context established until the process exits */
   mail$user_end (&muptr->VmsMailUserContext,
                  &CallMailNoSignalItem, &CallMailNullItem);
   muptr->VmsMailUserContext = 0;

   muptr->MailVmsStatus = status;
   return (status);
}

/*****************************************************************************/
/*
This function sets user profile information.
It's expects all relevant fields to have been populated.
*/

int CallMailSetUserProfile (VMS_MAIL_USER *muptr)

{
   int  status;
   unsigned long  UserNameLength;
   VMS_ITEM_LIST3  *itptr;
   /* reminder: these require one extra element for the null item */
   VMS_ITEM_LIST3 ProfileItem [13];
   VMS_ITEM_LIST3 SigFileItem [4];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "VmsMailSetUserProfile()\n");

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);
   if (muptr->VmsMailUserContext) ErrorExit (SS$_BUGCHECK, FI_LI);

   if (!muptr->VmsMailUserContext)
   {
      status = mail$user_begin (&muptr->VmsMailUserContext,
                                &CallMailNullItem, &CallMailNullItem);
      if (Debug) fprintf (stdout, "mail$user_begin() %%X%08.08X\n", status);
      if (VMSnok (status)) return (status);
   }

   itptr = &ProfileItem[0];
   memset (itptr, 0, sizeof(ProfileItem));

   itptr->buf_len = muptr->UserNameLength;
   itptr->item = MAIL$_USER_USERNAME;
   itptr->buf_addr = muptr->UserName;
   itptr->ret_len = &UserNameLength;
   itptr++;

   if (muptr->VmsMailAutoPurge)
      itptr->item = MAIL$_USER_SET_AUTO_PURGE;
   else
      itptr->item = MAIL$_USER_SET_NO_AUTO_PURGE;
   itptr++;

   if (muptr->VmsMailCcPrompt)
      itptr->item = MAIL$_USER_SET_CC_PROMPT;
   else
      itptr->item = MAIL$_USER_SET_NO_CC_PROMPT;
   itptr++;

   if (muptr->VmsMailCopyForward)
      itptr->item = MAIL$_USER_SET_COPY_FORWARD ;
   else
      itptr->item = MAIL$_USER_SET_NO_COPY_FORWARD ;
   itptr++;

   if (muptr->VmsMailCopyReply)
      itptr->item = MAIL$_USER_SET_COPY_REPLY;
   else
      itptr->item = MAIL$_USER_SET_NO_COPY_REPLY;
   itptr++;

   if (muptr->VmsMailCopySend)
      itptr->item = MAIL$_USER_SET_COPY_SEND;
   else
      itptr->item = MAIL$_USER_SET_NO_COPY_SEND;
   itptr++;

   if (muptr->VmsMailFormLength)
   {
      itptr->buf_len = muptr->VmsMailFormLength;
      itptr->item = MAIL$_USER_SET_FORM;
      itptr->buf_addr = muptr->VmsMailForm;
   }
   else
      itptr->item = MAIL$_USER_SET_NO_FORM;
   itptr++;

   if (muptr->VmsMailEditorLength)
   {
      itptr->buf_len = muptr->VmsMailEditorLength;
      itptr->item = MAIL$_USER_SET_EDITOR;
      itptr->buf_addr = muptr->VmsMailEditor;
   }
   else
      itptr->item = MAIL$_USER_SET_NO_EDITOR;
   itptr++;

   if (muptr->VmsMailForwardingLength)
   {
      itptr->buf_len = muptr->VmsMailForwardingLength;
      itptr->item = MAIL$_USER_SET_FORWARDING;
      itptr->buf_addr = muptr->VmsMailForwarding;
   }
   else
      itptr->item = MAIL$_USER_SET_NO_FORWARDING;
   itptr++;

   if (muptr->VmsMailPersonalNameLength)
   {
      itptr->buf_len = muptr->VmsMailPersonalNameLength;
      itptr->item = MAIL$_USER_SET_PERSONAL_NAME;
      itptr->buf_addr = muptr->VmsMailPersonalName;
   }
   else
      itptr->item = MAIL$_USER_SET_NO_PERSONAL_NAME;
   itptr++;

   if (muptr->VmsMailQueueLength)
   {
      itptr->buf_len = muptr->VmsMailQueueLength;
      itptr->item = MAIL$_USER_SET_QUEUE;
      itptr->buf_addr = muptr->VmsMailQueue;
   }
   else
      itptr->item = MAIL$_USER_SET_NO_QUEUE;
   itptr++;

   itptr->item = MAIL$_NOSIGNAL;

   status = mail$user_set_info (&muptr->VmsMailUserContext,
                                &ProfileItem,
                                &CallMailNullItem);
   if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);

   if (VmsVersion >= 700 && VMSok (status))
   {
      itptr = &SigFileItem[0];
      memset (itptr, 0, sizeof(SigFileItem));

      itptr->buf_len = muptr->UserNameLength;
      itptr->item = MAIL$_USER_USERNAME;
      itptr->buf_addr = muptr->UserName;
      itptr->ret_len = &UserNameLength;
      itptr++;

      itptr->buf_len = muptr->VmsMailSigFileLength;
      itptr->item = MAIL$_USER_SET_SIGFILE;
      itptr->buf_addr = muptr->VmsMailSigFile;
      itptr++;

      itptr->item = MAIL$_NOSIGNAL;

      status = mail$user_set_info (&muptr->VmsMailUserContext,
                                   &SigFileItem,
                                   &CallMailNullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);
   }

   /* leave the user context established until the process exits */
   mail$user_end (&muptr->VmsMailUserContext,
                  &CallMailNoSignalItem, &CallMailNullItem);
   muptr->VmsMailUserContext = 0;

   muptr->MailVmsStatus = status;
   return (status);
}

/*****************************************************************************/
/*
If messages are moved to or from, or deleted from the NEWMAIL folder update the
count of new messages.
*/

CallMailSetUserNewMessages
(
VMS_MAIL_USER *muptr,
int DeltaValue
)
{
   static short  NewMessages;

   static VMS_ITEM_LIST3  UserGetItem [] = {
      { sizeof(NewMessages), MAIL$_USER_NEW_MESSAGES, &NewMessages, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static VMS_ITEM_LIST3  UserSetItem [] = {
      { sizeof(NewMessages), MAIL$_USER_SET_NEW_MESSAGES, &NewMessages, 0 },
      { 0, MAIL$_USER_USERNAME, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   static VMS_ITEM_LIST3  UserNameItem [] = {
      { 0, MAIL$_USER_USERNAME, 0, 0 },
      { 0, MAIL$_NOSIGNAL, 0, 0 },
      { 0,0,0,0 }
   };

   int  status;
   unsigned long  UserContext;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailSetUserNewMessages()\n");

   UserContext = 0;

   status = mail$user_begin (&UserContext,
                             &CallMailNullItem, &CallMailNullItem);
   if (Debug) fprintf (stdout, "mail$user_begin() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   UserNameItem[0].buf_len = muptr->UserNameLength;
   UserNameItem[0].buf_addr = muptr->UserName;

   status = mail$user_get_info (&UserContext,
                                &UserNameItem,
                                &UserGetItem);
   if (Debug) fprintf (stdout, "mail$user_get_info() %%X%08.08X\n", status);

   if (VMSok (status))
   {
      if (Debug)
         fprintf (stdout, "NewMessages: %d (%d) %d\n",
                  NewMessages, DeltaValue, NewMessages+DeltaValue);
      NewMessages += DeltaValue;
      if (NewMessages < 0) NewMessages = 0;

      UserSetItem[1].buf_addr = muptr->UserName;
      UserSetItem[1].buf_len = muptr->UserNameLength;

      status = mail$user_set_info (&UserContext,
                                   &UserSetItem,
                                   &CallMailNullItem);
      if (Debug) fprintf (stdout, "mail$user_set_info() %%X%08.08X\n", status);

      if (VMSok (status)) muptr->VmsMailNewMessages = NewMessages;
   }

   mail$user_end (&UserContext, &CallMailNoSignalItem, &CallMailNullItem);

   return (status);
}

/****************************************************************************/
/*
This function retrieves the the mail file names in the user's Mail directory.
It places into the data area each null-terminated file name (i.e.
"FILE1\0FILE22\0"), and a count in 'MailFileListCount'.

CAUTION: RMS processing of these *.MAI to get the few mail files (as contrasted
to message files) can be VERY EXPENSIVE in disk activity and can introduce
noticable latency into folder listings.  This is why it has been done at this
low-level.  Can temporarily chew up a bit of memory.
*/ 

CallMailFileList (VMS_MAIL_USER *muptr)

{
   int  bytes, dfd, retval, rlen, status,
        FileCount,
        NameLength,
        NewListSize;
   char  *bptr, *cptr, *dzptr, *rptr, *sptr, *zptr,
         *BufferPtr,
         *NamePtr;
   char  DirFileName [256+128];
   stat_t  StatBuffer;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailFileList()\n");

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   /* a clumsy way of getting to the mail directory file via itself */
   zptr = (sptr = DirFileName) + sizeof(DirFileName)-1;
   for (cptr = muptr->VmsMailFullDirectory;
        *cptr && sptr < zptr-128;
        *sptr++ = *cptr++);
   *sptr = '\0';
   while (sptr > muptr->VmsMailFullDirectory && *sptr != ']') sptr--;
   if (*sptr != ']') exit (SS$_BUGCHECK);
   cptr = sptr;
   while (cptr > muptr->VmsMailFullDirectory && *cptr != '[')
   {
      if (*cptr == '.' && *(cptr-1) != '^') break;
      cptr--;
   }
   if (*cptr != '[' && *cptr != '.') exit (SS$_BUGCHECK);
   cptr++;
   *sptr++ = '.';
   *sptr++ = '-';
   *sptr++ = ']';
   while (*cptr != '.' && sptr < zptr)
   {
      if (*cptr == '^') cptr++;
      *sptr++ = *cptr++;
   }
   for (cptr = ".DIR"; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr > zptr) exit (SS$_BUGCHECK);
   *sptr = '\0';
   if (Debug) fprintf (stdout, "DirMailFile |%s|\n", DirFileName);

   if (muptr->MailFileListPtr)
   {
      free (muptr->MailFileListPtr);
      muptr->MailFileListPtr = NULL;
      muptr->MailFileListCount = muptr->MailFileListLength =
         muptr->MailFileListSize = 0;
   }

   if (stat (DirFileName, &StatBuffer) < 0)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "stat() %%X%08.08X\n", status);
      return (status);
   }

   bytes = StatBuffer.st_size;
   if (Debug) fprintf (stdout, "%d/%d bytes/blocks\n", bytes, bytes/512);

   dfd = open (DirFileName, O_RDONLY, 0, "ctx=stm", "shr=get");
   if (!dfd)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "open() %%X%08.08X\n", status);
      return (status);
   }

   BufferPtr = calloc (bytes, 1);
   if (!BufferPtr)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "calloc() %%X%08.08X\n", status);
      close (dfd);
      return (status);
   }

   retval = read (dfd, BufferPtr, bytes);
   if (Debug) fprintf (stdout, "retval: %d\n", retval);
   if (retval == -1)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "read() %%X%08.08X\n", status);
      close (dfd);
      return (status);
   }
   close (dfd);

   rlen = 0;
   rptr = bptr = BufferPtr;

   for (;;)
   {
      rptr += rlen;
      /* records are always on even byte (word) boundaries */
      if ((int)rptr & 1) rptr++;
      for (;;)
      {
         /* directory files have no-span record attributes */
         rlen = (unsigned int)*(unsigned short*)rptr;
         if (rlen != 0xffff) break;
         rptr = (bptr += 512);
         if (bptr >= BufferPtr + bytes) break;
      }
      if (bptr >= BufferPtr + bytes) break;
      rptr += sizeof(unsigned short);
      NameLength = (*(rptr+2) << 8) + *(rptr+3);
      NamePtr = rptr + 4;
      if (Debug)
         fprintf (stdout, "%d %02d %02d |%*.*s|\n",
                  (bptr-BufferPtr)/512+1, rlen,
                  NameLength, NameLength, NameLength, NamePtr);

      /* not much point proceding with this one */
      if (NameLength <= 4) continue;

      /* if it doesn't have a .MAI file type */
      cptr = NamePtr + NameLength - 4;
      if (*(unsigned long*)cptr != '.MAI') continue;

      /* if it's an externally stored mail message */
      if (NameLength == 21+4 && !strncmp (NamePtr, "MAIL$", 5)) continue;

      NewListSize = muptr->MailFileListSize + NameLength;
      if (NewListSize > muptr->MailFileListSize)
      {
         muptr->MailFileListSize += 256;
         muptr->MailFileListPtr = realloc (muptr->MailFileListPtr,
                                           muptr->MailFileListSize);
         if (!muptr->MailFileListPtr) exit (vaxc$errno);
      }

      /* terminate on the period (assigned above) */
      *cptr = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", NamePtr);

      sptr = muptr->MailFileListPtr + muptr->MailFileListLength;
      zptr = muptr->MailFileListPtr + muptr->MailFileListSize;
      for (cptr = NamePtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '\0';
      if (sptr >= zptr) exit (SS$_BUGCHECK);
      muptr->MailFileListCount++;
      muptr->MailFileListLength = sptr - muptr->MailFileListPtr;
   }

   free (BufferPtr);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
This function retrieves the the mail file names in the user's Mail directory.
It places into the data area each null-terminated file name (i.e.
"FILE1\0FILE22\0"), and a count in 'MailFileListCount'.

CAUTION: Listing *.MAI to get the few mail files (as contrasted to message
files) can be VERY EXPENSIVE in disk activity and can introduce noticable
latency into folder listings.  This is why it should only be done on demand.
*/

#ifdef FAR_TOO_EXPENSIVE  /* see my implmentation immediately above */

CallMailFileList (VMS_MAIL_USER *muptr)

{
   int  status,
        FileCount,
        NewListSize;
   char  *cptr, *dzptr, *sptr, *zptr;
   char  Directory [255+1],
         ExpFileName [255+1],
         ResFileName [255+1];
   struct FAB  SearchFab;
   struct NAM  SearchNam;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailFileList()\n");

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   if (muptr->MailFileListPtr)
   {
      free (muptr->MailFileListPtr);
      muptr->MailFileListPtr = NULL;
      muptr->MailFileListCount = muptr->MailFileListLength =
         muptr->MailFileListSize = 0;
   }

   SearchFab = cc$rms_fab;
   SearchFab.fab$l_dna = muptr->VmsMailFullDirectory;
   SearchFab.fab$b_dns = muptr->VmsMailFullDirectoryLength;
   SearchFab.fab$l_fna = "*.MAI;";
   SearchFab.fab$b_fns = 6;
   SearchFab.fab$l_nam = &SearchNam;

   SearchNam = cc$rms_nam;
   SearchNam.nam$l_esa = ExpFileName;
   SearchNam.nam$b_ess = sizeof(ExpFileName)-1;
   SearchNam.nam$l_rsa = ResFileName;
   SearchNam.nam$b_rss = sizeof(ResFileName)-1;

   if (VMSnok (status = sys$parse (&SearchFab, 0, 0)))
      return (status);
   SearchNam.nam$l_ver[SearchNam.nam$b_ver] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", ExpFileName);

   FileCount = 0;
   while (VMSok (status = sys$search (&SearchFab, 0, 0)))
   {
      SearchNam.nam$l_ver[SearchNam.nam$b_ver] = '\0';
      if (Debug) fprintf (stdout, "%d |%s|\n", FileCount, ResFileName);

      /* if it's an externally stored mail message */
      if (SearchNam.nam$b_name == 21 &&
          !strncmp (SearchNam.nam$l_name, "MAIL$", 5)) continue;

      SearchNam.nam$l_name[SearchNam.nam$b_name] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", SearchNam.nam$l_name);

      NewListSize = muptr->MailFileListSize + strlen(SearchNam.nam$l_name)+1;
      if (NewListSize > muptr->MailFileListSize)
      {
         muptr->MailFileListSize += 1024;
         muptr->MailFileListPtr = realloc (muptr->MailFileListPtr,
                                           muptr->MailFileListSize);
         if (!muptr->MailFileListPtr) exit (vaxc$errno);
      }

      sptr = muptr->MailFileListPtr + muptr->MailFileListLength;
      zptr = muptr->MailFileListPtr + muptr->MailFileListSize;
      for (cptr = SearchNam.nam$l_name;
           *cptr && sptr < zptr;
           *sptr++ = *cptr++);
      if (sptr < zptr) *sptr++ = '\0';
      if (sptr >= zptr) exit (SS$_BUGCHECK);
      muptr->MailFileListCount++;
      muptr->MailFileListLength = sptr - muptr->MailFileListPtr;

      SearchNam.nam$l_name[SearchNam.nam$b_name] = '.';
   }
   if (status == RMS$_FNF || status == RMS$_NMF) status = SS$_NORMAL;
   if (VMSnok (status))
   {
      /* ensure the list is empty in the case of an error */
      if (muptr->MailFileListPtr) muptr->MailFileListPtr[0] = '\0';
      muptr->MailFileListLength = muptr->MailFileListCount = 0;
   }
   return (status);
}

#endif  /* FAR_TOO_EXPENSIVE */

/*****************************************************************************/
/*
Supplied with a file name construct the full specification based on the user's
Mail directory and return a pointer to that.  This function massages the
file name to be ODS-2 compliant.
*/

char* CallMailFileIn
(
VMS_MAIL_USER *muptr,
char *NamePrefix,
char *FileName
)
{
   /* some elbow room so we can do things like 'strcat(ptr,";-3")' */
   static char  FullFileName [256+16];

   char  *cptr, *pptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CallMailFileIn() |%s|\n", FileName);

   zptr = (sptr = FullFileName) + 255;
   *sptr = '\0';
   for (cptr = muptr->VmsMailFullDirectory;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++);
   if (NamePrefix)
      for (cptr = NamePrefix; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (FileName)
   {
      pptr = NULL;
      for (cptr = FileName; *cptr && sptr < zptr; cptr++)
      {
         if (*cptr == '.') pptr = sptr;
         if (isalnum(*cptr)) *sptr++ = *cptr; else *sptr++ = '_';
      }
      if (pptr) *pptr = '.';
   }
   if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   *sptr = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", FullFileName);
   return (FullFileName);
}

/*****************************************************************************/
/*
Generates and returns a 32 bit hash of the major components of the VMS mail
message.  This is intended to help ensure that the message specified by
the 'MessageId' number is indeed the message intended.
*/

unsigned int CallMailHeaderHash (VMS_MAIL_MSG *vmptr)

{
   unsigned int  hash32;

   /*********/
   /* begin */
   /*********/

   CallMailElfHash (NULL, 0);  /* initialize */
   CallMailElfHash (vmptr->ExtId, vmptr->ExtIdLength);
   CallMailElfHash (vmptr->Date, vmptr->DateLength);
   CallMailElfHash (vmptr->From, vmptr->FromLength);
   CallMailElfHash (vmptr->To, vmptr->ToLength);
   hash32 = CallMailElfHash (vmptr->Subject, vmptr->SubjectLength);

   return (hash32);
}

/****************************************************************************/
/*
Hashing function that allows the hash to be developed over multiple calls. It
must be initialized before each hash generation with a CallMailElfHash(NULL,0).
Based on code by Arash Partow. 
*/ 

unsigned int CallMailElfHash
(
char *str,
int len
)
{
   static unsigned int  hash;

   unsigned int  x, i;

   /*********/
   /* begin */
   /*********/

   /* initialize */
   if (!str) return (hash = 0);

   for (i = 0; i < len; str++, i++)
   {
      hash = (hash << 4) + (*str);
      if ((x = hash & 0xF0000000L) != 0)
      {
         hash ^= (x >> 24);
         hash &= ~x;
      }
   }

   return (hash & 0x7FFFFFFF);
}

/*****************************************************************************/

