/******************************************************************************/
/*
                                 callmail.h
*/
/******************************************************************************/

#ifndef INCLUDE_CALLMAIL_H
#define INCLUDE_CALLMAIL_H

#include <descrip.h>

#include "mime.h"

#define CALLMAIL_MSG_SIZE 512*3  /* internal to MAIL.MAI = 3 blocks */

/******************************/
/* VMS mail message structure */
/******************************/

/*
Contains all of the information required to specify, represent and/or
manipulate a single VMS Mail based message.
*/

typedef struct STRUCT_VMS_MAIL_MSG
{
   struct STRUCT_VMS_MAIL_MSG  *PrevPtr,
                               *NextPtr;

   unsigned long  ConfirmHash,
                  CurrentId,
                  ErrorOpenIn,
                  HashMisMatch,
                  MessageContext,
                  MessageHash,
                  MessageId,
                  MessageSelectedCount,
                  RecordCount,
                  SearchNotHit,
                  SizeInRecords,
                  VmsForeign;

   unsigned long  BinDate [2];
   unsigned short  MessageFlags;

   char  *MessagePtr;
   unsigned long  MessageLength;

   char  *BodyPtr;
   unsigned long  BodyLength;

   char  Cc [255+1];
   unsigned long  CcLength;
   char  Date [23+1];
   unsigned long  DateLength;
   char  From [255+1];
   unsigned long  FromLength;
   char  Subject [255+1];
   unsigned long  SubjectLength;
   char  To [255+1];
   unsigned long  ToLength;
   char  TimeBefore [31+1];
   unsigned long  TimeBeforeLength;
   char  TimeSince [31+1];
   unsigned long  TimeSinceLength;
   char  ExtId [255+1];
   unsigned long  ExtIdLength;

   BOOL  LooksLikeRfc;
   char  *RfcHeaderPtr;
   int  RfcHeaderLength;
   char  *RfcBodyPtr;
   int  RfcBodyLength;

   char  *RfcFromPtr;
   int  RfcFromLength;
   char  *RfcToPtr;
   int  RfcToLength;
   char  *RfcCcPtr;
   int  RfcCcLength;
   char  *RfcSubjPtr;
   int  RfcSubjLength;
   char  *RfcDatePtr;
   int  RfcDateLength;
   char  *RfcReplyToPtr;
   int  RfcReplyToLength;
   char  *RfcReturnPathPtr;
   int  RfcReturnPathLength;

   /*
      These point to the actual data depending on whether it's native VMS
      Mail or one containing as leading RFC822 header.  They are set when the
      message body is scanned for the presence of an Internet mail header.
   */
   char  *MsgFromPtr,
         *MsgToPtr,
         *MsgCcPtr,
         *MsgSubjPtr,
         *MsgDatePtr,
         *MsgBodyPtr;

   /* a pointer to dynamically allocated MIME structure */
   MIME_DATA  *MimeDataPtr;

} VMS_MAIL_MSG;

/*****************/
/* VMS user data */
/*****************/

/*
This data structure represents all the information required to maintain the VMS
Mail user's identity within soyMAIL.
*/

#define DATA_BUFFER_MAX 4096

typedef struct STRUCT_VMS_MAIL_USER
{
   int  ListMessageIdStart,
        ListMessageIdCount,
        ListMessageIdMax,
        ListMessageIdVector,
        MailDotMai,
        MessageCopiedCount,
        MessageDeletedCount,
        MoveNewMailCount;

   char  UserName [USERNAME_MAX+1];
   unsigned long  UserNameLength;

   unsigned long  CurrentMessageCount,
                  TotalReclaimBytes;

   unsigned long  MailVmsStatus,
                  VmsMailUserContext;

   unsigned short  VmsMailNewMessages;
   unsigned long  VmsMailAutoPurge,
                  VmsMailCcPrompt,
                  VmsMailCopyForward,
                  VmsMailCopyReply,
                  VmsMailCopySend;

   char  VmsMailDirectory [255+1];
   unsigned long  VmsMailDirectoryLength;
   char  VmsMailEditor [255+1];
   unsigned long  VmsMailEditorLength;
   char  VmsMailForm [255+1];
   unsigned long  VmsMailFormLength;
   char  VmsMailForwarding [255+1];
   unsigned long  VmsMailForwardingLength;
   char  VmsMailFullDirectory [255+1];
   unsigned long  VmsMailFullDirectoryLength;
   char  VmsMailPersonalName [127+1];
   unsigned long  VmsMailPersonalNameLength;
   char  VmsMailQueue [255+1];
   unsigned long  VmsMailQueueLength;
   char  VmsMailSigFile [255+1];
   unsigned long  VmsMailSigFileLength;

   unsigned long  VmsMailFileContext;
   char VmsMailFileName [255+1];
   unsigned long  VmsMailFileNameLength;
   char  VmsMailFolderName [255+1];
   unsigned long  VmsMailFolderNameLength;
   char  VmsMailWasteBasketName [63+1];
   unsigned long  VmsMailWasteBasketNameLength;

   unsigned long  VmsMailDestFileContext;
   char VmsMailDestFileName [255+1];
   unsigned long  VmsMailDestFileNameLength;
   char  VmsMailDestFolderName [255+1];
   unsigned long  VmsMailDestFolderNameLength;

   char  *MailFileListPtr;
   int  MailFileListCount,
        MailFileListLength,
        MailFileListSize;

   char  *FolderListPtr;
   int  FolderListCount,
        FolderListLength,
        FolderListSize;

   VMS_MAIL_MSG  *MessageList;
   int  MessageListCount;

   char  *MsgIdListPtr;
   int  MsgIdListCount;

   VMS_MAIL_MSG  VmsMailMsg;

   char  DataBuffer [DATA_BUFFER_MAX];
   char  *DataBufferPtr;
   unsigned long  DataBufferLength;

   char  *SignaturePtr;
   int  SignatureLength;

   /* a pointer to the message (text) search function */
   int  (*SearchFunction)();
   void  *SearchDataPtr;
   /* a boolean to enable callable mail 'select' functionality */
   int  SelectNotSearch;

} VMS_MAIL_USER;

/**************/
/* prototypes */
/**************/

unsigned int CallMailElfHash (char*, int);
int CallMailFileOpen (VMS_MAIL_USER*);
int CallMailFileClose (VMS_MAIL_USER*);
int CallMailFileList (VMS_MAIL_USER*);
char* CallMailFileIn (VMS_MAIL_USER*, char*, char*);
int CallMailFolderList (VMS_MAIL_USER*);
int CallMailFolderListCallback (VMS_MAIL_USER*, struct dsc$descriptor*);
int CallMailGetUserProfile (VMS_MAIL_USER*);
int CallMailGoToNewMail (VMS_MAIL_USER*);
VMS_MAIL_MSG* CallMailMessageInit (VMS_MAIL_USER*);
int CallMailMessageContext (VMS_MAIL_USER*, VMS_MAIL_MSG*);
int CallMailMessageContextEnd (VMS_MAIL_USER*, VMS_MAIL_MSG*);
int CallMailMessageCopy (VMS_MAIL_USER*);
int CallMailMessageCopyMove (VMS_MAIL_USER*, BOOL);
int CallMailMessageDelete (VMS_MAIL_USER*);
void CallMailMessageFree (VMS_MAIL_MSG*);
int CallMailMessageGet (VMS_MAIL_USER*, VMS_MAIL_MSG*);
int CallMailMessageHeader (VMS_MAIL_USER*, VMS_MAIL_MSG*);
int CallMailMessageList (VMS_MAIL_USER*);
unsigned int CallMailHeaderHash (VMS_MAIL_MSG*);
int CallMailMessageMove (VMS_MAIL_USER*);
int CallMailMessageSelect (VMS_MAIL_USER*);
void CallMailMessageSelectReset (VMS_MAIL_USER*);
int CallMailSetUserProfile (VMS_MAIL_USER*);
int CallMailSetUserNewMessages (VMS_MAIL_USER*, int);

#endif /* INCLUDE_CALLMAIL_H */

/******************************************************************************/

