/*****************************************************************************/
#ifdef COMMENTS_WITH_COMMENTS
/*
                                  compose.c

All things involved in the composition and sending of a mail message.

When sending a message, if there are soyMAIL attachment files that have been
checked for inclusion a "multipart/mixed" MIME message is generated.  Each
attachment with a "text/.." content-type is quoted-printable encoded in a
separate part.  Attachments with non-text content are provided as base-64
encoded parts.  If there is a message body this is appropriately encoded and
supplied as the first part with a content-type of "text/plain".

If there are no attachments files present or checked the message body is
appropriately encoded and made into an in-line MIME message with a plain-text
content-type.

If any one of the address fields (To, CC or BCC) contain a non-VMS address the
SMTP Message Transport Agent is used to send the message.  If the address
fields only contain what looks like VMS Mail addresses the message is sent via
VMS callable Mail.


HTML EDITING
------------
This rather neat capability is provided using stand-along, JavaScript-driven,
HTML text editors that can be attached to the composition text area.  soyMAIL
has some hooks that recognise and support such editors, and when used to
compose a message the content is made into quoted-printable encoded HTML for
transmission.  They are made available using the configuration directives

  [html-editor-load]     JavaScript code to load an initialise the editor

So far it has been tried successfully with

  o  FCKeditor v2.2      http://www.fckeditor.net/
                         http://sourceforge.net/projects/fckeditor/

  o  TinyMCE v2.0        http://tinymce.moxiecode.com/
     MoxieCode Systems   http://www.moxiecode.com/

Make your choice.  Install it independently of soyMAIL.  Check configuration
requirements for the editor.  Modify the soyMAIL configuration directives to
provide the path to the editor and the initialization JavaScript code, both of
which will be provided in the editor installation guidelines, and voila!  The
function htmlEditorLoad() must be present (even if empty as in the case of
TinyMCE), being an onload= target for the document.


EXAMPLE FOR TINYMCE
-------------------
A functional configuration for TnyMCE.

  [html-editor-load]
  <script type="text/javascript" 
          src="/tinymce/jscripts/tiny_mce/tiny_mce.js"></script>
  <script type="text/javascript">

  tinyMCE.init({
     mode : "exact",
     elements : "compose_text",
     theme : "advanced",
     plugins: "preview,print,insertdatetime",
     theme_advanced_buttons1_add: "fontselect,fontsizeselect",
     theme_advanced_buttons2_add: "preview,print",
     theme_advanced_buttons3_add: "insertdate,inserttime",
     theme_advanced_buttons3_add_before: "forecolor,backcolor"
  });

  function htmlEditorLoad() { /* nothing required for TinyMCE! */ }
  </script>

This of course requires a mapping rule to provide access to the TinyMCE
installation.

  pass /tinymce/* /$1$disk/tinymce/*


EXAMPLE FOR FCKEDITOR
---------------------

  [html-editor-load]
  <script type="text/javascript" src="/FCKeditor/fckeditor.js"></script>
  <script type="text/javascript">
  function htmlEditorLoad()
  {
     var oFCKeditor = new FCKeditor( 'compose_text' ) ;
     oFCKeditor.BasePath = "/fckeditor/" ;
     oFCKeditor.ReplaceTextarea() ;
  }
  </script>

And a similar mapping requirement exists for FCKeditor.

  pass /fckeditor/* /$1$disk/fckeditor/*



COPYRIGHT
---------
Copyright (C) 2005,2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
04-APR-2006  MGD  add [send-close] and ComposeSendClose()
15-MAR-2006  MGD  use form field TABINDEX attribute to control focus
13-MAR-2006  MGD  bugfix; ComposeSelfAddress() 'len' initialization
11-MAR-2006  MGD  move input focus to the field most appropriate to the stage
                    of message composition, defaulting to the message text
                  bugfix; ComposeSelfAddress() with rationalization of 'self' 
                    address generation and using this for the "From:" header
25-FEB-2006  MGD  ComposePageRequest() support VMS copy-send/forward/reply
                  ComposeSendMessage() expand MX_.. logical names that
                    can be used to generate the "From:" address
                  ComposePage() [signature] button only if the user has a
                    signature file specified
14-FEB-2006  MGD  when used from the message page the main menu [compose]
                    button should provide a blank message, when used from the
                    compose page it should propagate any in-composition message
09-FEB-2006  MGD  bugfix; default charset should be ISO-8859-1 not IS0-8859-1
                    (actually went to using the macro SOY_DEFAULT_CHARSET)
06-FEB-2006  MGD  modified [reply-to] option to [smtp-from] and in conjunction
                    with [personal-name] have ComposePageRequest() generate a
                    "From:" header more in line with TCPIP$SMTP_FROM behaviour
                  add explicit [reply-to] button and field so that messages
                    can individually have a reply-to set against them
                  ComposeMime() add client traceability information to
                    "Received:" header field
01-FEB-2005  MGD  initial
*/
#endif /* COMMENTS_WITH_COMMENTS */
/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unixio.h>

/* VMS related header files */
#include <descrip.h>
#include <rmsdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header file */
#include "soymail.h"
#include "address.h"
#include "attach.h"
#include "callmail.h"
#include "cgilib.h"
#include "compose.h"
#include "config.h"
#include "html.h"
#include "mainmenu.h"
#include "mta.h"
#include "options.h"
#include "other.h"
#include "message.h"
#include "request.h"
#include "sendmail.h"

#define FI_LI __FILE__, __LINE__

/* global storage */

/* prototypes */

/* external storage */

extern BOOL  Debug,
             WatchEnabled;

extern int  VmsVersion;

extern char  CurrentVmsTimeString[],
             LocalHostName[],
             SoftwareId[],
             SoftwareVn[];

extern  CONFIG_DATA SoyMailConfig;
extern VMS_MAIL_USER VmsMailUser;

/*****************************************************************************/
/*
Assemble the request data related to this message composition.
*/

BOOL ComposePageRequest (REQUEST_DATA *rdptr)

{
   BOOL  MsgReply, MsgReplyAll,
         MsgForward, MsgEditAsNew;
   char  *cptr, *sptr,
         *CcPtr, *BccPtr, *ReplyToPtr;
   COMPOSE_DATA  *cdptr;
   MIME_DATA  *mdptr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("ComposePageRequest()");

   /* a bit of sanity checking */
   if (!rdptr->PrivateAccess) ErrorExit (SS$_BUGCHECK, FI_LI);

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;
   uoptr = &rdptr->UserOptions;
   cdptr = &rdptr->ComposeData;

   /* this also is used as the "from:" address */
   ComposeSelfAddress (rdptr);

   if (rdptr->PrevPageIdent == PAGE_COMPOSE)
   {
      /*************************/
      /* from the compose page */
      /*************************/

      /* propagate selected items from (any) previous request */
      cdptr->AlreadySent = sdptr->ComposeData.AlreadySent;
      cdptr->ContentTypeIs = sdptr->ComposeData.ContentTypeIs;
      cdptr->CopiedSelf = sdptr->ComposeData.CopiedSelf;
      cdptr->EditCols = sdptr->ComposeData.EditCols;
      cdptr->EditRows = sdptr->ComposeData.EditRows;
      cdptr->EditorType = sdptr->ComposeData.EditorType;
      cdptr->WrapAt = sdptr->ComposeData.WrapAt;

      CGIVAR (cdptr->ToPtr, "FORM_COMPOSE_TO");
      CGIVARNULL (CcPtr, "FORM_COMPOSE_CC");
      if (!(cdptr->CcPtr = CcPtr)) cdptr->CcPtr = "";
      CGIVARNULL (BccPtr, "FORM_COMPOSE_BCC");
      if (!(cdptr->BccPtr = BccPtr)) cdptr->BccPtr = "";
      CGIVARNULL (ReplyToPtr, "FORM_COMPOSE_REPLY_TO");
      if (!(cdptr->ReplyToPtr = ReplyToPtr)) cdptr->ReplyToPtr = "";

      CGIVAR (cdptr->SubjPtr, "FORM_COMPOSE_SUBJ");
      CGIVAR (cdptr->TextPtr, "FORM_COMPOSE_TEXT");
      CGIVAR (cdptr->CharSetPtr, "FORM_COMPOSE_CHARSET");

      CGIVARNULL (cptr, "FORM_CONTACT_ADDR");
      if (cptr) while (*cptr && isspace(*cptr)) cptr++;
      if (cptr && *cptr) cdptr->ContactListPtr = cptr;

      CGIVARNULL (cptr, "FORM_COMPOSE_BTN");
      if (cptr)
      {
         if (LangSame ("compose_to", cptr))
         {
            if (cptr = cdptr->ContactListPtr)
            {
               cdptr->ToPtr = AddressListAppend (cdptr->ToPtr, cptr);
               cptr = AddressListMassage (cptr, TRUE);
               StatusMessage (FI_LI, 0, "%s %s: %s",
                              LangFor("to"), LangFor("compose_add"),
                              HTML_ESCAPE(cptr));
            }
            else
               StatusMessage (FI_LI, 1, LangFor("compose_add_none"));
         }
         else
         if (LangSame ("compose_cc", cptr))
         {
            if (cptr = cdptr->ContactListPtr)
            {
               cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, cptr);
               cptr = AddressListMassage (cptr, TRUE);
               StatusMessage (FI_LI, 0, "%s %s: %s",
                              LangFor("cc"), LangFor("compose_add"),
                              HTML_ESCAPE(cptr));
            }
            else
            if (CcPtr && !CcPtr[0])
            {
               cdptr->ShowCcAddr = FALSE;
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
            if (!CcPtr)
            {
               cdptr->ShowCcAddr = TRUE;
               rdptr->FieldFocusPtr = "soymail_form.compose_cc.focus();";
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
               StatusMessage (FI_LI, 1, LangFor("compose_add_none"));
         }
         else
         if (LangSame ("compose_bcc", cptr))
         {
            if (cptr = cdptr->ContactListPtr)
            {
               if (AddressIsRfc (cptr))
               {
                  cdptr->BccPtr = AddressListAppend (cdptr->BccPtr, cptr);
                  cptr = AddressListMassage (cptr, TRUE);
                  StatusMessage (FI_LI, 0, "%s %s: %s",
                                 LangFor("bcc"), LangFor("compose_add"),
                                 HTML_ESCAPE(cptr));
               }
               else
                  StatusMessage (FI_LI, 1, "%s: %s",
                                 LangFor("compose_bcc_vms_address"),
                                 HTML_ESCAPE(cptr));
            }
            else
            if (BccPtr && !BccPtr[0])
            {
               cdptr->ShowBccAddr = FALSE;
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
            if (!BccPtr)
            {
               cdptr->ShowBccAddr = TRUE;
               rdptr->FieldFocusPtr = "soymail_form.compose_bcc.focus();";
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
               StatusMessage (FI_LI, 1, LangFor("compose_add_none"));
         }
         else
         if (LangSame ("compose_reply_to", cptr))
         {
            if (cptr = cdptr->ContactListPtr)
            {
               if (AddressIsRfc (cptr))
               {
                  sptr = calloc (1, strlen(cptr)+1);
                  if (!sptr) ErrorExit (vaxc$errno, FI_LI);
                  strcpy (cdptr->ReplyToPtr = sptr, cptr);
                  StatusMessage (FI_LI, 0, "%s %s: %s",
                                 LangFor("compose_reply_to"),
                                 LangFor("compose_add"),
                                 HTML_ESCAPE(cptr));
               }
               else
                  StatusMessage (FI_LI, 1, "%s: %s",
                                 LangFor("compose_reply_to_vms_address"),
                                 HTML_ESCAPE(cptr));
            }
            else
            if (ReplyToPtr && !ReplyToPtr[0])
            {
               cdptr->ShowBccAddr = FALSE;
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
            if (!ReplyToPtr)
            {
               cdptr->ShowReplyToAddr = TRUE;
               rdptr->FieldFocusPtr = "soymail_form.compose_reply_to.focus();";
               StatusMessage (FI_LI, 0, "%s %s",
                              LangFor("compose"), LangFor("page_opened"));
            }
            else
               StatusMessage (FI_LI, 1, LangFor("compose_add_none"));
         }
         else
         if (LangSame ("compose_send", cptr))
            ComposeSendMessage (rdptr);
         else
         if (LangSame ("compose_send_close", cptr))
         {
            ComposeSendMessage (rdptr);
            if (cdptr->AlreadySent)
            {
               ComposeSendClose (rdptr);
               return (TRUE);
            }
         }
         else
         if (LangSame ("compose_sign", cptr))
         {
            OptionsLoadSignature (rdptr);
            StatusMessage (FI_LI, 0, LangFor("compose_sign_done"));
         }
         else
         if (LangSame ("compose_save", cptr))
            ComposeSaveDraft (rdptr);
         else
         if (LangSame ("compose_restore", cptr))
         {
            ComposeRestoreDraft (rdptr);
            StatusMessage (FI_LI, 0, "%s %s.",
                           LangFor("compose_restore_done"),
                           cdptr->DraftSavedOnPtr);
         }
         else
         if (LangSame ("compose_clear", cptr))
         {
            cdptr->TextPtr = "";
            StatusMessage (FI_LI, 0, LangFor("compose_clear_done"));
         }
         else
         if (LangSame ("compose_new", cptr))
         {
            cdptr->FromPtr = cdptr->ToPtr = cdptr->CcPtr =
               cdptr->BccPtr = cdptr->ReplyToPtr = cdptr->SubjPtr =
               cdptr->TextPtr = cdptr->AttribPtr = NULL;
            cdptr->AlreadySent = cdptr->ContentTypeIs = 0;
            rdptr->FieldFocusPtr = "soymail_form.compose_to.focus();";
            StatusMessage (FI_LI, 0, LangFor("compose_new_done"));
         }
         else
         if (LangSame ("compose_wrap_on", cptr))
         {
            /* fudge this by changing the compose state */
            cdptr->WrapAt = cdptr->EditCols;
            StatusMessage (FI_LI, 0, "%s.", LangFor("compose_wrap_on"));
         }
         else
         if (LangSame ("compose_wrap_off", cptr))
         {
            /* fudge this by changing the compose state */
            cdptr->WrapAt = 0;
            StatusMessage (FI_LI, 0, "%s.", LangFor("compose_wrap_off"));
         }
         else
         if (LangSame ("compose_edit_html", cptr))
         {
            cdptr->EditorType = COMPOSE_EDITOR_HTML;
            StatusMessage (FI_LI, 0, "%s.", LangFor("compose_edit_html"));
         }
         else
         if (LangSame ("compose_edit_plain", cptr))
         {
            cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
            StatusMessage (FI_LI, 0, "%s.", LangFor("compose_edit_plain"));
         }
         else
         if (LangSame ("compose_edit_normal", cptr))
         {
            cdptr->EditCols = 72;
            StatusMessage (FI_LI, 0, "%s.", LangFor("compose_edit_normal"));
         }
         else
         if (LangSame ("compose_edit_wide", cptr))
         {
            cdptr->EditCols = 130;
            StatusMessage (FI_LI, 0, "%s.", LangFor("compose_edit_wide"));
         }
         else
         if (LangSame ("open", cptr))
         {
            /* open the contact list */
            StatusMessage (FI_LI, 0, "%s %s",
                           LangFor("compose"), LangFor("page_opened"));
         }
         else
            ErrorExit (SS$_BUGCHECK, FI_LI);
      }
      else
      {
         if (AttachComposeRequest (rdptr)) return (TRUE);

         if (!cdptr->CopiedSelf && muptr->VmsMailCopySend)
         {
            cdptr->CopiedSelf = TRUE;
            cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, cdptr->SelfPtr);
         }

         rdptr->FieldFocusPtr = "soymail_form.compose_to.focus();";

         StatusMessage (FI_LI, 0, "%s %s",
                        LangFor("compose"), LangFor("page_opened"));
      }
   }
   else
   if (rdptr->PrevPageIdent == PAGE_MESSAGE)
   {
      /**************************/
      /* from message read page */
      /**************************/

      MsgReply = MsgReplyAll = MsgForward = MsgEditAsNew = FALSE;
      CGIVARNULL (cptr, "FORM_MSG_REPLY");
      if (!cptr) CGIVARNULL (cptr, "FORM_MSG_REPLY_POPUP");
      if (cptr) MsgReply = TRUE;
      if (!cptr)
      {
         CGIVARNULL (cptr, "FORM_MSG_REPLY_ALL");
         if (!cptr) CGIVARNULL (cptr, "FORM_MSG_REPLY_ALL_POPUP");
         if (cptr) MsgReplyAll = TRUE;
      }
      if (!cptr)
      {
         CGIVARNULL (cptr, "FORM_MSG_FORWARD");
         if (!cptr) CGIVARNULL (cptr, "FORM_MSG_FORWARD_POPUP");
         if (cptr) MsgForward = TRUE;
         rdptr->FieldFocusPtr = "soymail_form.compose_to.focus();";
      }
      if (!cptr)
      {
         CGIVARNULL (cptr, "FORM_MSG_EDIT_NEW");
         if (!cptr) CGIVARNULL (cptr, "FORM_MSG_EDIT_NEW_POPUP");
         if (cptr) MsgEditAsNew = TRUE;
         rdptr->FieldFocusPtr = "soymail_form.compose_to.focus();";
      }

      if (MsgReply || MsgReplyAll || MsgForward || MsgEditAsNew)
      {
         vmptr = &muptr->VmsMailMsg;

         vmptr->MessageId = sdptr->MessageId;
         CallMailMessageGet (muptr, NULL);
         if (VMSnok (muptr->MailVmsStatus))
         {
            StatusMessage (FI_LI, 1, "MAIL: %s.",
                           SysGetMsg(muptr->MailVmsStatus));
            ComposePage (rdptr);
            return (TRUE);
         }

         InetMailParseHeader (vmptr);
         if (vmptr->MimeDataPtr)
            if (!MimeDecParseBody (vmptr->MimeDataPtr))
               vmptr->MimeDataPtr = NULL;

         /* note that this is using the state (i.e. previous request) data */
         if (SoyMailConfig.HtmlEditorLoad &&
             sdptr->ComposeData.ContentTypeIs == MIME_CONTENT_TEXT_HTML)
            mdptr = MessageBodyIdentify (vmptr, OPTIONS_READ_PLAIN0_HTML1);
         else
            mdptr = MessageBodyIdentify (vmptr, OPTIONS_READ_PLAIN1_HTML0);
         if (mdptr)
         {
            /* note what we have actually retrieved */
            cdptr->ContentTypeIs = mdptr->ContentTypeIs;
            if (MimeDecProcessPart (mdptr))
            {
               if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_PLAIN)
               {
                  if (sdptr->MassageMessage)
                     MessageMassage (mdptr->PartContentPtr);
                  cdptr->TextPtr = mdptr->PartContentPtr;
                  cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
               }
               else
               if (mdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
               {
                  HtmlSanitise (mdptr->PartContentPtr);
                  cdptr->TextPtr = mdptr->PartContentPtr;
                  cdptr->EditorType = COMPOSE_EDITOR_HTML;
               }
               else
                  cdptr->TextPtr = "";
            }
            else
               cdptr->TextPtr = "";
            rdptr->CharSetPtr = mdptr->CharsetPtr;
         }
         else
         {
            cdptr->TextPtr = "";
            cdptr->CharSetPtr = uoptr->CharSet;
            if (!vmptr->VmsForeign)
               StatusMessage (FI_LI, 1, "%s", LangFor("mime_type_unknown"));
         }

         if (MsgReply)
         {
            cdptr->ToPtr = AddressListAppend (cdptr->ToPtr, vmptr->MsgFromPtr);
            if (!cdptr->CopiedSelf && muptr->VmsMailCopyReply)
            {
               cdptr->CopiedSelf = TRUE;
               cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, cdptr->SelfPtr);
            }
            cdptr->AttribPtr = LangFor("compose_a_reply");
         }
         else
         if (MsgReplyAll)
         {
            cdptr->ToPtr = AddressListAppend (cdptr->ToPtr, vmptr->MsgFromPtr);
            cdptr->ToPtr = AddressListAppend (cdptr->ToPtr, vmptr->MsgToPtr);
            cdptr->ToPtr = AddressListAppend (cdptr->ToPtr, vmptr->MsgCcPtr);
            if (!cdptr->CopiedSelf && muptr->VmsMailCopyReply)
            {
               cdptr->CopiedSelf = TRUE;
               cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, cdptr->SelfPtr);
            }
            cdptr->AttribPtr = LangFor("compose_a_reply");
         }
         else
         if (MsgForward)
         {
            cdptr->AttribPtr = LangFor("compose_a_forward");
            if (!cdptr->CopiedSelf && muptr->VmsMailCopyForward)
            {
               cdptr->CopiedSelf = TRUE;
               cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, cdptr->SelfPtr);
            }
         }
         if (MsgEditAsNew)
         {
            if (!cdptr->CopiedSelf && muptr->VmsMailCopySend)
            {
               cdptr->CopiedSelf = TRUE;
               cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, cdptr->SelfPtr);
            }
         }

         cdptr->SubjPtr = vmptr->MsgSubjPtr;
         cdptr->FromPtr = vmptr->MsgFromPtr;
      }
      else
      if (!cdptr->CopiedSelf && muptr->VmsMailCopySend)
      {
         cdptr->CopiedSelf = TRUE;
         cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, cdptr->SelfPtr);
      }

      StatusMessage (FI_LI, 0, "%s %s",
                     LangFor("compose"), LangFor("page_opened"));
   }
   else
   {
      /*****************/
      /* anywhere else */
      /*****************/

      if (!cdptr->CopiedSelf && muptr->VmsMailCopySend)
      {
         cdptr->CopiedSelf = TRUE;
         cdptr->CcPtr = AddressListAppend (cdptr->CcPtr, cdptr->SelfPtr);
      }

      rdptr->FieldFocusPtr = "soymail_form.compose_to.focus();";

      StatusMessage (FI_LI, 0, "%s %s",
                     LangFor("compose"), LangFor("page_opened"));
   }

   ComposePage (rdptr);

   return (TRUE);
}

/*****************************************************************************/
/*
Display the message composition (edit) page.
*/

void ComposePage (REQUEST_DATA *rdptr)

{
   int  len,
        AttachFileCount,
        ButtonWidth,
        RowSpan,
        ContactsRows;
   char  *cptr, *sptr, *zptr,
         *AddrPtr,
         *ContactList,
         *NamePtr;
   COMPOSE_DATA  *cdptr;
   USER_OPTIONS  *uoptr;
   REQUEST_DATA  *sdptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("ComposePage()"); 

   muptr = &VmsMailUser;

   uoptr = &rdptr->UserOptions;
   sdptr = (REQUEST_DATA*)rdptr->StateDataPtr;
   vmptr = &muptr->VmsMailMsg;
   cdptr = &rdptr->ComposeData;

   /* if not otherwise set then focus on the message body */
   if (!rdptr->FieldFocusPtr)
      rdptr->FieldFocusPtr = "soymail_form.compose_text.focus();";

   /* get any contacts list */
   CGIVAR (cptr, "FORM_CONTACT_LIST");
   ContactsLoad (rdptr, cptr);
   ContactList = TrnLnm (SOY_CONTACT_LIST, NULL, 0);

   cdptr->ToPtr = AddressListMassage (cdptr->ToPtr, FALSE);
   cdptr->CcPtr = AddressListMassage (cdptr->CcPtr, FALSE);
   cdptr->BccPtr = AddressListMassage (cdptr->BccPtr, FALSE);

   if (!cdptr->FromPtr) cdptr->FromPtr = "";
   if (!cdptr->ToPtr) cdptr->ToPtr = "";
   if (!cdptr->CcPtr) cdptr->CcPtr = "";
   if (!cdptr->BccPtr) cdptr->BccPtr = "";
   if (!cdptr->ReplyToPtr) cdptr->ReplyToPtr = "";
   if (!cdptr->SubjPtr) cdptr->SubjPtr = "";
   if (!cdptr->TextPtr) cdptr->TextPtr = "";
   if (!cdptr->AttribPtr) cdptr->AttribPtr = "";

   if (cdptr->CcPtr[0]) cdptr->ShowCcAddr = TRUE;
   if (cdptr->BccPtr[0]) cdptr->ShowBccAddr = TRUE;
   if (cdptr->ReplyToPtr[0]) cdptr->ShowReplyToAddr = TRUE;

   if (AddressIsRfc(cdptr->ToPtr) ||
       AddressIsRfc(cdptr->CcPtr) ||
       AddressIsRfc(cdptr->BccPtr) ||
       AddressIsRfc(cdptr->ReplyToPtr))
      cdptr->LooksLikeRfc = TRUE;
   else
      cdptr->LooksLikeRfc = FALSE;

   RowSpan = 2;
   if (cdptr->ShowCcAddr) RowSpan++;
   if (cdptr->ShowBccAddr) RowSpan++;
   if (cdptr->ShowReplyToAddr) RowSpan++;

   ContactsRows = 3;
   if (cdptr->ShowCcAddr) ContactsRows += 2;
   if (cdptr->ShowBccAddr) ContactsRows += 2;
   if (cdptr->ShowReplyToAddr) ContactsRows += 2;

   if (!cdptr->EditorType)
   {
      /* not explicitly requesting one way or the other */
      if (SoyMailConfig.HtmlEditorLoad)
      {
         if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
            cdptr->EditorType = COMPOSE_EDITOR_HTML;
         else
         if (uoptr->ComposeHTML)
            cdptr->EditorType = COMPOSE_EDITOR_HTML;
         else
            cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
      }
      else
         cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
   }
   /* obviously x 2 */
   if (!SoyMailConfig.HtmlEditorLoad) cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
   if (!rdptr->ClientHasJavaScript) cdptr->EditorType = COMPOSE_EDITOR_PLAIN;

   cptr = LangFor("compose_button_width");
   ButtonWidth = atoi(cptr);
   if (ButtonWidth <= 0 || ButtonWidth > 12) ButtonWidth = 6;
   /* allow for the width of "HTML" in the change editor button */
   if (SoyMailConfig.HtmlEditorLoad && rdptr->ClientHasJavaScript)
      ButtonWidth++;

   if (cdptr->EditRows != 20 && cdptr->EditRows != 40)
      cdptr->EditRows = 20; 
   if (cdptr->EditCols != 72 && cdptr->EditCols != 130)
      cdptr->EditCols = cdptr->WrapAt = 72; 
   if (cdptr->WrapAt && cdptr->WrapAt != cdptr->EditCols)
      cdptr->WrapAt = cdptr->EditCols;

   /**************/
   /* begin page */
   /**************/

   rdptr->CharSetPtr = cdptr->CharSetPtr;
   if (!rdptr->CharSetPtr || !rdptr->CharSetPtr[0])
      rdptr->CharSetPtr = SOY_DEFAULT_CHARSET;

   MainMenuPageBegin (rdptr, PAGE_COMPOSE);
   MainMenuBar (rdptr);
   StatusInfoPanel (rdptr);

   /******************/
   /* message header */
   /******************/

   fputs ("<!-- message header begin -->\n", stdout);

   if (rdptr->ClientHasJavaScript)
   {
      fprintf (stdout,
"<SCRIPT LANGUAGE=\"JavaScript\">\n\
function resizeContacts() {\n\
   tdHeight = document.getElementById(\'contact_td\').scrollHeight;\n\
   document.soymail_form.contact_addr.style.height = tdHeight + \'px\';\n\
}\n\
function contactNone() {\n\
   refreshStatus (\"status_info\", \
\"<SPAN CLASS=\'css_warning\'>%s</SPAN>\");\n\
   return false;\n\
}\n\
function contactNonInternet(addr) {\n\
   refreshStatus (\"status_info\", \
\"<SPAN CLASS=\'css_warning\'>%s: \" + \
addr.replace(/&/g,\"&amp;\").replace(/</g,\"&lt;\").replace(/>/g,\"&gt;\") + \
\"</SPAN>\");\n\
   return false;\n\
}\n\
function contactAdded(descr,addr) {\n\
   refreshStatus (\"status_info\", descr + \" %s: \" + \
addr.replace(/&/g,\"&amp;\").replace(/</g,\"&lt;\").replace(/>/g,\"&gt;\"));\n\
   return false;\n\
}\n\
function contactAppend (toCcBcc,descr) {\n\
   var list = document.getElementById(toCcBcc);\n\
   if (list == null) return true;\n\
   var cad = document.soymail_form.contact_addr;\n\
   if (cad.selectedIndex < 0) {\n\
      if (list.value == \"\") return true;\n\
      return contactNone();\n\
   }\n\
   var addr = cad.options[cad.selectedIndex].value;\n\
   if (toCcBcc.indexOf(\"bcc\") != -1 && addr.indexOf(\"@\") == -1) \
return contactNonInternet(addr);\n\
   cad.selectedIndex = -1;\n\
   if (list.value.indexOf(addr) >= 0) return false;\n\
   if (list.value != \"\") list.value += \"\\n\";\n\
   list.value += addr;\n\
   return contactAdded(descr,addr);\n\
}\n\
</SCRIPT>\n",
              LangFor("compose_add_none"),
              LangFor("compose_bcc_vms_address"),
              LangFor("compose_add"));
   }

   fprintf (stdout,
"<P CLASS=\"css_p08\">\
<TABLE WIDTH=\"100%%\" CELLPADDING=\"1\" CELLSPACING=\"5\" \
CLASS=\"css_panel_1\" ALIGN=\"center\">\n\
\
<TR>\n\
<TD CLASS=\"css_tabhdr\">%s:</TD>\n\
<TD STYLE=\"vertical-align:top;\">\
<TEXTAREA CLASS=\"css_text_edit\" STYLE=\"vertical-align:top;width:40em;\" \
ID=\"compose_to\" NAME=\"compose_to\" \
ROWS=\"2\" COLS=\"40\" WRAP=\"off\" TABINDEX=\"1\">%s</TEXTAREA></TD>\n\
<TD ROWSPAN=\"%d\" STYLE=\"vertical-align:top;padding-left:15;\">\n\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"return contactAppend(\'compose_to\',\'%s\')\">\n\
<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"return contactAppend(\'compose_cc\',\'%s\')\">\n\
<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"return contactAppend(\'compose_bcc\',\'%s\')\">\n\
<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n\
</TD>\n\
<TD ROWSPAN=\"%d\" ID=\"contact_td\" STYLE=\"vertical-align:top;\">\n",
            LangFor("to"), HTML_ADDR(cdptr->ToPtr),
            RowSpan,
            ButtonWidth,
            LangFor("compose_to"), LangFor("to"),
            ButtonWidth,
            LangFor("compose_cc"), LangFor("cc"),
            ButtonWidth,
            LangFor("compose_bcc"), LangFor("bcc"),
            ButtonWidth,
            LangFor("compose_reply_to"),
            RowSpan);

   ContactsCompose (rdptr, ContactsRows);
   fprintf (stdout, "</TD>\n");

   if (cdptr->ShowCcAddr)      
      fprintf (stdout,
"</TR>\n\
<TR>\n\
<TD  CLASS=\"css_tabhdr\">%s:</TD>\n\
<TD STYLE=\"vertical-align:top;\">\
<TEXTAREA CLASS=\"css_text_edit\" STYLE=\"width:40em;\" \
ID=\"compose_cc\" NAME=\"compose_cc\" ROWS=\"2\" COLS=\"40\" \
WRAP=\"off\" TABINDEX=\"2\">%s</TEXTAREA></TD></TR>\n",
               LangFor("cc"), HTML_ADDR(cdptr->CcPtr));

   if (cdptr->ShowBccAddr)
      fprintf (stdout,
"<TR>\n\
<TD CLASS=\"css_tabhdr\">%s:</TD>\
<TD STYLE=\"vertical-align:top;\">\
<TEXTAREA CLASS=\"css_text_edit\" STYLE=\"width:40em;\" \
ID=\"compose_bcc\" NAME=\"compose_bcc\" ROWS=\"2\" COLS=\"40\" \
WRAP=\"off\" TABINDEX=\"3\">%s</TEXTAREA></TD></TR>\n",
               LangFor("bcc"), HTML_ADDR(cdptr->BccPtr));

   if (cdptr->ShowReplyToAddr)
      fprintf (stdout,
"<TR>\n\
<TD  CLASS=\"css_tabhdr\">%s:</TD>\n\
<TD STYLE=\"vertical-align:top;\">\
<INPUT CLASS=\"css_text_edit\" STYLE=\"width:40em;\" \
NAME=\"compose_reply_to\" SIZE=\"40\" \
VALUE=\"%s\" TABINDEX=\"4\"></TD>\n\
</TR>\n",
               LangFor("compose_reply_to"), HTML_ADDR(cdptr->ReplyToPtr));

   fprintf (stdout,
"<TR>\n\
<TD CLASS=\"css_tabhdr\">%s:</TD>\n\
<TD STYLE=\"vertical-align:top;\">\
<INPUT CLASS=\"css_text_edit\" STYLE=\"width:40em;\" \
NAME=\"compose_subj\" SIZE=\"40\" \
VALUE=\"%s%s%s\" TABINDEX=\"5\"></TD>\n\
<TD WIDTH=\"100%%\"></TD>\n\
</TR>\n\
</TABLE>\n\
<!-- message header end -->\n\n",
            LangFor("subject"),
            cdptr->AttribPtr, *cdptr->AttribPtr ? " " : "",
            HTML_TEXT(cdptr->SubjPtr));

   /****************/
   /* message body */
   /****************/

   if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
   {
      fprintf (stdout,
"<!-- HTML editor load -->\n\
%s\n\
<!-- HTML editor loaded! -->\n\
\n",
               SoyMailConfig.HtmlEditorLoad);
      /* good golly miss molly!  is that all it takes?!! */
   }

   fputs ("<!-- message body begin -->\n", stdout);

   if (rdptr->ClientHasJavaScript)
   {
      fprintf (stdout,
"<SCRIPT LANGUAGE=\"JavaScript\">\n\
function confirmSend () {\n\
   var send = true;\n\
   if (%d) \
send = confirm(\"%s %s\");\n\
   if (send && document.soymail_form.compose_to.value == \"\") \
send = confirm(\"%s %s\");\n\
   if (send && document.soymail_form.compose_subj.value == \"\") \
send = confirm(\"%s %s\");\n\
   if (send && document.soymail_form.compose_text.value == \"\") \
send = confirm(\"%s %s\");\n\
   return send;\n\
}\n\
</SCRIPT>\n",
               cdptr->AlreadySent,
               LangFor("compose_already_sent"), LangFor("sure?"),
               LangFor("compose_no_recipient"), LangFor("sure?"),
               LangFor("compose_no_subject"), LangFor("sure?"),
               LangFor("compose_no_text"), LangFor("sure?"));
   }

   fprintf (stdout,
"<P CLASS=\"css_p08\">\
<TABLE CELLPADDING=\"1\" CELLSPACING=\"5\" WIDTH=\"100%%\" \
CLASS=\"css_panel_1\" ALIGN=\"center\">\n\
\
<TR><TD WIDTH=\"5%%\" STYLE=\"vertical-align:top;\">\n\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"if (!confirmSend()) return false; else return startWorking();\" \
TABINDEX=\"7\">\n\
\
<SCRIPT LANGUAGE=\"JavaScript\">\
if (window.opener) document.write(\'<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%dem;\" TYPE=\"submit\" NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"if (!confirmSend()) return false; else return startWorking();\">\');\
</SCRIPT>\n",
               ButtonWidth,
               LangFor("compose_send"),
               ButtonWidth,
               LangFor("compose_send_close"));

   if (muptr->VmsMailSigFile[0])
      fprintf (stdout,
"<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n",
                  ButtonWidth,
                  LangFor("compose_sign"));

   if (cdptr->TextPtr[0])
      fprintf (stdout,
"<P CLASS=\"css_p08\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" \
TYPE=\"reset\" VALUE=\"%s\" \
ONCLICK=\"return confirm(\'%s %s\')\">\n",
                  ButtonWidth,
                  LangFor("compose_reset"),
                  LangFor("compose_reset_confirm"), LangFor("sure?")); 

   fprintf (stdout,
"<P CLASS=\"css_p08\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n\
<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"return confirm(\'%s %s\')\">\n\
<P CLASS=\"css_p08\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"return confirm(\'%s %s\')\">\n\
<P CLASS=\"css_p02\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\" \
ONCLICK=\"return confirm(\'%s %s\')\">\n",
               ButtonWidth,
               LangFor("compose_save"),
               ButtonWidth,
               LangFor("compose_restore"),
               LangFor("compose_restore_confirm"), LangFor("sure?"), 
               ButtonWidth,
               LangFor("compose_clear"),
               LangFor("compose_clear_confirm"), LangFor("sure?"), 
               ButtonWidth,
               LangFor("compose_new"),
               LangFor("compose_new_confirm"), LangFor("sure?")); 

   if (SoyMailConfig.HtmlEditorLoad &&
       rdptr->ClientHasJavaScript)
   {
         fprintf (stdout,
"<P CLASS=\"css_p08\">\
<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n\
<P CLASS=\"css_p02\">",
               ButtonWidth,
               cdptr->EditorType == COMPOSE_EDITOR_PLAIN ?
                  LangFor("compose_edit_html") :
                  LangFor("compose_edit_plain"));
   }
   else
      fprintf (stdout, "<P CLASS=\"css_p08\">");

   if (cdptr->EditorType == COMPOSE_EDITOR_PLAIN)
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n\
<P CLASS=\"css_p02\">",
               ButtonWidth,
               cdptr->WrapAt ? LangFor("compose_wrap_off") :
                               LangFor("compose_wrap_on"));

   fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" STYLE=\"width:%dem;\" TYPE=\"submit\" \
NAME=\"compose_btn\" VALUE=\"%s\">\n\
</TD>\
<TD ROWSPAN=\"2\" WIDTH=\"5%%\" \
STYLE=\"height:%dem;padding-left:1em;vertical-align:top;\">\n\
\
<TEXTAREA CLASS=\"css_text_edit\" \
ID=\"compose_text_id\" NAME=\"compose_text\" \
ROWS=\"%d\" COLS=\"%d\" WRAP=\"%s\" TABINDEX=\"6\">",
            ButtonWidth,
            cdptr->EditCols == 72 ? LangFor("compose_edit_wide") :
                                    LangFor("compose_edit_normal"),
            /* plus five to maintain the height when using FCKeditor */
            cdptr->EditRows + 5,
            cdptr->EditRows, cdptr->EditCols,
            cdptr->WrapAt > 0 ? "virtual" : "off");

   if (cdptr->AttribPtr &&
       LangSame ("compose_a_reply", cdptr->AttribPtr))
   {
      cptr = LangFor ("compose_quote");
      fprintf (stdout, cptr,
               HTML_ESCAPE(vmptr->MsgDatePtr),
               HTML_ESCAPE(cdptr->FromPtr));
      cptr = HTML_ESCAPE(cdptr->TextPtr);
      if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      {
         fputs ("<BR>\n", stdout);
         fputs (cptr, stdout);
      }
      else
      {
         fputs ("\n", stdout);
         while (*cptr)
         {
            sptr = cptr;
            while (*cptr && *cptr != '\n') cptr++;
            if (*cptr) cptr++;
            fprintf (stdout, "%s%*.*s",
                     uoptr->QuoteChar, cptr-sptr, cptr-sptr, sptr);
         }
      }
   }
   else
   if (cdptr->AttribPtr &&
       LangSame ("compose_a_forward", cdptr->AttribPtr))
   {
      if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      {
         fprintf (stdout,
"<BR><BR>--------------- %s ---------------\n\
<BR>%s: %s\n\
<BR>%s: %s\n\
<BR>%s: %s\n\
<BR>%s: %s\n\
<BR><BR>\n",
                  LangFor("compose_original"),
                  LangFor("subject"), HTML_TEXT(vmptr->MsgSubjPtr),
                  LangFor("to"), HTML_ADDR(vmptr->MsgToPtr),
                  LangFor("from"), HTML_ADDR(vmptr->MsgFromPtr),
                  LangFor("date"), HTML_ADDR(vmptr->MsgDatePtr));
      }
      else
      {
         fprintf (stdout,
"\n\n--------------- %s ---------------\n\
%s: %s\n\
%s: %s\n\
%s: %s\n\
%s: %s\n\
\n",
                  LangFor("compose_original"),
                  LangFor("subject"), HTML_TEXT(vmptr->MsgSubjPtr),
                  LangFor("to"), HTML_ADDR(vmptr->MsgToPtr),
                  LangFor("from"), HTML_ADDR(vmptr->MsgFromPtr),
                  LangFor("date"), HTML_TEXT(vmptr->MsgDatePtr));
      }

      if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
         cptr = cdptr->TextPtr;
      else
         cptr = HTML_ESCAPE(cdptr->TextPtr);
      fputs (cptr, stdout);
   }
   else
   if (cdptr->ContentTypeIs == MIME_CONTENT_TEXT_HTML)
      fputs (cdptr->TextPtr, stdout);
   else
      fputs (HTML_ESCAPE(cdptr->TextPtr), stdout);

   if (muptr->SignaturePtr)
   {
      if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
      {
         cptr = muptr->SignaturePtr;
         while (*cptr)
         {
            sptr = cptr;
            while (*cptr && *cptr != '\n') cptr++;
            if (*cptr) cptr++;
            fprintf (stdout, "<BR>%*.*s", cptr-sptr, cptr-sptr, sptr);
         }
      }
      else
         fputs (HTML_TEXT(muptr->SignaturePtr), stdout);
   }

   fprintf (stdout,
"</TEXTAREA>\n\
</TD><TD WIDTH=\"90%%\"></TD>\n\
</TABLE>\n\
<!-- message body end -->\n\n");

   /***************/
   /* attachments */
   /***************/

   AttachComposeList (rdptr);

   /************/
   /* end page */
   /************/

   MainMenuPageEnd (rdptr);
}

/*****************************************************************************/
/*
Save a copy of the current composition (draft) to the file-system.

Only a single draft can be saved at any one time.

So that address lists can be saved as a single line in the draft file the
newlines delimitting each address are converted to bells (0x07) before storage
and then converted back when restoring the draft.
*/

void ComposeSaveDraft (REQUEST_DATA *rdptr)

{
   int  retval, status;
   char  *cptr, *fnptr, *sptr;
   FILE  *fp;
   COMPOSE_DATA  *cdptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeSaveDraft()\n");

   cdptr = &rdptr->ComposeData;
   muptr = &VmsMailUser;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   fnptr = CallMailFileIn (muptr, NULL, SOY_DRAFT);

   fp = fopen (fnptr, "w");
   if (!fp)
   {
      status = vaxc$errno;
      StatusMessage (FI_LI, 1, "%s: %s.", SOY_MAIL, SysGetMsg(status));
      return;
   }

   status = SS$_NORMAL;

   fprintf (fp,
"# soyMAIL draft as at %s\n\
\n\
[version]  %s\n\
[saved]  %s\n\
[editor]  %s\n\
[columns]  %d\n\
[wrap-at]  %d\n",
            CurrentVmsTimeString, SoftwareVn,
            LocalRfcBinDate(NULL),
            cdptr->EditorType == COMPOSE_EDITOR_PLAIN ? "plain" : "HTML",
            cdptr->EditCols, cdptr->WrapAt);

   cptr = AddressListMassage (cdptr->ToPtr, FALSE);
   retval = fprintf (fp, "[to]  ");
   if (retval < 0) status = vaxc$errno;
   ConfigWriteEscaped (fp, cptr);

   cptr = AddressListMassage (cdptr->CcPtr, FALSE);
   fprintf (fp, "[cc]  ");
   ConfigWriteEscaped (fp, cptr);

   cptr = AddressListMassage (cdptr->BccPtr, FALSE);
   fprintf (fp, "[bcc]  ");
   ConfigWriteEscaped (fp, cptr);

   fprintf (fp, "[reply-to]  ");
   ConfigWriteEscaped (fp, cdptr->ReplyToPtr);

   fprintf (fp, "[subj]  ");
   ConfigWriteEscaped (fp, cdptr->SubjPtr);

   cptr = calloc (1, strlen(cdptr->TextPtr)+1);
   if (!cptr) ErrorExit (vaxc$errno, FI_LI);
   strcpy (cptr, cdptr->TextPtr);
   retval = fprintf (fp, "[text]\n%s", cptr);
   if (retval < 0) status = vaxc$errno;

   fclose (fp);

   if (VMSok (status))
   {
      /* purge the versions back to one */
      strcat (fnptr, ";-1");
      while (!remove (fnptr));
      StatusMessage (FI_LI, 0, LangFor("compose_save_draft"));
   }
   else
   {
      /* remove this latest one */
      strcat (fnptr, ";0");
      remove (fnptr);
      StatusMessage (FI_LI, 1, "%s: %s.", SOY_MAIL, SysGetMsg(status));
   }
}                              

/*****************************************************************************/
/*
Read the saved draft from the file-system.
See comments in ComposeSaveDraft().
*/

void ComposeRestoreDraft (REQUEST_DATA *rdptr)

{
   int  status,
        DraftTextLength;
   char  *cptr, *fnptr, *sptr,
         *DraftTextPtr,
         *EditorPtr,
         *NamePtr,
         *ValuePtr,
         *VersionPtr;
   COMPOSE_DATA  *cdptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeRestoreDraft()\n");

   cdptr = &rdptr->ComposeData;
   muptr = &VmsMailUser;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   fnptr = CallMailFileIn (muptr, NULL, SOY_DRAFT);

   status = ReadFileIntoMemory (fnptr, &DraftTextPtr, &DraftTextLength);
   if (VMSnok (status))
   {
      StatusMessage (FI_LI, 1, "%s: %s.", SOY_MAIL, SysGetMsg(status));
      return;
   }

   EditorPtr = VersionPtr = NULL;
   cdptr->DraftSavedOnPtr = "?";

   while (DraftTextPtr && *DraftTextPtr)
   {
      cptr = ConfigTextParse (&DraftTextPtr, &NamePtr, &ValuePtr);
      if (!cptr) break;

      if (strsame (NamePtr, "version", -1))
         VersionPtr = ValuePtr;
      else
      if (strsame (NamePtr, "editor", -1))
         EditorPtr = ValuePtr;
      else
      if (strsame (NamePtr, "saved", -1))
         cdptr->DraftSavedOnPtr = ValuePtr;
      else
      if (strsame (NamePtr, "columns", -1))
         cdptr->EditCols = atoi(ValuePtr);
      else
      if (strsame (NamePtr, "wrap-at", -1))
         cdptr->WrapAt = atoi(ValuePtr);
      else
      if (strsame (NamePtr, "to", -1))
         cdptr->ToPtr = ValuePtr;
      else
      if (strsame (NamePtr, "cc", -1))
         cdptr->CcPtr = ValuePtr;
      else
      if (strsame (NamePtr, "bcc", -1))
         cdptr->BccPtr = ValuePtr;
      else
      if (strsame (NamePtr, "reply-to", -1))
         cdptr->ReplyToPtr = ValuePtr;
      else
      if (strsame (NamePtr, "subj", -1))
      {
         cdptr->SubjPtr = ValuePtr;

         /********************/
         /* loop breaks here */
         /********************/

         break;
      }
      else
         DraftTextPtr = NULL;
   }

   if (DraftTextPtr)
   {
      /*************************/
      /* hand-parse draft text */
      /*************************/

      for (cptr = DraftTextPtr; *cptr && *cptr == '\n'; *cptr++);
      if (strsame (cptr, "[text]\n", 6))
         cdptr->TextPtr = cptr + 6;
      else
         DraftTextPtr = NULL;
   }

   if (EditorPtr && *EditorPtr == 'P')
      cdptr->EditorType = COMPOSE_EDITOR_PLAIN;
   else
   if (EditorPtr && *EditorPtr == 'H')
      cdptr->EditorType = COMPOSE_EDITOR_HTML;
   else
      cdptr->EditorType = 0;

   if (!DraftTextPtr)
   {
      cdptr->ToPtr = cdptr->CcPtr = cdptr->BccPtr =
         cdptr->ReplyToPtr = cdptr->SubjPtr = cdptr->TextPtr = "";
      cdptr->EditorType = 0;
      StatusMessage (FI_LI, 1, "%s", LangFor("file_format_error"));
   }
}                              

/*****************************************************************************/
/*
Ensure the text of a message is reasonably formatted.
This includes wrapping lines at 72/130/~ characters if not quoted.
The 'escape sequence' of a backslash ('\') plus the (user-option) quote
character suppresses wrapping of that line (the 'escape sequence' is removed).
*/

char* ComposeBodyTextMassage
(
char *TextPtr,
int MaxLineLen,
char QuoteChar
)
{
   int  cnt, lcnt, llen;
   char  *cptr, *sptr, *wsptr,
         *NewTextPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "ComposeBodyTextMassage() %d %c\n",
               QuoteChar, QuoteChar);

   if (MaxLineLen <= 0) MaxLineLen = 72;

   /* first, a quick scan to see if any changes are necessary */
   lcnt = 0;
   cptr = TextPtr;
   while (*cptr)
   {
      if (QuoteChar == *cptr)
         while (*cptr && *cptr != '\n' && *cptr != '\r') cptr++;
      else
      {
         llen = 0;
         while (*cptr && *cptr != '\n' && *cptr != '\r')
         {
            llen++;
            cptr++;
            if (llen <= MaxLineLen) continue;
            lcnt++;
            llen = 0;
         }
      }
      while (*cptr == '\n' || *cptr == '\r') cptr++;
      if (*cptr == '\\' && *(cptr+1) == QuoteChar) lcnt++;
   }
   /* if none required then just return the original message */
   if (!lcnt) return (TextPtr);

   cnt = cptr - TextPtr + lcnt + 1;
   sptr = NewTextPtr = calloc (cnt, 10);
   if (!sptr) ErrorExit (vaxc$errno, FI_LI); 

   cptr = TextPtr;
   while (*cptr)
   {
      if (*cptr == '\\' && *(cptr+1) == QuoteChar)
      {
         /* remove the 'escape sequence' and do not wrap the line */
         cptr += 2;
         while (*cptr && *cptr != '\n' && *cptr != '\r') *sptr++ = *cptr++;
      }
      else
      if (*cptr == QuoteChar)
      {
         /* 'quoted' - do not wrap the line */
         while (*cptr && *cptr != '\n' && *cptr != '\r') *sptr++ = *cptr++;
      }
      else
      {
         llen = 0;
         while (*cptr && *cptr != '\n' && *cptr != '\r')
         {
            wsptr = NULL;
            while (*cptr && *cptr != '\n' && *cptr != '\r')
            {
               if (*cptr == ' ' || *cptr == '\t') wsptr = sptr;
               *sptr++ = *cptr++;
               llen++;
               if (llen >= MaxLineLen && wsptr) break;
            }
            if (llen >= MaxLineLen && wsptr) *(sptr = wsptr) = '\n';
            llen = 0;
            while (*sptr)
            {
               llen++;
               sptr++;
            }
         }
      }
      while (*cptr == '\n' || *cptr == '\r') *sptr++ = *cptr++;
   }
   *sptr = '\0';

   return (NewTextPtr);
}                              

/*****************************************************************************/
/*
Create a 'self' address for use as the [SELF] contact and as the SMTP/RFC
"From:" header field.
*/

void ComposeSelfAddress (REQUEST_DATA *rdptr)

{
   int  len;
   char  *cptr, *fhptr, *pnptr, *sptr, *zptr;
   char  UserNameLo [64];
   COMPOSE_DATA  *cdptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeSelfAddress()\n");

   muptr = &VmsMailUser;
   cdptr = &rdptr->ComposeData;
   uoptr = &rdptr->UserOptions;

   len = 0;
   pnptr = NULL;
   if (uoptr->PersonalName[0])
      len += strlen(pnptr = uoptr->PersonalName) + 1;
   else
   if (muptr->VmsMailPersonalName[0])
      len += strlen(pnptr = muptr->VmsMailPersonalName) + 1;

   fhptr = NULL;
   /* user option can just be a host/domain name for user@host/domain */
   if (uoptr->SmtpFrom[0]) fhptr = uoptr->SmtpFrom;
   if (!fhptr) fhptr = SoyMailConfig.SmtpFromHost;
   /* if not then try for MX logical names */
   if (!fhptr) fhptr = TrnLnm ("MX_FROM_HOST", NULL, 0);
   if (!fhptr) fhptr = TrnLnm ("MX_ENVELOPE_FROM_HOST", NULL, 0);
   if (!fhptr) fhptr = TrnLnm ("MX_NODE_NAME", NULL, 0);
   /* fall back to the IP host name */
   if (!fhptr) fhptr = LocalHostName;

   if (AddressIsRfc(fhptr))
   {
      /* full 'whomever@wherever' specified (usually in the user option) */
      len += strlen(fhptr) + 16;
      sptr = calloc (1, len);
      if (!sptr) ErrorExit (vaxc$errno, FI_LI);
      if (pnptr)
         sprintf (sptr, "%s <%s>", pnptr, fhptr);
      else
         strcpy (sptr, fhptr);
   }
   else
   {
      /* just the host/domain name */
      zptr = (sptr = UserNameLo) + sizeof(UserNameLo)-1;
      for (cptr = rdptr->UserName;
           *cptr && sptr < zptr;
           *sptr++ = tolower(*cptr++));
      *sptr = '\0';
      len = sptr - UserNameLo;
      len += strlen(fhptr) + 16;
      sptr = calloc (1, len);
      if (!sptr) ErrorExit (vaxc$errno, FI_LI);
      if (pnptr)
         sprintf (sptr, "%s <%s@%s>",
                  pnptr, UserNameLo, fhptr);
      else
         sprintf (sptr, "%s@%s", UserNameLo, fhptr);
   }

   cdptr->SelfPtr = sptr;

   if (WatchEnabled) WatchThis ("SELF !AZ", cdptr->SelfPtr); 
}                              

/*****************************************************************************/
/*
This rather over-size function creates a MIME compliant RFC822 message.

If there are soyMAIL attachment files that have been checked for inclusion a
"multipart/mixed" MIME message is generated.  Each attachment with a "text/.."
content-type is quoted-printable encoded in a separate part.  Attachments with
non-text content are provided as base-64 encoded parts.  If there is a message
body this is appropriately encoded and supplied as the first part with a
content-type of "text/plain".

If there are no attachments files present or checked the message body is
appropriately encoded and made into an in-line MIME message.
*/

BOOL ComposeMime (REQUEST_DATA *rdptr)

{
#define STRINIT(ptr,size)  zptr = (sptr = (char*)ptr) + size;
#define STRCAT(str)        for (cptr=str; *cptr && sptr<zptr; *sptr++=*cptr++);
#define STRNULL            if (sptr<zptr) *sptr = '\0';
#define STROVF             sptr >= zptr

static char  ThisMessageIsInMimeFormat [] =
"This message is in MIME format.  If your mail reader does not understand\r\n\
this format, some or all of the message may not be legible.\r\n";

   int sys$gettim (__unknown_params);

   int  cnt, len, status,
        AttachFileCount,
        DataLength,
        FileSizeBytes;
   char  *cptr, *fnptr, *sptr, *zptr,
         *CgiHttpXForwardedForPtr,
         *CgiRemoteAddrPtr,
         *CgiRemoteHostPtr,
         *CgiRemotePortPtr,
         *CgiServerNamePtr,
         *CharSetPtr,
         *ContentTypePtr,
         *DataPtr,
         *HtmlTextPtr,
         *MessageTextPtr,
         *MimeEncodedPtr,
         *NamePtr,
         *PlainTextPtr,
         *QuotedPrintableTextPtr,
         *TransferEncodingPtr;
   char  AttId [64],
         CbxName [256],
         HeaderBuffer [8192],
         MimeBoundary [64],
         ScratchBuffer [512];
   unsigned long  BinTime [2];
   stat_t  FstatBuffer;
   FILE  *fp;
   COMPOSE_DATA  *cdptr;
   USER_OPTIONS  *uoptr;
   TEXT_DATA  *tl1ptr, *tl2ptr, *tl3ptr;
   VMS_MAIL_MSG  *vmptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeMime()\n");

   muptr = &VmsMailUser;
   vmptr = &muptr->VmsMailMsg;
   cdptr = &rdptr->ComposeData;
   uoptr = &rdptr->UserOptions;

   if (!muptr->VmsMailFullDirectoryLength) ErrorExit (SS$_BUGCHECK, FI_LI);

   sys$gettim (&BinTime);
   MimeBoundary[0] = '\0';
   CharSetPtr = cdptr->CharSetPtr;
   if (!CharSetPtr || !CharSetPtr[0]) CharSetPtr = SOY_DEFAULT_CHARSET;
   if (Debug) fprintf (stdout, "charset: %s\n", CharSetPtr);

   /****************/
   /* message text */
   /****************/

   if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
      MessageTextPtr = cdptr->TextPtr;
   else
   if (cdptr->WrapAt)
      MessageTextPtr = ComposeBodyTextMassage (cdptr->TextPtr,
                                               cdptr->WrapAt,
                                               *uoptr->QuoteChar);
   else
      MessageTextPtr = cdptr->TextPtr;

   HtmlTextPtr = MimeEncodedPtr =
      PlainTextPtr = QuotedPrintableTextPtr = NULL;
   if (cdptr->EditorType == COMPOSE_EDITOR_HTML)
   {
      /* the HTML text body requires 'quoted-printable'ing */
      MimeEncodedPtr = MimeEncQuotedPrintable (MessageTextPtr, -1);
      QuotedPrintableTextPtr = MimeEncodedPtr;
      HtmlTextPtr = MessageTextPtr;
   }
   else
   {
      /* look for 8 bit character(s) */
      for (cptr = MessageTextPtr; *cptr && isascii(*cptr); cptr++);
      if (!*cptr)
      {
         /* look for excessively long lines */
         for (cptr = sptr = MessageTextPtr; *cptr && (cptr-sptr<=72); cptr++)
            if (*cptr == '\r' || *cptr == '\n') sptr = cptr + 1;
      }
      if (*cptr)
      {
         /* the plain text body requires 'quoted-printable'ing */
         MimeEncodedPtr = MimeEncQuotedPrintable (MessageTextPtr, -1);
         QuotedPrintableTextPtr = MimeEncodedPtr;
      }
      else
         PlainTextPtr = MimeEncodedPtr = MessageTextPtr;
   }

   tl1ptr = tl2ptr = NULL;

   /*******************************/
   /* search for attachment files */
   /*******************************/

   AttachFileCount = 0;
   fnptr = FileSpecSearch (FALSE, muptr->VmsMailFullDirectory, "SOYMAIL_*.ATT");
   while (fnptr)
   {
      /* get the digits from the file name */
      for (cptr = fnptr; *cptr; cptr++);
      while (cptr > fnptr && *cptr != ']') cptr--;
      if (*cptr == ']') cptr++;
      while (*cptr && !isdigit(*cptr)) cptr++;
      zptr = (sptr = AttId) + sizeof(AttId);
      while (isdigit(*cptr) && sptr < zptr) *sptr++ = *cptr++;
      if (sptr > zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      *sptr = '\0';

      if (WatchEnabled) WatchThis ("ATTACH !AZ", fnptr); 

      /* check if the attachment checkbox is currently checked */
      sprintf (CbxName, "FORM_PART_CBX_%s", AttId);
      CGIVARNULL (sptr, CbxName);
      /* if not checked just continue */
      if (!sptr)
      {
         fnptr = FileSpecSearch (FALSE, NULL, NULL);
         continue;
      }

      /********************/
      /* checkbox checked */
      /********************/

      if (!AttachFileCount)
      {
         sprintf (MimeBoundary, "----_=_%08.08X%08.08X",
                  BinTime[1], BinTime[0]);

         if (HtmlTextPtr && HtmlTextPtr[0])
         {
            ContentTypePtr = "text/html";
            MimeEncodedPtr = QuotedPrintableTextPtr;
            TransferEncodingPtr = "quoted-printable";
         }
         else
         if (PlainTextPtr && PlainTextPtr[0])
         {
            ContentTypePtr = "text/plain";
            MimeEncodedPtr = PlainTextPtr;
            TransferEncodingPtr = "7bit";
         }
         else
         if (QuotedPrintableTextPtr && QuotedPrintableTextPtr[0])
         {
            ContentTypePtr = "text/plain";
            MimeEncodedPtr = QuotedPrintableTextPtr;
            TransferEncodingPtr = "quoted-printable";
         }
         else
            ContentTypePtr = MimeEncodedPtr = TransferEncodingPtr = NULL;

         /************************************/
         /* "this message is in MIME format" */
         /************************************/

         /* use the 'tl2ptr' expressly so that it can be used to link later */
         TEXT_DATA_CALLOC (tl2ptr, 0);
         if (!tl2ptr) ErrorExit (vaxc$errno, FI_LI);

         TEXT_DATA_TEXT(tl2ptr) = ThisMessageIsInMimeFormat;

         /* this IS the first part of the MIME body */
         TEXT_DATA_NEXT (&cdptr->MimeBody) = tl2ptr;

         if (MimeEncodedPtr)
         {
            /****************/
            /* message text */
            /****************/

            if (WatchEnabled)
            {
               WatchThis ("ATTACH (message text)", fnptr); 
               WatchThis ("MIME TYPE !AZ", ContentTypePtr); 
               WatchThis ("MIME TRANSFER !AZ", TransferEncodingPtr); 
            }

            /* first will point to the MIME part header */
            TEXT_DATA_CALLOC (tl1ptr, 512);
            if (!tl1ptr) ErrorExit (vaxc$errno, FI_LI);

            /* link the "this message..." to this new part */
            if (tl2ptr) tl2ptr->NextPtr = tl1ptr;

            /* second will point to the MIME part encoded content */
            TEXT_DATA_CALLOC (tl2ptr, 0);
            if (!tl2ptr) ErrorExit (vaxc$errno, FI_LI);

            TEXT_DATA_TEXT(tl1ptr) = TEXT_DATA_SELF(tl1ptr);
            STRINIT (TEXT_DATA_TEXT(tl1ptr), TEXT_DATA_SIZE(tl1ptr)) 

            STRCAT ("\r\n--")
            STRCAT (MimeBoundary)
            STRCAT ("\r\nContent-Type: ")
            STRCAT (ContentTypePtr)
            STRCAT ("; charset=\"")
            STRCAT (CharSetPtr)
            STRCAT ("\"\r\nContent-Transfer-Encoding: ")
            STRCAT (TransferEncodingPtr)
            STRCAT ("\r\n\r\n")
            STRNULL
            TEXT_DATA_LENGTH (tl1ptr) = sptr - TEXT_DATA_TEXT (tl1ptr);
            if (STROVF) ErrorExit (SS$_BUGCHECK, FI_LI);

            TEXT_DATA_TEXT (tl2ptr) = MimeEncodedPtr;
            TEXT_DATA_LENGTH (tl2ptr) = strlen(TEXT_DATA_TEXT (tl2ptr));

            TEXT_DATA_NEXT (tl1ptr) = tl2ptr;

            if (Debug)
               fprintf (stdout, "%d\n|%s|\n|%s|\n",
                        AttachFileCount, TEXT_DATA_TEXT(tl1ptr),
                        TEXT_DATA_TEXT(tl2ptr));
         }
      }

      /*******************/
      /* attachment file */
      /*******************/

      AttachFileCount++;

      fp = fopen (fnptr, "r", "ctx=bin");
      if (!fp)
      {
         status = vaxc$errno;
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(status));
         return (FALSE);
      }

      if (fstat (fileno(fp), &FstatBuffer))
      {
         status = vaxc$errno;
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(status));
         return (FALSE);
      }
      FileSizeBytes = FstatBuffer.st_size;

      DataPtr = calloc (1, FileSizeBytes);
      if (!DataPtr)
      {
         status = vaxc$errno;
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(status));
         return (FALSE);
      }

      DataLength = read (fileno(fp), DataPtr, FileSizeBytes);
      if (DataLength <= 0)
      {
         status = vaxc$errno;
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(status));
         return (FALSE);
      }
      fclose (fp);

      NamePtr = ContentTypePtr = NULL;

      if (!strncmp (DataPtr, "[soyMAIL-version]", 17))
      {
         sptr = DataPtr + 17;
         while (*sptr) sptr++;
         sptr++;
         if (!strncmp (sptr, "[name]", 6))
         {
            NamePtr = sptr += 6;
            while (*sptr) sptr++;
            sptr++;
            if (!strncmp (sptr, "[type]", 6))
            {
               ContentTypePtr = sptr += 6;
               while (*sptr) sptr++;
               sptr++;
               if (!strncmp (sptr, "[content]", 9)) sptr += 9;
               FileSizeBytes -= sptr - DataPtr;
            }
         }
      }

      if (!NamePtr || !ContentTypePtr)
      {
         StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(SS$_BUGCHECK));
         return (FALSE);
      }

      if (!strncmp (cptr, "text/", 5))
      {
         if (WatchEnabled)
            WatchThis ("MIME QUOTED-PRINTABLE !UL bytes", FileSizeBytes); 
         MimeEncodedPtr = MimeEncQuotedPrintable (sptr, FileSizeBytes);
         if (!MimeEncodedPtr)
         {
            StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(SS$_BUGCHECK));
            return (FALSE);
         }
         TransferEncodingPtr = "quoted-printable";
      }
      else
      {
         if (WatchEnabled) WatchThis ("MIME BASE64 !UL bytes", FileSizeBytes); 
         MimeEncodedPtr = MimeEncBase64 (sptr, FileSizeBytes, "\r\n");
         if (!MimeEncodedPtr)
         {
            StatusMessage (FI_LI, 1, "%s: %s.", AttId, SysGetMsg(SS$_BUGCHECK));
            return (FALSE);
         }
         TransferEncodingPtr = "base64";
      }

      /* free the attachment file content */
      free (DataPtr);

      if (WatchEnabled)
      {
         WatchThis ("MIME NAME !AZ", NamePtr); 
         WatchThis ("MIME TYPE !AZ", ContentTypePtr); 
         WatchThis ("MIME TRANSFER !AZ", TransferEncodingPtr); 
      }

      /* first will point to the MIME part header */
      TEXT_DATA_CALLOC (tl1ptr, 512);
      if (!tl1ptr) ErrorExit (vaxc$errno, FI_LI);

      /* if there was a previous part then link it to this new part */
      if (tl2ptr) tl2ptr->NextPtr = tl1ptr;

      /* second will point to the MIME part encoded content */
      TEXT_DATA_CALLOC (tl2ptr, 0);
      if (!tl2ptr) ErrorExit (vaxc$errno, FI_LI);

      TEXT_DATA_TEXT(tl1ptr) = TEXT_DATA_SELF(tl1ptr);
      STRINIT (TEXT_DATA_TEXT(tl1ptr), TEXT_DATA_SIZE(tl1ptr)) 
      STRCAT ("\r\n--")
      STRCAT (MimeBoundary)
      STRCAT ("\r\nContent-Type: ")
      STRCAT (ContentTypePtr)
      STRCAT (";\r\n  name=\"")
      STRCAT (NamePtr)
      STRCAT ("\"\r\nContent-Transfer-Encoding: ")
      STRCAT (TransferEncodingPtr)
      STRCAT ("\r\nContent-Disposition: attachment;\r\n  filename=\"")
      STRCAT (NamePtr)
      STRCAT ("\"\r\n\r\n")
      STRNULL
      TEXT_DATA_LENGTH (tl1ptr) = sptr - TEXT_DATA_TEXT (tl1ptr);
      if (STROVF) ErrorExit (SS$_BUGCHECK, FI_LI);

      TEXT_DATA_TEXT (tl2ptr) = MimeEncodedPtr;
      TEXT_DATA_LENGTH (tl2ptr) = strlen(TEXT_DATA_TEXT (tl2ptr));

      /* if this is the first part of the MIME body */
      if (!TEXT_DATA_NEXT (&cdptr->MimeBody))
         TEXT_DATA_NEXT (&cdptr->MimeBody) = tl1ptr;

      TEXT_DATA_NEXT (tl1ptr) = tl2ptr;

      if (Debug)
         fprintf (stdout, "%d\n|%s|\n|%s|\n",
                  AttachFileCount, TEXT_DATA_TEXT(tl1ptr),
                  TEXT_DATA_TEXT(tl2ptr));

      fnptr = FileSpecSearch (FALSE, NULL, NULL);
   }

   /******************************/
   /* end of (any) attachment(s) */
   /******************************/

   if (AttachFileCount)
   {
      /* add the final trailing MIME boundary */
      TEXT_DATA_CALLOC (tl3ptr, 128);
      if (!tl3ptr) ErrorExit (vaxc$errno, FI_LI);
      TEXT_DATA_TEXT(tl3ptr) = TEXT_DATA_SELF(tl3ptr);
      STRINIT (TEXT_DATA_SELF(tl3ptr), TEXT_DATA_SIZE(tl3ptr)) 
      STRCAT ("\r\n--")
      STRCAT (MimeBoundary)
      STRCAT ("--\r\n")
      STRNULL
      TEXT_DATA_LENGTH (tl3ptr) = sptr - TEXT_DATA_TEXT (tl3ptr);
      if (STROVF) ErrorExit (SS$_BUGCHECK, FI_LI);
      TEXT_DATA_NEXT (tl2ptr) = tl3ptr;
   }
   else
   {
      /* make the text body in-line */
      TEXT_DATA_CALLOC (tl1ptr, 0);
      if (!tl1ptr) ErrorExit (vaxc$errno, FI_LI);
      TEXT_DATA_TEXT (tl1ptr) = MimeEncodedPtr;
      TEXT_DATA_LENGTH (tl1ptr) = strlen(TEXT_DATA_TEXT (tl1ptr));
      /* if this is the only part of the MIME body */
      TEXT_DATA_NEXT (&cdptr->MimeBody) = tl1ptr;
   }

   /***********************************/
   /* create an RFC822 message header */
   /***********************************/

   zptr = (sptr = HeaderBuffer) + sizeof(HeaderBuffer);

   CGIVARNULL (CgiServerNamePtr, "SERVER_NAME");
   if (!CgiServerNamePtr) CgiServerNamePtr = "?";

   CGIVARNULL (CgiRemoteHostPtr, "REMOTE_HOST");
   if (!CgiRemoteHostPtr) CgiRemoteHostPtr = "?";

   CGIVARNULL (CgiRemoteAddrPtr, "REMOTE_ADDR");
   if (!CgiRemoteAddrPtr) CgiRemoteAddrPtr = "?.?.?.?";

   CGIVARNULL (CgiRemotePortPtr, "REMOTE_PORT");
   if (!CgiRemotePortPtr) CgiRemotePortPtr = "?";

   CGIVARNULL (CgiHttpXForwardedForPtr, "HTTP_X_FORWARDED_FOR");

   zptr = (sptr = ScratchBuffer) + sizeof(ScratchBuffer);

   STRCAT ("from ")
   STRCAT (CgiRemoteHostPtr)
   if (strcmp (CgiRemoteHostPtr, CgiRemoteAddrPtr))
   {
      /* resolved host name in REMOTE_HOST */
      STRCAT (" (")
      STRCAT (CgiRemoteAddrPtr)
      if (CgiHttpXForwardedForPtr)
      {
         STRCAT (" forwarded for ")
         STRCAT (CgiHttpXForwardedForPtr)
      }
      STRCAT (")")
   }
   else
   if (CgiHttpXForwardedForPtr)
   {
      /* IP address in REMOTE_HOST */
      STRCAT ("(forwarded for ")
      STRCAT (CgiHttpXForwardedForPtr)
      STRCAT (")")
   }
   STRCAT (" for ")
   STRCAT (rdptr->UserName)
   STRCAT ("@")
   STRCAT (CgiServerNamePtr)
   STRCAT (" by ")
   STRCAT (LocalHostName)
   STRCAT (" (")
   STRCAT (SoftwareId)
   STRCAT (") with HTTP; ")
   STRCAT (MtaRfcBinDate(NULL))
   STRNULL

   if (STROVF)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }

   zptr = (sptr = HeaderBuffer) + sizeof(HeaderBuffer);

   len = InetMailFoldRfc822Header (sptr, zptr-sptr,
                                   "Received:", ScratchBuffer); 
   if (len < 0)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }
   sptr += len;

   sptr += sprintf (sptr,
"Message-ID: <%s.%s.%08.08x%08.08x.soymail@%s>\r\nDate: %s\r\n",
                    CgiRemoteAddrPtr, CgiRemotePortPtr,
                    BinTime[1], BinTime[0], LocalHostName,
                    MtaRfcBinDate(NULL));

   len = InetMailFoldRfc822Header (sptr, zptr-sptr, "Subject:",
                                   MimeEncIsoString(cdptr->SubjPtr,
                                                    CharSetPtr)); 
   if (len < 0)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }
   if (len)
      sptr += len;
   else
      STRCAT ("Subject: (none)\r\n")

   len = InetMailFoldRfc822Header (sptr, zptr-sptr, "From:",
                                   MimeEncIsoString(cdptr->FromPtr,
                                                    CharSetPtr)); 
   if (len < 0)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }
   sptr += len;

   if (cdptr->ReplyToPtr && cdptr->ReplyToPtr[0])
   {
      len = InetMailFoldRfc822Header (sptr, zptr-sptr, "Reply-to:",
                                      MimeEncIsoString(cdptr->ReplyToPtr,
                                                       CharSetPtr));
      if (len < 0)
      {
         StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
         return (FALSE);
      }
      sptr += len;
   }

   len = InetMailFoldRfc822Header (sptr, zptr-sptr, "To:",
                                   MimeEncIsoString(cdptr->ToPtr,
                                                    CharSetPtr));
   if (len < 0)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }
   sptr += len;

   len = InetMailFoldRfc822Header (sptr, zptr-sptr, "CC:",
                                   MimeEncIsoString(cdptr->CcPtr,
                                                    CharSetPtr)); 
   if (len < 0)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }
   sptr += len;

   if (cdptr->BccPtr && cdptr->BccPtr[0]) STRCAT ("BCC:\r\n")

   STRCAT ("MIME-Version: 1.0\r\n")

   if (MimeBoundary[0])
   {
      STRCAT ("Content-Type: multipart/mixed;\r\n  boundary=\"")
      STRCAT (MimeBoundary)
      STRCAT ("\"\r\n")
   }
   else
   if (HtmlTextPtr)
   {
      STRCAT ("Content-Type: text/html; charset=")
      STRCAT (CharSetPtr)
      STRCAT ("\r\nContent-Transfer-Encoding: quoted-printable\r\n")
   }
   else
   if (QuotedPrintableTextPtr)
   {
      STRCAT ("Content-Type: text/plain; charset=")
      STRCAT (CharSetPtr)
      STRCAT ("\r\nContent-Transfer-Encoding: quoted-printable\r\n")
   }
   else
   {
      STRCAT ("Content-Type: text/plain; charset=")
      STRCAT (CharSetPtr)
      STRCAT ("\r\nContent-Transfer-Encoding: 7bit\r\n")
   }

   STRCAT ("X-Mailer: ")
   STRCAT (SoftwareId)
   STRCAT ("\r\n\r\n")
   STRNULL

   if (STROVF)
   {
      StatusMessage (FI_LI, 1, "RFC822: %s.", SysGetMsg(SS$_RESULTOVF));
      return (FALSE);
   }

   len = sptr - HeaderBuffer;
   sptr = calloc (1, len);
   if (!sptr) ErrorExit (vaxc$errno, FI_LI);
   memcpy (sptr, HeaderBuffer, len+1);

   if (WatchEnabled) WatchThis ("RFC822 HEADER\n!AZ", HeaderBuffer);

   TEXT_DATA_TEXT (&cdptr->MimeHeader) = sptr;
   TEXT_DATA_SIZE (&cdptr->MimeHeader) = len+1;
   TEXT_DATA_LENGTH (&cdptr->MimeHeader) = len;

   return (TRUE);
}                              

/*****************************************************************************/
/*
Send the mail message contained in the 'compose data'.
*/

void ComposeSendMessage (REQUEST_DATA *rdptr)

{
   int  len;
   char  *cptr, *sptr, *zptr,
         *AddrListPtr,
         *SmtpServerPtr;
   COMPOSE_DATA  *cdptr;
   USER_OPTIONS  *uoptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeSendMessage()\n");

   cdptr = &rdptr->ComposeData;
   uoptr = &rdptr->UserOptions;

   for (cptr = cdptr->ToPtr; *cptr && isspace(*cptr); cptr++);
   if (!*cptr)
   {
      StatusMessage (FI_LI, 1, "%s", LangFor("compose_no_recipient"));
      return;
   }
   if (!rdptr->ClientHasJavaScript)
   {
      for (cptr = cdptr->SubjPtr; *cptr && isspace(*cptr); cptr++);
      if (!*cptr)
      {
         StatusMessage (FI_LI, 1, "%s", LangFor("compose_no_subject"));
         return;
      }
      for (cptr = cdptr->TextPtr; *cptr && isspace(*cptr); cptr++);
      if (!*cptr)
      {
         StatusMessage (FI_LI, 1, "%s", LangFor("compose_no_text"));
         return;
      }
   }

   /* ensure each of the addresses are comma-delimitted */
   cdptr->ToPtr = AddressListMassage (cdptr->ToPtr, TRUE);
   cdptr->CcPtr = AddressListMassage (cdptr->CcPtr, TRUE);
   cdptr->BccPtr = AddressListMassage (cdptr->BccPtr, TRUE);

   if (!cdptr->FromPtr) cdptr->FromPtr = cdptr->SelfPtr;

   /* if it looks like one or more RFC822 addresses */ 
   if (AddressIsRfc(cdptr->ToPtr) ||
       AddressIsRfc(cdptr->CcPtr) ||
       AddressIsRfc(cdptr->BccPtr) ||
       AddressIsRfc(cdptr->ReplyToPtr))
   {
      /************************/
      /* send via an SMTP MTA */
      /************************/

      if (AddressIsVms(cdptr->ToPtr) ||
          AddressIsVms(cdptr->CcPtr) ||
          AddressIsVms(cdptr->BccPtr) ||
          AddressIsVms(cdptr->ReplyToPtr))
      {
         StatusMessage (FI_LI, 1, "%s", LangFor("compose_either_or"));
         return;
      }

      ComposeMime (rdptr);

      /* create a single list of all the addresses for the MTA SMTP server */
      len = strlen(cdptr->ToPtr) +
            strlen(cdptr->CcPtr) +
            strlen(cdptr->BccPtr) + 8;
      AddrListPtr = calloc (1, len);
      if (!AddrListPtr) ErrorExit (vaxc$errno, FI_LI);
      sprintf (AddrListPtr, "%s%s%s%s%s",
               cdptr->ToPtr,
               cdptr->CcPtr[0] ? "\n" : "", cdptr->CcPtr,
               cdptr->BccPtr[0] ? "\n" : "", cdptr->BccPtr);

      SmtpServerPtr = SoyMailConfig.SmtpServerHost;
      if (!SmtpServerPtr) SmtpServerPtr = "localhost";

      cptr = MtaMsgSmtpToServer (SmtpServerPtr,
                                 cdptr->FromPtr,
                                 AddrListPtr,
                                 NULL,
                                 &cdptr->MimeHeader,
                                 NULL,
                                 &cdptr->MimeBody);
      if (cptr) StatusMessage (FI_LI, 1, "SMTP: %s", HTML_ESCAPE(cptr));
   }
   else
   {
      /****************************************/
      /* otherwise send via VMS callable mail */
      /****************************************/

      /* ensure that the body of the message is correctly wrapped */
      cdptr->TextPtr = ComposeBodyTextMassage (cdptr->TextPtr,
                                               cdptr->WrapAt,
                                               *uoptr->QuoteChar);

      cptr = SendMailMessage (rdptr->RemoteUser,
                              cdptr->ToPtr,
                              cdptr->CcPtr,
                              cdptr->SubjPtr,
                              cdptr->TextPtr);
      if (cptr) StatusMessage (FI_LI, 1, "MAIL: %s", cptr);
   }

   if (!cptr)
   {
      StatusMessage (FI_LI, 0, "%s", LangFor("compose_sent_ok"));
      cdptr->AlreadySent = TRUE;
   }
}                              

/*****************************************************************************/
/*
*/

void ComposeSendClose (REQUEST_DATA *rdptr)

{
   int  NewMailCount;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ComposeSendClose()\n");

   CgiLibResponseHeader (200, "text/html",
                         "Script-Control: X-content-encoding-gzip=0\n");

   fprintf (stdout,
"<SCRIPT LANGUAGE=\"JavaScript\">window.close()</SCRIPT>\n");
}

/*****************************************************************************/

