/******************************************************************************/
/*
                                  compose.h
*/
/******************************************************************************/

#ifndef INCLUDE_COMPOSE_H
#define INCLUDE_COMPOSE_H

#define BOOL int
#define TRUE 1
#define FALSE 0

#define COMPOSE_EDITOR_PLAIN 1
#define COMPOSE_EDITOR_HTML  2

/* forward references */
typedef struct STRUCT_REQUEST_DATA REQUEST_DATA;
typedef struct STRUCT_MIME_DATA MIME_DATA;

typedef struct STRUCT_COMPOSE_DATA
{
   /* this storage is used from request-to-request */

   BOOL  /* message has already been sent at least once */
         AlreadySent,
         /* the [SELF] has already been placed into the CC or BCC */
         CopiedSelf,
         /* looks like RFC addressing (if the TO or CC fields contain a '@') */
         LooksLikeRfc,
         /* a CC address has been provided (or requested) */
         ShowCcAddr,
         /* a BCC address has be provided (or requested) */
         ShowBccAddr,
         /* a reply-to address has be provided (or requested) */
         ShowReplyToAddr;

   int  /* the MIME content-type (plain or HTML) */
        ContentTypeIs,
        /* number of columns in edit window */
        EditCols,
        /* number of rows in edit window */
        EditRows,
        /* default plain-text, or optional HTML */
        EditorType,
        /* wrap plain-text message body at column */
        WrapAt;

   char  /* "RE:", "FW:" */
         *AttribPtr,
         /* list of addresses the message is to be sent BLIND to */
         *BccPtr,
         /* list of addresses the message is to be CCed to */
         *CcPtr,
         /* character set of the plain-text message */
         *CharSetPtr,
         /* list of contacts loaded from file */
         *ContactListPtr,
         /* date/time restored draft was saved */
         *DraftSavedOnPtr,
         /* ostensible source (sender) */
         *FromPtr,
         /* explicit address to reply to */
         *ReplyToPtr,
         /* provides [SELF] in list of contacts */
         *SelfPtr,
         /* subject line */
         *SubjPtr,
         /* message body plain-text */
         *TextPtr,
         /* list of addresses the message is to be sent to */
         *ToPtr;

   TEXT_DATA  MimeBody,
              MimeHeader;

} COMPOSE_DATA;

/**************/
/* prototypes */
/**************/

BOOL ComposeMime (REQUEST_DATA*);
void ComposePage (REQUEST_DATA*);
BOOL ComposePageRequest (REQUEST_DATA*);
void ComposeRestoreDraft (REQUEST_DATA*);
void ComposeSaveDraft (REQUEST_DATA*);
void ComposeSelfAddress (REQUEST_DATA*);
void ComposeSendMessage (REQUEST_DATA*);
void ComposeSendClose (REQUEST_DATA*);
int ComposeStripNonPrint (char*);
char* ComposeBodyTextMassage (char*, int, char);

#endif /* INCLUDE_COMPOSE_H */

/******************************************************************************/

