/*****************************************************************************/
/*
                                  folder.c

Provide HTML form selection lists and buttons for VMS Mail files/folders.


COPYRIGHT
---------
Copyright (C) 2005,2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
09-APR-2006  MGD  FolderOpenSingleFileListEach() support public wildcard
                    folder configuration with MAIL and NEWMAIL prohibition
06-APR-2006  MGD  remove default MAIL and NEWMAIL folders
02-APR-2006  MGD  FolderOpenSelect() modifications for public wildcard access
07-FEB-2006  MGD  folder buttons for left-side of message listing
01-FEB-2005  MGD  initial
*/

/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* VMS related header files */
#include <descrip.h>
#include <mailmsgdef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header file */
#include "soymail.h"
#include "attach.h"
#include "callmail.h"
#include "cgilib.h"
#include "compose.h"
#include "folder.h"
#include "mainmenu.h"
#include "newmail.h"
#include "other.h"
#include "message.h"

#define FI_LI __FILE__, __LINE__

/********************/
/* external storage */
/********************/

extern BOOL  Debug,
             WatchEnabled;

extern int  VmsVersion;

extern char  CurrentVmsTimeString[],
             SoftwareId[],
             SoftwareVn[];

extern USER_OPTIONS  UserOptions;
extern REQUEST_DATA  RequestData;
extern VMS_MAIL_USER  VmsMailUser;


/*****************************************************************************/
/*
Reset (empty) the mail file and folder lists.
Called upon certain triggers and periodically.
*/

void FolderListReset (REQUEST_DATA *rdptr)

{
   int  cnt, status;
   char  *cptr, *sptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   muptr = &VmsMailUser;

   if (WatchEnabled) WatchThis ("FolderListReset()");

   if (muptr->MailFileListPtr)
   {
      free (muptr->MailFileListPtr);
      muptr->MailFileListPtr = NULL;
      muptr->MailFileListSize =
         muptr->MailFileListLength =
         muptr->MailFileListCount = 0;
   }

   if (muptr->FolderListPtr)
   {
      free (muptr->FolderListPtr);
      muptr->FolderListPtr = NULL;
      muptr->FolderListSize =
         muptr->FolderListLength =
         muptr->FolderListCount = 0;
   }
}

/*****************************************************************************/
/*
Get the list of mail files and a list of folders in the default mail file.
*/

void FolderListInit (REQUEST_DATA *rdptr)

{
   int  cnt, status;
   char  *cptr, *sptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   if (WatchEnabled)
      WatchThis ("FolderListInit() !AZ",
                 muptr->FolderListPtr ? "REUSE" : "GET");

   if (!uoptr->UseOnlyMAILFile &&
       !muptr->MailFileListPtr)
   {
      status = CallMailFileList (muptr);
      if (VMSnok (status))
         StatusMessage (FI_LI, 1, "MAIL: %s.", SysGetMsg(status));
      if (WatchEnabled) WatchThis ("MAILFILES !UL", muptr->MailFileListCount);
      time ((unsigned long*)&rdptr->FolderRefreshSeconds);
   }

   if (!muptr->FolderListPtr)
   {
      CallMailFolderList (muptr);
      if (VMSnok (muptr->MailVmsStatus))
         StatusMessage (FI_LI, 1, "MAIL: %s.",
                        SysGetMsg(muptr->MailVmsStatus));
      if (WatchEnabled) WatchThis ("FOLDERS !UL", muptr->FolderListCount);
      time ((unsigned long*)&rdptr->FolderRefreshSeconds);
   }
}

/*****************************************************************************/
/*
Generate a selection list for the available mail folders and [open] button. 
If private access then include the [9 new] button.
*/

void FolderOpenSelect (REQUEST_DATA *rdptr)

{
   int  status;
   char  newBuf [256];
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderOpenSelect()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   fprintf (stdout,
"<!-- folder open begin -->\n\
<INPUT TYPE=\"hidden\" NAME=\"folder_list_change\" VALUE=\"\">\
<SELECT CLASS=\"css_select_2\" NAME=\"folder_list_open\" \
ONCHANGE=\"folder_list_change.value=true;startWorking();submit();\">\n");

   if (uoptr->UseOnlyMAILFile || muptr->MailFileListCount <= 1)
      fprintf (stdout,
"<OPTION VALUE=\"%s\" STYLE=\"padding-left:0.5em;\">%s\n",
            LangFor("(none)"), LangFor("(none)"));
   else
      fprintf (stdout,
"<OPTION VALUE=\"%s/%s\" STYLE=\"padding-left:0.5em;\">%s\n",
            rdptr->MailFileName, LangFor("(none)"), LangFor("(none)"));

   FolderOpenListEach (rdptr);

   sprintf (newBuf, "%d %s", rdptr->NewMessages, LangFor("new"));

   fprintf (stdout,
"</SELECT>&nbsp;&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"main_btn\" VALUE=\"%s\"%s>&nbsp;\n",
            LangFor("open"),
            CLICK_STARTWORKING);

   if (rdptr->PrivateAccess)
      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"main_btn\" VALUE=\"%s\"%s>&nbsp; \n",
               newBuf,
               NewMailInclude(rdptr) ?
                  " ONCLICK=\"refreshNewMail(-1);return false;\"" :
                  CLICK_STARTWORKING);

   fprintf (stdout,
"<!-- folder open end -->\n\n");
}

/*****************************************************************************/
/*
*/

void FolderOpenListEach (REQUEST_DATA *rdptr)

{
   int  cnt;
   char  *cptr, *sptr, *tptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderOpenListEach()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   if (rdptr->PublicWildcard ||
       uoptr->UseOnlyMAILFile ||
       muptr->MailFileListCount <= 1)
      FolderOpenSingleFileListEach (rdptr);
   else
      FolderOpenMultiFileListEach (rdptr);
}

/*****************************************************************************/
/*
Only a single MAIL.MAI file or optioned to only show MAIL.MAI.
List each folder.
*/

void FolderOpenSingleFileListEach (REQUEST_DATA *rdptr)

{
   int  fldcnt, idx;
   char  *cptr, *fldptr, *escptr, *selptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderOpenSingleFileListEach()");

   muptr = &VmsMailUser;

   if (!rdptr->NewMessages && !strcmp (rdptr->FolderName, "NEWMAIL"))
      fputs ("<OPTION VALUE=\"NEWMAIL\" STYLE=\"padding-left:0.5em;\" \
SELECTED>NEWMAIL\n", stdout);

   fldptr = muptr->FolderListPtr;
   for (fldcnt = muptr->FolderListCount; fldcnt; fldcnt--)
   {
      if (rdptr->PublicWildcard &&
          rdptr->PublicNoNewMail &&
          (!strcmp (fldptr, "MAIL") ||
           !strcmp (fldptr, "NEWMAIL")))
      {
         /* next folder in list */
         while (*fldptr) fldptr++;
         fldptr++;
         continue;
      }

      if (!strcmp (fldptr, rdptr->FolderName))
         selptr = " SELECTED";
      else
         selptr = "";
      escptr = HTML_ESCAPE(fldptr);
      fprintf (stdout,
"<OPTION VALUE=\"%s\" STYLE=\"padding-left:0.5em;\"%s>%s\n",
               escptr, selptr, escptr);

      /* next folder in list */
      while (*fldptr) fldptr++;
      fldptr++;
   }
}

/*****************************************************************************/
/*
Multiple mail .MAI files.
List each folder for the currently selected mail .MAI file.
*/

void FolderOpenMultiFileListEach (REQUEST_DATA *rdptr)

{
   int  filcnt, fldcnt, idx;
   char  *cptr, *filptr, *fldptr, *escptr, *selptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderOpenMultiFileListEach()");

   muptr = &VmsMailUser;

   if (!rdptr->NewMessages && !strcmp (rdptr->FolderName, "NEWMAIL"))
      fputs ("<OPTION VALUE=\"NEWMAIL\" STYLE=\"padding-left:0.5em;\" \
SELECTED>NEWMAIL\n", stdout);

   filptr = muptr->MailFileListPtr;
   for (filcnt = muptr->MailFileListCount; filcnt; filcnt--)
   {
      fprintf (stdout,
"<OPTION VALUE=\"%s/MAIL\" STYLE=\"padding-left:0.5em;\">%s/\n",
               filptr, filptr);

      /* if it's not equal to the current (destination) mail file */
      if (strcmp (filptr, rdptr->MailFileName))
      {
         /* next file in list */
         while (*filptr) filptr++;
         filptr++;
         continue;
      }

      fldptr = muptr->FolderListPtr;
      for (fldcnt = muptr->FolderListCount; fldcnt; fldcnt--)
      {
         /* selected if it's the current (destination) folder name */
         if (!strcmp (fldptr, rdptr->FolderName))
            selptr = " SELECTED";
         else
            selptr = "";
         escptr = HTML_ESCAPE(fldptr);
         fprintf (stdout,
"<OPTION VALUE=\"%s/%s\"STYLE=\"padding-left:2em;\"%s>/%s\n",
                     filptr, escptr, selptr, escptr);

         /* next folder in list */
         while (*fldptr) fldptr++;
         fldptr++;
      }

      /* next mail file in list */
      while (*filptr) filptr++;
      filptr++;
   }
}

/*****************************************************************************/
/*
Differences in the CSS rendering of (only) buttons inside table data (the CSS
lay-up technicalities are well outside any expertise I might have) for Firefox
1.5, Mozilla 1.7, Opera 8.5 and MSIE 6.0 make this guestimate of the width of
the required buttons absolutely necessary.  I played around with the various
bits of CSS I understood and still got vastly differing behaviours between the
agents and eventually resorted to this nightmare.  Talk about making it up as
you go along!
*/

void FolderButtonListEach (REQUEST_DATA *rdptr)

{
   int  em, filcnt, fldcnt, idx, lcem, maxem;
   char  *cptr, *filptr, *fldptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderButtonListEach()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   CGIVAR (cptr, "HTTP_USER_AGENT");
   /* MSIE seems to render the fonts slightly wider (?) */
   if (strstr (cptr, "MSIE")) lcem = 6; else lcem = 5;

   /* guess the maximum number of 'em's required for the button width */
   maxem = 0;
   if (!(uoptr->UseOnlyMAILFile || muptr->MailFileListCount <= 1))
   {
      /* mail file names */
      filptr = muptr->MailFileListPtr;
      for (filcnt = muptr->MailFileListCount; filcnt; filcnt--)
      {
         em = 0;
         for (cptr = filptr; *cptr; cptr++)
            if (islower(*cptr)) em += lcem; else em += 8;
         if (em > maxem) maxem = em;
         /* next mail file */
         while (*filptr) filptr++;
         filptr++;
      }
   }
   /* folder names */
   fldptr = muptr->FolderListPtr;
   for (fldcnt = muptr->FolderListCount; fldcnt; fldcnt--)
   {
      em = 0;
      for (cptr = fldptr; *cptr; cptr++)
         if (islower(*cptr)) em += lcem; else em += 8;
      if (em > maxem) maxem = em;
      /* next folder */
      while (*fldptr) fldptr++;
      fldptr++;
   }

   if (uoptr->UseOnlyMAILFile || muptr->MailFileListCount <= 1)
   {
      maxem = (maxem - 16) / 8;
      FolderButtonSingleFileListEach (rdptr, maxem);
   }
   else
   {
      maxem = maxem / 8;
      FolderButtonMultiFileListEach (rdptr, maxem);
   }
}

/*****************************************************************************/
/*
Only a single MAIL.MAI file or optioned to only show MAIL.MAI.
Provide a button for each folder.
*/

void FolderButtonSingleFileListEach
(
REQUEST_DATA *rdptr,
int ButtonWidth
)
{
   int  cnt, fldcnt, idx;
   char  *cptr, *fldptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderButtonSingleFileListEach()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   cnt = 0;
   fldptr = muptr->FolderListPtr;
   for (fldcnt = muptr->FolderListCount; fldcnt; fldcnt--)
   {
      if (cnt && !(cnt % uoptr->FolderButtons)) 
         fprintf (stdout, "<P CLASS=\"css_p02\">");
      cnt++;

      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%dem;text-align:left;padding-left:0.5em;\" \
TYPE=\"submit\" NAME=\"folder_btn\" VALUE=\"%s\">\n",
               ButtonWidth, HTML_ESCAPE(fldptr));

      /* next folder in list */
      while (*fldptr) fldptr++;
      fldptr++;
   }
}

/*****************************************************************************/
/*
Multiple mail .MAI files.
Provide a button for each mail file and each folder for the currently selected
mail .MAI file.
*/

void FolderButtonMultiFileListEach
(
REQUEST_DATA *rdptr,
int ButtonWidth
)
{
   int  cnt, filcnt, fldcnt, idx;
   char  *cptr, *filptr, *fldptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderButtonMultiFileListEach()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   /* folder buttons don't contain the mail file, supply using this */
   fprintf (stdout,
"<INPUT TYPE=\"hidden\" NAME=\"folder_btn_mailfile\" VALUE=\"%s\">\n",
            rdptr->MailFileName); 

   cnt = 0;
   filptr = muptr->MailFileListPtr;
   for (filcnt = muptr->MailFileListCount; filcnt; filcnt--)
   {
      if (cnt && !(cnt % uoptr->FolderButtons)) 
         fprintf (stdout, "<P CLASS=\"css_p02\">");
      cnt++;

      fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%dem;text-align:left;padding-left:0.5em;\" \
TYPE=\"submit\" NAME=\"folder_btn\" VALUE=\"%s/\">\n",
               ButtonWidth, filptr);

      /* if it's not equal to the current (destination) mail file */
      if (strcmp (filptr, rdptr->MailFileName))
      {
         /* next file in list */
         while (*filptr) filptr++;
         filptr++;
         continue;
      }

      fldptr = muptr->FolderListPtr;
      for (fldcnt = muptr->FolderListCount; fldcnt; fldcnt--)
      {
         if (cnt && !(cnt % uoptr->FolderButtons)) 
            fprintf (stdout, "<P CLASS=\"css_p02\">");
         cnt++;

         fprintf (stdout,
"<INPUT CLASS=\"css_button_panel_1\" \
STYLE=\"width:%dem;text-align:left;padding-left:1.5em;\" \
TYPE=\"submit\" NAME=\"folder_btn\" VALUE=\"/%s\">\n",
                  ButtonWidth, HTML_ESCAPE(fldptr));
         /* next folder in list */
         while (*fldptr) fldptr++;
         fldptr++;
      }

      /* next mail file in list */
      while (*filptr) filptr++;
      filptr++;
   }
}

/*****************************************************************************/
/*
Provide a selection list for the message action destination folders.
*/

void FolderDestListEach (REQUEST_DATA *rdptr)

{
   int  cnt;
   char  *cptr, *sptr, *tptr;
   USER_OPTIONS  *uoptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderDestListEach()");

   muptr = &VmsMailUser;
   uoptr = &rdptr->UserOptions;

   if (uoptr->UseOnlyMAILFile || muptr->MailFileListCount <= 1)
      FolderDestSingleFileListEach (rdptr);
   else
      FolderDestMultiFileListEach (rdptr);
}

/*****************************************************************************/
/*
List each folder available as a destination for [copy] or [move].
*/

void FolderDestSingleFileListEach (REQUEST_DATA *rdptr)

{
   int  fldcnt, idx;
   char  *cptr, *fldptr, *escptr, *selptr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderDestSingleFileListEach()");

   muptr = &VmsMailUser;

   fldptr = muptr->FolderListPtr;
   for (fldcnt = muptr->FolderListCount; fldcnt; fldcnt--)
   {
      /* it's not any of the default folders */
      if (!strcmp (fldptr, rdptr->DestFolderName))
         selptr = " SELECTED";
      else
         selptr = "";
      escptr = HTML_ESCAPE(fldptr);
      fprintf (stdout,
"<OPTION VALUE=\"%s\" STYLE=\"padding-left:0.5em;\"%s>%s\n",
               escptr, selptr, escptr);
      /* next folder in list */
      while (*fldptr) fldptr++;
      fldptr++;
   }
}

/*****************************************************************************/
/*
List each mail file and possible folder available as a destination for [copy]
or [move].
*/

void FolderDestMultiFileListEach (REQUEST_DATA *rdptr)

{
   int  filcnt, fldcnt, idx;
   char  *cptr, *filptr, *fldptr, *escptr, *selptr,
         *SoyTempPtr;
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderDestMultiFileListEach()");

   muptr = &VmsMailUser;

   SoyTempPtr = LangFor("folder_soy_temp");

   filptr = muptr->MailFileListPtr;
   for (filcnt = muptr->MailFileListCount; filcnt; filcnt--)
   {
      fprintf (stdout,
"<OPTION VALUE=\"%s/_%s\" STYLE=\"padding-left:0.5em;\"%s>%s/\n",
               filptr, SoyTempPtr, "", filptr);

      /* if it's not equal to the current mail file */
      if (strcmp (filptr, rdptr->MailFileName))
      {
         /* next file in list */
         while (*filptr) filptr++;
         filptr++;
         continue;
      }

      fldptr = muptr->FolderListPtr;
      for (fldcnt = muptr->FolderListCount; fldcnt; fldcnt--)
      {
         /* selected if it's the current (destination) folder name */
         if (!strcmp (fldptr, rdptr->DestFolderName))
            selptr = " SELECTED";
         else
            selptr = "";
         escptr = HTML_ESCAPE(fldptr);

         fprintf (stdout,
"<OPTION VALUE=\"%s/%s\"STYLE=\"padding-left:2em;\"%s>/%s\n",
                  filptr, escptr, selptr, escptr);

         /* next folder in list */
         while (*fldptr) fldptr++;
         fldptr++;
      }

      /* next mail file in list */
      while (*filptr) filptr++;
      filptr++;
   }
}

/*****************************************************************************/
/*
A faked folder select specifically for the help menu bar page.
*/

void FolderOpenHelp (REQUEST_DATA *rdptr)

{
   int  status;
   char  newBuf [256];
   VMS_MAIL_USER  *muptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderOpenHelp()");

   fprintf (stdout,
"<!-- folder open help begin -->\n\
<TD NOWRAP ALIGN=\"left\" VALIGN=\"middle\">\n\
<SELECT CLASS=\"css_select_2\" NAME=\"folder_list_help\">\n\
<OPTION VALUE=\"%s\" SELECTED>&nbsp;%s&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\n\
</SELECT>&nbsp;&nbsp;\n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"main_btn\" VALUE=\"%s\">&nbsp; \n\
<INPUT CLASS=\"css_button_panel_2\" TYPE=\"submit\" \
NAME=\"main_btn\" VALUE=\"0 %s\">&nbsp; \n\
</TD>\n\
<!-- folder open help end -->\n\n",
            LangFor("help"), LangFor("help"),
            LangFor("open"),
            LangFor("new"));
}

/*****************************************************************************/
/*
Generate an appropriate folder description (i.e. if it's not the default mail
file then include it in the description) and return a pointer to dynamic
storage.
*/

char* FolderDescribe
(
char *MailFileName,
char *FolderName
)
{
   char  *cptr, *sptr, *zptr;
   char  buf [256];

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("FolderDescribe()");

   zptr = (sptr = buf) + sizeof(buf)-1;
   if (strcmp (MailFileName, "MAIL"))
   {
      /* not the default MAIL.MAI */
      for (cptr = MailFileName; *cptr && sptr < zptr; *sptr++ = *cptr++);
      for (cptr = " / "; *cptr && sptr < zptr; *sptr++ = *cptr++);
   }
   for (cptr = FolderName; *cptr && sptr < zptr; *sptr++ = *cptr++);
   *sptr = '\0';
   cptr = calloc (1, sptr-buf+1);
   strcpy (cptr, buf);
   return (cptr);
}

/*****************************************************************************/

