/*****************************************************************************/
/*
                                  HTML.c

Processing associated with reading and composing HTML-format messages.


COPYRIGHT
---------
Copyright (C) 2005,2006 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under the conditions of the GNU GENERAL PUBLIC LICENSE, version 2.


VERSION HISTORY
---------------
14-MAR-2006  MGD  refine HtmlNumericDeEntify() and HtmlCharacterDeEntify()
13-MAR-2006  MGD  HtmlSanitise() suppress "background-image"
01-FEB-2005  MGD  initial
*/

/*****************************************************************************/

#ifdef SOYMAIL_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60000000
#undef __CRTL_VER
#define __CRTL_VER 60000000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stat.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <types.h>

/* VMS related header files */
#include <descrip.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header file */
#include "soymail.h"
#include "config.h"
#include "html.h"
#include "message.h"

#define FI_LI __FILE__, __LINE__

/******************/
/* global storage */
/******************/

/********************/
/* external storage */
/********************/

extern BOOL  Debug,
             WatchEnabled;

extern CONFIG_DATA  SoyMailConfig;

/*****************************************************************************/
/*
Reduce HTML markup down to a semblance of plain-text.  Relies on the underlying
HTML to provide white-space, indenting, etc.  Essentially just strips the
markup out and adds carriage-control as necessary.  Will format ordered and
unordered lists.
*/

int HtmlToPlain (char *HtmlPtr)

{
#define MAX_LIST_DEPTH 16

   int  ListIndex;
   int  ListType [MAX_LIST_DEPTH+1];
   char  *cptr, *sptr, *tptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("HtmlToPlain()");

   /* ensure all potentially 'interesting' entities are revealed */
   HtmlNumericDeEntify (HtmlPtr);

   ListIndex = 0;
   memset (&ListType, 0, sizeof(ListType));

   cptr = sptr = HtmlPtr;
   while (*cptr)
   {
      while (*cptr && *cptr != '<') *sptr++ = *cptr++;
      if (!*cptr) break;

      if (!memcmp (cptr, "<!--", 4))
      {
         /* absorb all comments */
         cptr += HtmlSkipTagTo (cptr, "-->");
         continue;
      }
      if (HtmlThisTag (cptr, "SCRIPT"))
      {
         cptr += HtmlSkipTagTo (cptr, "</SCRIPT>");
         continue;
      }
      if (HtmlThisTag (cptr, "STYLE"))
      {
         cptr += HtmlSkipTagTo (cptr, "</STYLE>");
         continue;
      }
      if (HtmlThisTag (cptr, "IMG"))
      {
         cptr += HtmlSkipTagTo (cptr, ">");
         continue;
      }

      if (HtmlThisTag (cptr, "P"))
         *sptr++ = '\n';
      else
      if (HtmlThisTag (cptr, "B"))
         *sptr++ = '*';
      else
      if (HtmlThisTag (cptr, "U"))
         *sptr++ = '_';
      else
      if (HtmlThisTag (cptr, "I"))
         *sptr++ = '/';
      else
      if (HtmlThisEndTag (cptr, "LI"))
         *sptr = '\0';
      else
      if (HtmlThisTag (cptr, "LI"))
      {
         if (ListType[ListIndex] < 0)
         {
            *sptr++ = ' ';
            *sptr++ = '*';
            *sptr++ = ' ';
            *sptr++ = ' ';
         }
         else
         {
            *sptr++ = ' ';
            *sptr++ = ListType[ListIndex];
            *sptr++ = '.';
            *sptr++ = ' ';
            if (ListType[ListIndex] < 'z') ListType[ListIndex]++;
            if (ListType[ListIndex] == '9' + 1) ListType[ListIndex] = 'a';
         }
      }
      else
      if (HtmlThisEndTag (cptr, "UL"))
      {
         if (ListIndex > 1) ListIndex--;
      }
      else
      if (HtmlThisTag (cptr, "UL"))
      {
         if (ListIndex < MAX_LIST_DEPTH) ListIndex++;
         ListType[ListIndex] = -1;
      }
      else
      if (HtmlThisEndTag (cptr, "OL"))
      {
         if (ListIndex > 1) ListIndex--;
      }
      else
      if (HtmlThisTag (cptr, "OL"))
      {
         if (ListIndex < MAX_LIST_DEPTH) ListIndex++;
         ListType[ListIndex] = '1';
      }

      cptr += HtmlSkipTagTo (cptr, ">");
   }
   *sptr = '\0';

   HtmlCharacterDeEntify (HtmlPtr);

   return (MessageCollapseLineBreaks (HtmlPtr, 2));
}

/*****************************************************************************/
/*
Remove all HTML tags except those (probably) harmless tags permitted in the
tests.  Ensure only known anchor HREFs schemes are permitted.  In particular
try to disable all scripting capabilities.  Bit rude and crude and with lots of
assumptions!  Done in situ because resultant text will always be smaller or the
same size.  Return number of characters remaining.

CAUTION!!  Sanitising anything is like skating on thin ice!
*/

int HtmlSanitise (char *HtmlPtr)

{
   char  *cptr, *sptr, *tptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("HtmlSanitise()");

   /* ensure all potentially 'interesting' entities are revealed */
   HtmlNumericDeEntify (HtmlPtr);

   /* sanitise */
   cptr = sptr = HtmlPtr;
   while (*cptr)
   {
      while (*cptr && *cptr != '<') *sptr++ = *cptr++;
      if (!*cptr) break;

      if (!memcmp (cptr, "<!--", 4))
      {
         /* absorb all comments */
         cptr += HtmlSkipTagTo (cptr, "-->");
         continue;
      }

      if (HtmlThisTag (cptr, "TITLE"))
      {
         /* absorb the title */
         cptr += HtmlSkipTagTo (cptr, "</TITLE>");
         continue;
      }

      if (HtmlThisTag (cptr, "SCRIPT"))
      {
         /**********/
         /* script */
         /**********/

         /* absorb and add the following warning to the displayed page */
         static char  ScriptHere [] = "<B>&lt;SCRIPT&gt;</B>";
         cptr += HtmlSkipTagTo (tptr = cptr, "</SCRIPT>");
         if (cptr - tptr >= sizeof(ScriptHere))
            for (tptr = ScriptHere; *tptr; *sptr++ = *tptr++);
         continue;
      }

      if (HtmlThisTag (cptr, "IMG"))
      {
         /*********/
         /* image */
         /*********/

         /* absorb and add the following informational to the displayed page */
         static char  ImgHere [] = "<B>&lt;IMG&gt;</B>";
         cptr += HtmlSkipTagTo (tptr = cptr, ">");
         if (cptr - tptr >= sizeof(ImgHere))
            for (tptr = ImgHere; *tptr; *sptr++ = *tptr++);
         continue;
      }

      if (HtmlThisTag (cptr, "STYLE"))
      {
         /*********/
         /* style */
         /*********/

         /* lose all attributes other than what's specifically allowed */
         while (*cptr && *cptr != '>' && !isspace(*cptr)) *sptr++ = *cptr++;

         while (*cptr)
         {
            while (isspace (*cptr)) cptr++;
            if (!*cptr || *cptr == '>') break;

            if (strsame (cptr, "TYPE=", 5)) 
            {
               *sptr++ = ' ';
               cptr += HtmlSpanAttrib (tptr = cptr);
               while (tptr < cptr) *sptr++ = *tptr++;
               continue;
            }

            /* absorb the attribute */
            cptr += HtmlSpanAttrib (cptr);
         }

         /* now, include any intervening embedded style */
         cptr += HtmlSkipTagTo (tptr = cptr, "</STYLE>");
         while (tptr < cptr)
         {
            /* do not allow any importing of style sheets! */
            if (*tptr == '@' && strsame (tptr, "@import", 7))
            {
               memcpy (sptr, "_imp0rt", 7);
               sptr += 7;
               tptr += 7;
            }
            else
            if (tolower(*tptr) == 'b' &&
                strsame (tptr, "background-image", 16))
            {
               memcpy (sptr, "backgr0und_image", 16);
               sptr += 16;
               tptr += 16;
            }
            else
               *sptr++ = *tptr++;
         }
         continue;
      }

      if (HtmlThisTag (cptr, "A"))
      {
         /**********/
         /* anchor */
         /**********/

         /* lose all attributes other than what's specifically allowed */
         while (*cptr && *cptr != '>' && !isspace(*cptr)) *sptr++ = *cptr++;

         while (*cptr)
         {
            while (isspace (*cptr)) cptr++;
            if (!*cptr || *cptr == '>') break;

            /* lose everything but a known HREF */
            if (strsame (cptr, "HREF=", 5))
            {
               /* only allow this subset of known schemes to be specified */
               if (strsame (cptr+6, "http://", 7) ||
                   strsame (cptr+6, "https://", 8) ||
                   strsame (cptr+6, "ftp://", 6) ||
                   strsame (cptr+6, "mailto:", 7))
               {
                  *sptr++ = ' ';
                  cptr += HtmlSpanAttrib (tptr = cptr);
                  while (tptr < cptr) *sptr++ = *tptr++;
               }
               else
                  cptr += HtmlSpanAttrib (cptr);
               continue;
            }

            /* absorb the attribute */
            cptr += HtmlSpanAttrib (cptr);
         }

         continue;
      }

      if (HtmlThisTag (cptr, "B") ||
          HtmlThisTag (cptr, "BLOCKQUOTE") ||
          HtmlThisTag (cptr, "BODY") ||
          HtmlThisTag (cptr, "BR") ||
          HtmlThisTag (cptr, "EM") ||
          HtmlThisTag (cptr, "HR") ||
          HtmlThisTag (cptr, "I") ||
          HtmlThisTag (cptr, "LI") ||
          HtmlThisTag (cptr, "NOSCRIPT") ||
          HtmlThisTag (cptr, "OL") ||
          HtmlThisTag (cptr, "P") ||
          HtmlThisTag (cptr, "PRE") ||
          HtmlThisTag (cptr, "SPAN") ||
          HtmlThisTag (cptr, "STRONG") ||
          HtmlThisTag (cptr, "TABLE") ||
          HtmlThisTag (cptr, "TD") ||
          HtmlThisTag (cptr, "TH") ||
          HtmlThisTag (cptr, "TR") ||
          HtmlThisTag (cptr, "U") ||
          HtmlThisTag (cptr, "UL"))
      {
         /***********/
         /* allowed */
         /***********/

         /* lose all attributes other than what's specifically allowed */
         while (*cptr && *cptr != '>' && !isspace(*cptr)) *sptr++ = *cptr++;

         while (*cptr)
         {
            while (isspace (*cptr)) cptr++;
            if (!*cptr || *cptr == '>') break;

            /* of course not all of these apply to every tag! */
            if (strsame (cptr, "ALIGN=", 6) || 
                strsame (cptr, "ALINK=", 6) || 
                strsame (cptr, "BGCOLOR=", 8) || 
                strsame (cptr, "BORDER=", 7) || 
                strsame (cptr, "CELLSPACING=", 12) || 
                strsame (cptr, "CELLPADDING=", 12) || 
                strsame (cptr, "CLASS=", 6) || 
                strsame (cptr, "COLSPAN=", 8) || 
                strsame (cptr, "HEIGHT=", 7) || 
                strsame (cptr, "ID=", 3) || 
                strsame (cptr, "LINK=", 5) || 
                strsame (cptr, "NAME=", 5) || 
                strsame (cptr, "NOWRAP", 6) || 
                strsame (cptr, "RULES=", 6) || 
                strsame (cptr, "ROWSPAN=", 8) || 
                strsame (cptr, "STYLE=", 6) || 
                strsame (cptr, "TEXT=", 5) || 
                strsame (cptr, "VALIGN=", 7) || 
                strsame (cptr, "VLINK=", 6) || 
                strsame (cptr, "WIDTH=", 6)) 
            {
               *sptr++ = ' ';
               cptr += HtmlSpanAttrib (tptr = cptr);
               while (tptr < cptr) *sptr++ = *tptr++;
               continue;
            }

            /* absorb the attribute */
            cptr += HtmlSpanAttrib (cptr);
         }

         continue;
      }

      /**************/
      /* disallowed */
      /**************/

      cptr += HtmlSkipTagTo (cptr, ">");
   }
   *sptr = '\0';

   return (MessageCollapseLineBreaks (HtmlPtr, 1));
}

/*****************************************************************************/
/*
Return true if the next characters in the string represent the specified tag.
*/

int HtmlThisTag
(
char *sptr,
char *tagptr
)
{
   /*********/
   /* begin */
   /*********/

   /* if oops */
   if (*sptr++ != '<') return (FALSE);
   /* if end of tag */
   if (*sptr == '/') sptr++;
   while (*sptr && *tagptr)
      if (toupper (*sptr++) != toupper (*tagptr++)) return (FALSE);
   if (*tagptr || (*sptr != '>' && !isspace(*sptr))) return (FALSE);
   return (TRUE);
}

/*****************************************************************************/
/*
Return true if the next characters in the string represent the specified
end-of-tag.
*/

int HtmlThisEndTag
(
char *sptr,
char *tagptr
)
{
   /*********/
   /* begin */
   /*********/

   /* if oops */
   if (*sptr++ != '<') return (FALSE);
   /* if not end of tag */
   if (*sptr != '/') return (FALSE);
   sptr++;
   while (*sptr && *tagptr)
      if (toupper (*sptr++) != toupper (*tagptr++)) return (FALSE);
   if (*tagptr || (*sptr != '>' && !isspace(*sptr))) return (FALSE);
   return (TRUE);
}

/*****************************************************************************/
/*
Find the specified tag (to close a tag with intervening text) taking into
account quoting.  Return the number of characters consumed.
*/

int HtmlSkipTagTo
(
char *HtmlPtr,
char *CloseTag
)
{
   int  taglen;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("HtmlSkipTagTo() !AZ", CloseTag);

   cptr = HtmlPtr;
   taglen = strlen(CloseTag);
   while (*cptr)
   {
      while (*cptr && *cptr != *CloseTag &&
             *cptr != '\"' && *cptr != '\'') cptr++;
      if (*cptr == '\"')
      {
         cptr++;
         while (*cptr && *cptr != '\"') cptr++;
         if (*cptr) cptr++;
         continue;
      }
      if (*cptr == '\'')
      {
         cptr++;
         while (*cptr && *cptr != '\'') cptr++;
         if (*cptr) cptr++;
         continue;
      }
      if (taglen == 1 || strsame (cptr, CloseTag, taglen))
      {
         cptr += taglen;
         break;
      }
      cptr++;
   }
   return (cptr-HtmlPtr);
}

/*****************************************************************************/
/*
*/

int HtmlSpanAttrib (char *HtmlPtr)

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("HtmlSpanAttrib()");

   cptr = HtmlPtr;
   while (*cptr)
   {
      while (*cptr && !isspace(*cptr) && *cptr != '>' &&
             *cptr != '\"' && *cptr != '\'') cptr++;
      if (*cptr == '\"')
      {
         cptr++;
         while (*cptr && *cptr != '\"') cptr++;
         if (*cptr) cptr++;
         continue;
      }
      if (*cptr == '\'')
      {
         cptr++;
         while (*cptr && *cptr != '\'') cptr++;
         if (*cptr) cptr++;
         continue;
      }
      break;
   }
   return (cptr-HtmlPtr);
}

/*****************************************************************************/
/*
Ensure all potentially 'interesting' and numeric entities are revealed in order
to prevent characters they might represent from being 'hidden' from HTML
stripping.  Done in situ because resultant text will always be smaller or the
same size.  Return number of characters remaining.
*/

int HtmlNumericDeEntify (char *HtmlPtr)

{
   int  enval;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("HtmlNumericDeEntify() !AZ", HtmlPtr);

   cptr = sptr = HtmlPtr;
   while (*cptr)
   {
      if (*(unsigned short*)cptr != '&#')
      {
         *sptr++ = *cptr++;
         continue;
      }
      if (!memcmp (cptr, "&#8194;", 7) ||
          !memcmp (cptr, "&#8195;", 7) ||
          !memcmp (cptr, "&#8216;", 7) ||
          !memcmp (cptr, "&#8217;", 7) ||
          !memcmp (cptr, "&#8220;", 7) ||
          !memcmp (cptr, "&#8221;", 7) ||
          !memcmp (cptr, "&#8230;", 7))
      {
         /* a few, common UTF-8 exceptions */
         *sptr++ = *cptr++;
         continue;
      }
      /* convert a numeric entity to it's character equivalent */
      cptr += 2;
      if (*cptr == 'x')
         enval = strtol(cptr+1, NULL, 16);
      else
         enval = atol(cptr);
      while (*cptr && *cptr != ';') cptr++;
      /* if badly formed entity */
      if (!*cptr) break;
      cptr++;
      /* only cater for 8 bits, untranslated character an &iquest; */
      if (enval <= 0 || enval >= 256) enval = 191;
      *sptr++ = enval;
   }
   *sptr = '\0';

   return (sptr-HtmlPtr);
}

/*****************************************************************************/
/*
Convert known HTML character entities to the character equivalents.  Unknown
ones are converted to question marks.  Done in situ because resultant text will
always be smaller or the same size.  Return number of characters remaining. 
For creating plain-text from HTML.
*/

int HtmlCharacterDeEntify (char *HtmlPtr)

{
   static struct  {
      int len;
      char *en;
      char ch;
   } HtmlEnt [] = {
      { 5, "&amp;", '&' },
      { 6, "&nbsp;", ' ' },
      { 4, "&lt;", '<' },
      { 4, "&gt;", '>' },
      { 6, "&quot;", '\"' },
      { 6, "&apos;", '\'' },
      { 7, "&rsquo;", '\'' },
      { 7, "&lsquo;", '\'' },
      { 6, "&iexl;", 161 },
      { 6, "&cent;", 162 },
      { 7, "&pound;", 163 },
      { 8, "&curren;", 164 },
      { 6, "&copy;", 169 },
      { 5, "&reg;", 174 },
      { 8, "&iquest;", 191 },
      /* some common UTF-8s allowed through */
      { 7, "&#8194;", ' '  },
      { 7, "&#8195;", ' ' },
      { 7, "&#8216;", '\'' },
      { 7, "&#8217;", '\'' },
      { 7, "&#8220;", '\"' },
      { 7, "&#8221;", '\"' },
      { 7, "&#8230;", '.' },
      { 0, NULL, 0 }
   };

   int  idx;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis ("HtmlCharacterDeEntify() !AZ", HtmlPtr);

   cptr = sptr = HtmlPtr;
   while (*cptr)
   {
      if (*cptr != '&')
      {
         *sptr++ = *cptr++;
         continue;
      }
      for (idx = 0; HtmlEnt[idx].len; idx++)
         if (!memcmp (cptr, HtmlEnt[idx].en, HtmlEnt[idx].len)) break;
      if (HtmlEnt[idx].len)
      {
         /* known entity - substitute character */
         cptr += HtmlEnt[idx].len;
         *sptr++ = HtmlEnt[idx].ch;
      }
      else
      {
         /* unknown entity - represent using an &iquest; */
         while (*cptr && *cptr != ';') cptr++;
         /* if badly formed entity */
         if (!*cptr) break;
         cptr++;
         *sptr++ = 191;
      }
   }
   *sptr = '\0';

   return (sptr-HtmlPtr);
}

/*****************************************************************************/

